#! /usr/bin/env python
#from SEP import *
import commands as cmd
import re,string
import sys, os
import os.path as path


sys.path.append('./pycode')
from pyLocLmuv3 import *


import pyInvmodRegOneDir
import pyfirstiterRegOneDir

# ----------------------------------------------------------------
# ----- Project specific Definition -------- *** Need Changing ***
# ----------------------------------------------------------------

# Please make sure pyAdjoint and pyForward include the proper pyLOC.py file that contains the location of the nodes
# L = [[0],['A',GxA,GyA,GzA],['B',GxB,GyB,GzB],['C',GxC,GyC,GzC], ...]

#maindir = "."
maindir = "." # where souwfld, syndat, Input, bin 
projdir = maindir + "/LSM/H"            # Project directory, where all intermediate output stored
plet = 'H'                                  # Project letter (to avoid overwrite in scr)

# ---- OBS Horizontal location -----------------------------------
K=[1,2,3,4,5,6,7,8,9,10,11,12,13,14,15]                        # shot number to calculate

# ---- File names and project directories ----------------
velfile= 'zigsmt.H'              # vel file (located at projdir/Input/.)
parAdj= 'AdjOffLmuF.par'                   # Adjoint param file (located at projdir/Par/.)
parFor= 'ForOffLmuF.par'                   # Forward param file (located at projdir/Par/.)
dirdata = 'syndat/H/sdNoiH'

soufile =  'souwfld/H/swLmuH' 
DWfile  = 'syndat/DatWgt1/tmWgtdnF'       # Data weighting files
MWfile  =  'souwfld/MWfile/MWdnF'         # Model weighting files

# ---------- Preconditioning and Regularization -----------------------
Rname = 'Iden.x'                           # Program name for regularization operation
eps   = 0.2                                # Epsilon is 20% between ratio of initial ggmDggm and ggDgg
ratio = 0.                                 # Initialization for ggdDggd/ggmDggm
smooth = 1                                    
lpfil  = 0
gauss  = 0
VDer   = 0
DW     = 1                                 # Data space weighting

# ---- Image space parameters ---------------
modax = 3                                  # number of axis in image file
nmz = 280                                   # nz 
omz = 300.                                # oz
dmz = 15.                                  # dz
nmx = 799.                               # nx
omx = 0.                                # ox
dmx = 15.                                  # dx
nhx  = 0                                   # Number of subsurface offset slice.


# --- Data space parameters ----------------
datax = 2                               # number of axis in data space

# ----- Iteration loop ----------------------
biter = 1                             # Beginning iteration
eiter = 51                           # End iteration


# ----------------------------------------------------------------
# ------- End of Definition --------------------------------------
# ----------------------------------------------------------------

dirm  = projdir+'/m/mLmu'+plet
dirs  = projdir+'/s/sLmu'+plet
dirdm = projdir+'/dm/dmLmu'+plet
dirgg = projdir+'/gg/ggLmu'+plet
dirss = projdir+'/ss/ssLmu'+plet
dirrr = projdir+'/rr/rrLmu'+plet

dirggm = projdir+'/ggm/ggmLmu'+plet
dirssm = projdir+'/ssm/ssmLmu'+plet
dirrrm = projdir+'/rrm/rrmLmu'+plet
dmm    = projdir + '/dm/dmMLmu'+plet

os.system("cd ..")
os.system("make %s/modcgstep.x"%(sbin))
os.system("make %s/rtmOff2d_Main.x"%(sbin))
os.system("make %s/ForOff2d_Main.x"%(sbin))
os.system("make %s/updateAB.x"%(sbin))
os.system("make %s/Solver_ops.x"%(sbin))
os.system("make %s/ApplyWgt.x"%(sbin))
os.system("make %s/ModWgt.x"%(sbin))

cmd1= 'make %s/%s'%(sbin,Rname)
os.system(cmd1)




if (biter == 1):
	pyfirstiterRegOneDir.run(maindir,projdir,velfile,parFor,parAdj, plet,K,nmz,omz,dmz,nmx,omx,dmx,nhx,modax,datax,dirdata,smooth,lpfil,VDer, gauss, DW, soufile, MWfile, DWfile, Rname,eps,ratio)
	biter = biter+1


# At this point, iter 1 is already finished, so calculate ratio
ratio = pyInvmodRegOneDir.Ratio(maindir,projdir,plet,K,Rname,modax,datax,1)
print 'ggmDggm vs ggdDggd ratio', ratio

for iter in range(biter,eiter+1):



	print 'Running iteration#', iter
	
	# Step 2: Loop shots  dm = F'Rd  + eps A' Rm
	if (iter != -1*biter):
		pyInvmodRegOneDir.Adjoint(maindir, projdir, velfile, parAdj, plet, K, smooth, lpfil,VDer, gauss, DW, soufile, MWfile, DWfile, Rname, eps, ratio, modax, nhx, iter)

        # Step 3: Calculate dmTotal
	if (iter != -1*biter):
		pyInvmodRegOneDir.Add(dirdm, iter, modax, K,dmm)

	# Step 4: Loop shots  gg = Fdm and ggm= eps A dm        
	if (iter != -1*biter):
		pyInvmodRegOneDir.Forward(maindir, projdir, velfile, parFor, plet, K, smooth, lpfil,VDer, gauss, DW, soufile, MWfile, DWfile, Rname, eps, ratio, modax, nhx, iter)
			        
        
        
        # Step 5: Calculate the alpha and beta in conjugate direction
        alpha, beta, resid = pyInvmodRegOneDir.DPcalCD(projdir, plet, K,iter,datax,modax, dirdm, dirggm, dirgg, dirrr, dirrrm, dirss, dirssm)
	print alpha, beta, resid

	

	#  Step 6: Update the residual files
	print 'update data residual, iter', iter
	for i in K[:]:
        	print 'residual update, iter', iter, 'shot', i
                #  calculate your ss file ss=alpha*dr+beta*ss	
        	C1 = "%s/updateAB.x alpha=%s beta=%s naxis=%s file1=%s%siter%s.H file2=%s%siter%s.H outfile=%s%siter%s.H >/dev/null"%(sbin,alpha,beta,datax,dirgg,i,iter,dirss,i,iter,dirss,i,iter+1)
		os.system(C1)
			             
		#  update your current residual here  rr=rr+ss
		C1 = "%s/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%s%siter%s.H file2=%s%siter%s.H outfile=%s%siter%s.H >/dev/null"%(sbin,datax,dirrr,i,iter,dirss,i,iter+1,dirrr,i,iter+1)	
		os.system(C1)


	#Step 6b: Update the  model residual files
	print 'update model residual, iter', iter
	# calculate ssm = alpha*ggm 
	C1 = "%s/updateAB.x alpha=%s cg=1 naxis=%s file1=%siter%s.H outfile=%siter%s.H >/dev/null"%(sbin,alpha,modax,dirggm,iter,dirssm,iter+1)
	print C1	
	os.system(C1)  
	# Calculate your new rrm = rrm + ssm
	C1 = "%s/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%siter%s.H file2=%siter%s.H outfile=%siter%s.H >/dev/null"%(sbin,modax,dirrrm,iter,dirssm,iter+1,dirrrm,iter+1)	
	print C1
	os.system(C1)

	
	# Step 7: Update the model files
	# Calculate your new dm s=alpha*dm+beta*s
	print 'model update, iter', iter
	C1 = "%s/updateAB.x alpha=%s beta=%s naxis=%s file1=%sTotaliter%s.H file2=%sTotaliter%s.H outfile=%sTotaliter%s.H >/dev/null"%(sbin,alpha,beta,modax,dirdm,iter,dirs,iter,dirs,iter+1)	
	print C1
	os.system(C1)  
	# Calculate your new m
	C1 = "%s/updateAB.x alpha=1.0 beta=1.0 naxis=%s file1=%sTotaliter%s.H file2=%sTotaliter%s.H outfile=%sTotaliter%s.H >/dev/null"%(sbin,modax,dirm,iter,dirs,iter+1,dirm,iter+1)	
	print C1
	os.system(C1)
       
        # Step 9, remove files
##	pyInvmodRegOneDir.removefiles(maindir,nblock,dboc,osx,dsx,iter,nitersave,saveshot)


# Calculate final residuals and save to report
iiter = iter+1
resid = pyInvmodRegOneDir.ResidCal(projdir, plet,K,iiter,datax,dirrr)
print  resid
file = open(projdir+'/Reportsave.txt', 'a')              # save to report
cmd1 = " %s na na %s \n"%(iter, resid)
file.writelines(cmd1);
file.close()
