program DSOzxh
! Apply an DSO operator to the input file 
! Basically output = eps * abs(hx) * m(z,x,hx)


  use sep
  use waveprop_mod
  use type_mod
  use smooth_mod

  logical                    :: adj
  integer                   :: naxis
  integer                   :: n1,n2,n3
  integer                   :: i3
  real                      :: eps, hmax
  real                      :: hx, factor
  real                      :: o1, o2, o3
  real                      :: d1, d2, d3
  real, dimension(:,:,:), allocatable :: inArray, outArray
  character(len=128)        :: label1, label2, label3
  character(len=128)        :: infile_, outfile_


  call sep_init()

  call from_param("adj", adj);
  call from_param("naxis",naxis);
  call from_param("eps",eps,1.0);

  call from_param('outfile',outfile_) 
  call from_param('infile',infile_) 

  if ((naxis .ne. 3) ) call erexit('naxis must be 3')
  
  n1=1.; n2=1.; n3=1.; 

     call from_aux(infile_, "n1", n1);    call from_aux(infile_,"d1",d1);     call from_aux(infile_, "o1",o1);
     call from_aux(infile_, "label1",label1);

     call from_aux(infile_, "n2", n2);    call from_aux(infile_,"d2",d2);     call from_aux(infile_, "o2",o2);
     call from_aux(infile_, "label2",label2);

     call from_aux(infile_, "n3", n3);    call from_aux(infile_,"d3",d3,1.);     call from_aux(infile_, "o3",o3,0.);
     call from_aux(infile_, "label3",label3);



  allocate(inArray(n1,n2,n3), outArray(n1,n2,n3))
  inArray = 0.
  outArray = 0.

  call sep_read(inArray, infile_);
  
  hmax = max(abs(o3+(n3-1)*d3),o3);
  eps = eps/hmax
  
  do i3=1, n3
      factor = eps*abs(o3+(i3-1)*d3)
      outArray(:,:,i3) = inArray(:,:,i3)*factor
  end do

         call auxputch("n1",'i',n1, trim(outfile_));     
         call auxputch("d1",'f',d1, trim(outfile_));   
         call auxputch("o1",'f',o1,trim(outfile_));
         call auxputch("label1",'s',label1,trim(outfile_));

	  call auxputch("n2",'i',n2, trim(outfile_));     
	  call auxputch("d2",'f',d2, trim(outfile_));   
	  call auxputch("o2",'f',o2,trim(outfile_));
          call auxputch("label2",'s',label2,trim(outfile_));

	  call auxputch("n3",'i',n3, trim(outfile_));     
	  call auxputch("d3",'f',d3, trim(outfile_));   
	  call auxputch("o3",'f',o3,trim(outfile_));
          call auxputch("label3",'s',label3,trim(outfile_));
     
     call sep_write(outArray, outfile_);

end program
