program EstFil_Main
! Estimate a convolution filter
!      a*b = c
! convolution expressed in matrix multiplication
!      aB=c
! input:
!      na = length of filter
!     bfile = filename for b series
!     cfile = filename for c series
! Output:
!     out = filename for a series


  use sep
  use dottest_mod
  use solver_tiny_mmod
  use cgstep_mod
  use EstFil_mod
  
  
  implicit none
  
  

  
  character(len=128)    :: bfile_,cfile_, cout_
  
  logical              :: dott
  integer              :: op, na, nb, nc, niter, stat
  integer              :: ia, ib, ic
  
  real                 :: dc, oc
  real, pointer :: a(:), b(:), c(:), cout(:)
  
  call sep_init()
  
  call from_param("dottest",dott,.false.)

  call from_param("bfile",bfile_);   call from_param("cfile",cfile_);
  call from_param("cout",cout_,"cout.H");
  call from_param("niter", niter,10)
  call from_param("op",op,1)
  
  call from_aux(bfile_,"n1",nb); 
  call from_aux(cfile_,"n1",nc);  call from_aux(cfile_,"d1",dc); call from_aux(cfile_,"o1",oc);
  
  if (op .eq. 1) then
    call from_param("na",na,10);   ! length of filter a
  else
    na = nc*nc
  end if

  if (op .eq. 1) then    
    call to_history("n1",na); call to_history("d1",1.); call to_history("o1",0.)
  else
      call to_history("n1",nc); call to_history("d1",1.); call to_history("o1",0.)
    call to_history("n2",nc); call to_history("d2",1.); call to_history("o2",0.)      
  end if 
 
  call auxputch("n1","i",nc,cout_); call auxputch("d1",'f',dc,cout_); call auxputch("o1",'f',oc,cout_);
  
  call sep_close()  
  
  write(0,*) '**************************'
  write(0,*) 'dottest, op', dott, op
  write(0,*) 'na, nb, nc', na, nb, nc
  write(0,*) 'bfile', bfile_
  write(0,*) 'cfile', cfile_
  write(0,*) '**************************'
  
  
  if (nb .gt. 1000000) call erexit('nb > 1e6')
  if (nc .gt. 1000000) call erexit('nc > 1e6')
  if (na .gt. 10000000) call erexit('na > 1e6')
  
  allocate(a(na),b(nb),c(nc), cout(nc))
  
  do ia=1,na
     a(ia) = 0.
  end do
  do ib=1,nb
     b(ib) = 0.
  end do
  do ic=1,nc
     c(ic) = 0.
     cout(ic) = 0.
  end do
    
  call sep_read(b,bfile_)
  call sep_read(c,cfile_)
  write(0,*) 'sum b c', sum(b), sum(c)
  
  call estfil_init(na, nb, nc, b);
  
  if (dott) then ! perform dot product test
     write(0,*) "dottest"
     if (op .eq. 1) then
       call dottest(estfil_lop,a,c)
     else
       call dottest(estfil2_lop,a,c)
     end if
     write(0,*) "dottest OK"
  
  else ! perform inverse    
      write(0,*) 'calling solver_smp'  
      if (op .eq. 1) then
        call solver_tiny(a,c,estfil_lop,cgstep,niter)
      else 
        call solver_tiny(a,c,estfil2_lop,cgstep,niter)   
      end if         
  end if
  
        call sep_write(a)
        
  ! Now apply filter to b to get cout
  stat = estfil2_lop(.false.,.false.,a,cout)
  call sep_write(cout,cout_);      
  call estfil_close()
  
  write(0,*) 'deallocate'
!  deallocate(a)
!  deallocate(a, stat=stat)
!  if(stat /=0) call erexit("cant deallocate a")
 ! deallocate(b, stat=stat)
!  if(stat /=0) call erexit("cant deallocate b")
!  deallocate(c, stat=stat)
!  if(stat /=0) call erexit("cant deallocate c")  


  end program 
