program ForOff2d_Main

  use sep
  use rtmOff2d_oper
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod
  use smooth_mod


  implicit none

  type(velo) :: vel
  type(sou_grid)  :: sougrid
  type(rec_grid)  :: recgrid
  type(bc_values) :: bc

  
  logical                   :: smooth, verb, wfldout
  integer                   :: it, id, stat
  integer                   :: iz, ix, nz, nx, nt, jsnap
  integer                   :: bndx, bndz, nnz, nnx, nw, nmz, nmx, nhx, nm3
  character(len=128)        :: refl_, velout_, wfldsou_, wfldrec_, wfldmod_, out_
  character(len=128)        :: vp_, MWfile_, DWfile_
  real                      :: rho
  real, dimension(:,:), allocatable  :: tmpdata
  real, dimension(:,:,:), allocatable  :: tmpmodel


  call sep_init()


  
  
  !!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.);  call from_param("smooth",smooth,.false.);
   call from_param("wfmout", bc%wfmout,.false.);     call from_param("wfrout",bc%wfrout,.false.);
   call from_param("DW", bc%DW, .false.);
   
   !! preconditioning
   call from_param("lpfil", bc%lpfil, .false.);     call from_param("VDer", bc%VDer, .false.);
   call from_param("gauss", bc%gauss, .false.);


   ! Read velin, source and rec file and depth
   call from_param("velin", vp_);
   call from_param("wfldsou", wfldsou_);     call from_param("reflfile",refl_);
  

   ! read model and data weighting file
   call from_param("MWfile", MWfile_);     
   if (bc%DW) then  
  	 call from_param("DWfile",DWfile_); 
   end if
   
   ! read offset range
   call from_param('nhx', nhx, 0);   
   vel%h%n = nhx

   ! Read back ground velocity param
   call from_aux(vp_, "n1", vel%z%n);  call from_aux( vp_, "d1", vel%z%d);  call from_aux( vp_, "o1",vel%z%o);
   call from_aux( vp_, "n2", vel%x%n);  call from_aux( vp_, "d2", vel%x%d);  call from_aux( vp_, "o2",vel%x%o);
   call from_param('zw', vel%zw);
   nw = (vel%zw-vel%z%o)/vel%z%d + 0.5
   write(0,*) 'nw', nw
   call from_aux(refl_,"n1", nmz);  call from_aux(refl_,"n2",nmx);   call from_aux(refl_,"n3",nm3);
   



   !All shots has the same source wavelet
   call from_aux(wfldsou_, "n3", sougrid%t%n);  call from_aux(wfldsou_,"d3", sougrid%t%d);   call from_aux(wfldsou_,"o3", sougrid%t%o);
   sougrid%ntup = sougrid%t%n; 
   sougrid%dtup = sougrid%t%d;
   sougrid%otup = sougrid%t%o;

   sougrid%x%d = 1.
   call from_param('Gx',sougrid%x%o);     call from_param('Gz',sougrid%z%o);

   
   call from_param('nrx', recgrid%x%n);    call from_param('drx',recgrid%x%d );     call from_param('orx', recgrid%x%o);
   call from_param('nt', recgrid%t%n);     call from_param('dt', recgrid%t%d);      call from_param('ot', recgrid%t%o);
   call from_param('zr', recgrid%z%o);      
   
   

   ! Read in Density
   call from_param('rho', rho);

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   bc%nop = 2
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;

   ! Data weighting and Masking
   call from_param("bmask",bc%bmask,recgrid%x%o);  call from_param("emask",bc%emask,recgrid%x%o+(recgrid%x%n-1)*recgrid%x%d);


   !! Auxillary output files
   call from_param("outtag", out_);
   call from_param("velout", velout_, "velfile.H");
   call from_param("wfldrec", wfldrec_, "wfldrec.H");
   call from_param("wfldmod", wfldmod_, "wfldmod.H");
   call from_Param("jsnap", jsnap, 100);



   ! Error check


   if (jsnap .le. 0) jsnap=50
   
 !  if (recgrid%x%d .ne. vel%x%d) call erexit('rec dx does not match vel dx')
   
   if (sougrid%t%n .gt. recgrid%t%n) sougrid%t%n = recgrid%t%n

   write(0,*) 'nz, nw+nnz', vel%z%n, nw, nmz

   if (vel%z%n .ne. (nw+nmz)) call erexit('vel nz not equal to model nz + nw')
      
   if (vel%x%n .ne. (nmx)) call erexit('vel nx not equal to model nx')  
   
   if ((nm3 .ne. 1)  .and. (nm3 .ne. 2*nhx+1)) call erexit("dim 3 of refl file doesn't match")
   
   if (recgrid%t%n .gt. 9000) call erexit('too many time steps, nt < 9000')
   
 !  if (sougrid%t%d .ne. recgrid%t%d) call erexit('rec dt does not match sou dt')

   if ((recgrid%z%n .eq. 2) .and. (recgrid%z%d .eq. 0)) call erexit('Expect over/under but found dz=0')
   

   !! Definition
   vel%smooth = smooth
   sougrid%x%n = 1
   nnz = vel%z%n
   nnx = vel%x%n
   bndx = bc%bndx
   bndz = bc%bndz
   vel%z%n = vel%z%n + 2*bndz; 
!!@   vel%z%n = vel%z%n + bndz + bc%nop; 
   vel%x%n = vel%x%n+2*bndx
   nz = vel%z%n;
   nx = vel%x%n;
   nt=recgrid%t%n
   nw = (vel%zw-vel%z%o)/vel%z%d + 0.5


  
   !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nz, vel%z%d, vel%z%o
   write(0,*) 'vel: nx, dx, ox',  nx, vel%x%d, vel%x%o
   write(0,*) 'vel: zw, nw'         ,  vel%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sougrid%x%n, sougrid%x%d, sougrid%x%o  
   write(0,*)  'sou: nt, dt',     sougrid%t%n, sougrid%t%d
   write(0,*) 'rec: nx, dx, ox', recgrid%x%n, recgrid%x%d, recgrid%x%o
   write(0,*) 'nt, dt',          recgrid%t%n, recgrid%t%d
   write(0,*) 'nrz', recgrid%z%n
   write(0,*) 'bcOp, verb', bc%Op, verb
   write(0,*) 'smooth, wfldout, residout', smooth, bc%wfldout, bc%residout
   write(0,*) 'orz, drz', recgrid%z%o, recgrid%z%d
   write(0,*) 'bndz, bndx, nop', bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco', bc%xtapco, bc%ztapco
   write(0,*) 'jsnap', jsnap
   write(0,*) 'lpfil, VDer,guass', bc%lpfil, bc%VDer, bc%gauss
   write(0,*) 'DW, bmask, emask', bc%DW, bc%bmask, bc%emask
   write(0,*) '**************************'

   !----------------------------------------------------------------

   ! Output file
   
  
   call auxputch("n1",'i', recgrid%x%n, trim(out_));     call auxputch("label1",'s','x', trim(out_));
   call auxputch("d1",'f', recgrid%x%d, trim(out_));     call auxputch("o1",'f',recgrid%x%o, trim(out_)); 

   call auxputch("n2",'i',  nt, trim(out_));             call auxputch("label2",'s',  't', trim(out_));
   call auxputch("d2",'f', recgrid%t%d, trim(out_));           call auxputch("o2",'f', 0., trim(out_));


   ! Allocation
   write(0,*) '*** allocating tmpdata ***', recgrid%x%n,nt
   allocate(tmpdata(recgrid%x%n,nt))
      write(0,*) '*** allocating tmpmodel ***', (nnz-nw),nnx,(2*nhx+1)
   allocate(tmpmodel((nnz-nw),nnx,(2*nhx+1)))


   ! initialize
   tmpdata=0.;
   tmpmodel=0.;

   write(0,*) 'before reading refl model', refl_
   
   ! Read in reflectivity model 
   if (nm3 .eq. 1) then  ! read only zero offset cube
    call sep_read(tmpmodel(1:(nnz-nw),1:nnx,nhx+1), refl_, dim1=(nnz-nw), dim2=nnx,esize=4)
  
   else    ! read entire sub-surface offset cube     
    call sep_read(tmpmodel, refl_);
   end if

write(0,*) 'pass 3', sum(tmpmodel)


   if (bc%wfrout) then
      
!      if (exist_file("wfldrec")) then
         call auxputch("n1",'i',nz,trim(wfldrec_));   
         call auxputch("n2",'i',nx,trim(wfldrec_)); 
  !       call auxputch("n1",'i',nz-2*bndz,trim(wfldrec_));   
         call auxputch("d1",'f',vel%z%d,trim(wfldrec_));   call auxputch("o1",'f',vel%z%o,trim(wfldrec_));

  !       call auxputch("n2",'i',nx-2*bndx,trim(wfldrec_));   
         call auxputch("d2",'f',vel%x%d,trim(wfldrec_));   call auxputch("o2",'f',vel%x%o,trim(wfldrec_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldrec_));   
         call auxputch("d3",'f',recgrid%t%d*jsnap,trim(wfldrec_));   call auxputch("o3",'f',recgrid%t%o+(jsnap-1)*recgrid%t%d,trim(wfldrec_));

   end if


   if (bc%wfmout) then
!      if (exist_file("wfldmod")) then

         call auxputch("n1",'i',nz,trim(wfldmod_));   
         call auxputch("d1",'f',vel%z%d,trim(wfldmod_));   call auxputch("o1",'f',vel%z%o,trim(wfldmod_));

         call auxputch("n2",'i',nx,trim(wfldmod_));   
         call auxputch("d2",'f',vel%x%d,trim(wfldmod_));   call auxputch("o2",'f',vel%x%o,trim(wfldmod_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldmod_));   
         call auxputch("d3",'f',recgrid%t%d*jsnap,trim(wfldmod_));   call auxputch("o3",'f',recgrid%t%o,trim(wfldmod_));

   end if



! write(0,*) ' size of the output', size(tmpdata), recgrid%x%n, nt, 2
! call erexit('Exithere')

   ! Call L_initiate
   stat=L_init(sougrid, recgrid, vel, wfldsou_, vp_,rho, bc, wfldrec_in=wfldrec_, wfldmod_in=wfldmod_, verb_in=verb,jsnap_in=jsnap, velout_in=velout_, MWfile_in=MWfile_, DWfile_in=DWfile_)



   !! Forward Model: reflectivity ----> data 
   !! adj = false
   stat = L_oper(.false.,.false.,tmpmodel, tmpdata)

   call sep_write(tmpdata, out_)



     call sep_close()
     
     deallocate(tmpdata, tmpmodel);

end program
