program GenVel

! Add a water volume on top of the velocity model
! nw = grid cell for the ocean
! nsb = grid cell to start defining the seabottom
! Options to add seabed
! mode = 1 : horizontal
! mode = 2 : slanted
! slope = slope of slanted line
! mode = 3 : exponential
!! a = height of Exp bump ( in number of grid cell);  
! b= width of Exp bump (x=c +/- b when z=amp*0.37);  c=center of Exp bump

  use sep
!  use waveprop_mod
  use type_mod

implicit none




 integer    :: ix, iz,  nx, n1, nz, nw,nsb, mode
 real       :: wvel, x
 real       :: dx, dz, ox, oz
 real       :: slope, upperbound, a, b, c
 integer, dimension(:), allocatable :: zloc
 real, dimension(:), allocatable   :: zloc2
 real, dimension(:,:), allocatable :: vp, vpout
 character(len=128)                :: zlocfile_
 
 call sep_init()


   call from_param('nw', nw, 0)
   call from_param('nsb', nsb, 0)
   call from_param('wvel',wvel,1500.)
   call from_param('mode',mode,1);

   call from_param("zlocfile", zlocfile_,'zloc.H')
   



  ! Read back ground velocity param
   call from_history("n1", n1);  call from_history("d1", dz);  call from_history("o1",oz);
   call from_history("n2", nx);  call from_history("d2", dx);  call from_history("o2",ox);

  !! Read in parameter for the slanted slope
  call from_param('slope',slope,(1.0*nx)/nsb);;
  if (slope .gt. (1.0*nx)/nsb) call erexit("Slope must be less than nx/nsb")

  !! Read the parameter for the Exponential bump
  call from_param("a",a, 1.*nsb);      call from_param("b",b,(nx/8)*dx);    call from_param("c",c,(nx/4)*dx+ox);
  if (a .gt. nsb) call erexit("Height of bump(a) must be less than nsb")

   nz = nw+n1;


   write(0,*) '****************'
   write(0,*) 'nw, nsb', nw, nsb
   write(0,*) 'dx, dz', dx, dz
   write(0,*) 'mode', mode
if (mode .eq. 2) write(0,*) 'slope ', slope   
if (mode .eq. 3) write(0,*) 'a, b, c ', a, b, c   
   write(0,*) '****************'
   
   allocate( vp(n1,nx))
   allocate(zloc(nx))
   allocate(zloc2(nx))
   allocate( vpout(nz,nx))

write(0,*) 'pass 1'

   ! Read in the background velocity
    call sep_read( vp(1:n1,1:nx))
 
 write(0,*) 'pass 2'
    
   ! Define the location of the seabed
   if (mode .eq. 1) then
   	zloc(:) = nw+  nsb +1
   else if (mode .eq. 2) then
   	do ix=1,nx
   	   zloc(ix) = nw +1+   nsb + slope*(nx-ix)
   	end do
   else if (mode .eq. 3) then
   	   do ix=1,nx
              x=ox+dx*(ix-1)
              zloc(ix) = nw +1 + nsb + -a*exp(-1*((x-c)/b)**2);         
           end do
   end if

write(0,*) 'pass 3'
   
   vpout(1:nw,:) = wvel

   do ix=1,nx
!      write(0,*) ix, zloc(ix)
      do iz=nw+1,nz
         if (iz .lt. zloc(ix)) then
          vpout(iz,ix)=wvel
         else
          vpout(iz,ix)=vp(iz-nw,ix)
         end if
      end do
   end do 
     
write(0,*) 'pass 4'

   
!----------------------------------------------
! Output

  call sep_write(vpout,dim1=nz, dim2=nx, esize=4)
  
  call to_history("n1",nz); call to_history("d1",dz);  call to_history("o1",oz);
  call to_history("n2",nx); call to_history("d2",dx);  call to_history("o2",ox);

!----------------------------------------------
! Output zloc file for later use
  do ix=1,nx
      zloc2(ix) = zloc(ix)
  end do     

  call sep_write(zloc2,trim(zlocfile_))
  call auxputch("n1",'i',nx,zlocfile_); 
  call auxputch("d1", 'f', dx, zlocfile_); call auxputch("o1", 'f', ox, zlocfile_);
  call auxputch("label1", 's', 'x', zlocfile_); 
     
end program
