program ModWgt
! Output the data space weighting file
! Dimension: n1 = nx
!            n2 = nx
!            in =  Amplitude file;  outfile, weightfile          
! ampba = [0] no amplitude balancing, [1] Joint; [2] down=only; [3] up-only


  use sep
  use waveprop_mod
  use smooth_mod

  
  implicit none
  
  logical                   :: edgemute
  integer                   :: ampba 
  integer                   :: nz, nx, n3
  integer                   :: iz, ix, i3
  integer                   :: wd, igz, ibz, iez, ibx, iex, igx
  real                      :: oz, ox, o3
  real                      :: dz, dx, d3
  real                      :: amp, Eps, Epsup, Epsdn, pow, pi=3.14159265359
  real                      :: limup, limdn, Gz, Gx
  character(len=128)        :: outfile_, Wgtfile_, ampfile_
  integer, dimension(:), allocatable :: tapebd
  real, dimension(:), allocatable     :: xtapeMod, ztapeMod
  real, dimension(:,:), allocatable   ::  WgtArray, Wblank, CtapeMod
  real, dimension(:,:,:), allocatable   ::  ampArray, inArray, outArray

  call sep_init()

  call from_param("edgemute", edgemute, .true.);  
  call from_param("ampba", ampba, 0);
  call from_param("pow", pow, 0.);
  call from_param("outfile", outfile_);
  call from_param("Wgtfile", Wgtfile_);
  if (ampba .gt. 0) call from_param("ampfile", ampfile_);
  call from_param('Gx',Gx,0.);   call from_param('Gz',Gz,0.);
  


  call from_history("n1",nz);     call from_history("d1",dz);    call from_history("o1",oz);
  call from_history("n2", nx);     call from_history("d2",dx);    call from_history("o2",ox);
  call from_history("n3", n3,1);

  ! Errror Check
  if ((ampba .lt. 0) .or. (ampba .gt. 3)) call erexit('Ampba: [0] No amp balance, 1=Joint, 2=down, 3=up')

  allocate(ampArray(nz,nx,2))
  allocate(inArray(nz,nx,n3), outArray(nz,nx,n3), WgtArray(nz,nx), Wblank(nz,nx), tapebd(nx))

  if (edgemute) then
    wd  = 10   ! Model tapering width
    allocate(xtapeMod(wd), ztapeMod(wd))
    allocate(ctapeMod(wd,wd))
    xtapeMod=0.; ztapeMod=0.;
    ctapeMod=0.;
  end if 

  ! initialization
  WgtArray = 1.
  Wblank = 1.
  outArray = 0.

  write(0,*) 'nz, dz, oz', nz, dz, oz
  write(0,*) 'nx, dx, ox', nx, dx, ox
  write(0,*) 'n3, Gx, Gz' , n3, Gx, Gz
  write(0,*) 'edgemute, ampba', edgemute, ampba
  write(0,*) 'pow', pow

  call sep_read(inArray, dim1=nz, dim2=nx, esize=4) 

  write(0,*) 'pass1', nz, nx, 2
  write(0,*) 'ampfile', ampfile_

  if (ampba .ne. 0) then
 	 call sep_read(ampArray,ampfile_, dim1=nz, dim2=nx, dim3=2, esize=4)
  end if

  write(0,*) 'pass 2'

  ! Amplitude balancing

  write(0,*) 'pass 3'

    Epsup = 2.0e2 ;   Epsdn=1.e1
    limup = 0.;       limdn=3.e5

  if (ampba .gt. 0) then
        do ix=1,nx
           do iz=1, nz
                if (ampba .eq. 1) then 
	                amp = sqrt(ampArray(iz,ix,1)  + 2*ampArray(iz,ix,2)) 
		        Eps = max(Epsup, Epsdn)
                else if (ampba .eq. 2) then
 			amp = sqrt(ampArray(iz,ix,2))
 			if (amp .lt. limdn) amp=0.
			Eps = Epsdn 
                else if (ampba .eq. 3) then
			amp = sqrt(ampArray(iz,ix,1))  
			Eps = Epsup
                end  if
                
                if (amp .eq. 0) then
                    WgtArray(iz,ix) = 0.
                 !   write(0,*) 'exclude', iz, ix
                else
  		   WgtArray(iz,ix) = WgtArray(iz,ix)/(amp+Eps)            
                end if

           end do
        end do
    end if

   write(0,*) 'pass 4'


    ! Mute out the edges for WgtArray
    if (edgemute) then
       call WmodeltaperSetup2(wd, wd, ztapeMod, xtapeMod, ctapeMod, 0.5, 0.5)
       call taperingCos(nz,nx,wd,wd,ztapeMod,xtapeMod,ctapeMod, WgtArray)
    end if

 write(0,*) 'pass 5'

  ! Apply a power factor to the weight
  if (pow .gt. 0.) then
     do iz=1, nz

        WgtArray(iz,:) = WgtArray(iz,:)*(1.0+ 1.0*iz/nz)**pow

     end do
  end if

  write(0,*) 'pass 5.5'

  if (ampba .eq. 3) then
  ! Apply a cos tapering to zero near to top 
    do ix=1,nx
       tapebd(ix) = 50 - 50*sin(pi*(ix-1)/(nx-1))**2
    end do
    call WmodeltaperSetup(tapebd,50,Wblank,nz,nx,0.55,0.45)
    WgtArray = WgtArray*Wblank
  end if
  
  write(0,*) 'pass 6'


   ! Smooth the weighting matrix
    if (2 .eq. 2) then
      write(0,*) 'smoothing now'
      call triangle2(10, 10, nz, nx, WgtArray, WgtArray)
      call triangle2(10, 10, nz, nx, WgtArray, WgtArray)
    end if

   if (ampba .eq. 3) then
    ! Mute out the node position in WgtArray
	igx = (Gx - ox)/dx + 1.5       
	ibx = max(1,igx-20);  iex = min(nx,igx+20) 
	igz = (Gz - oz)/dz + 1.5 
	ibz = max(1,igz-5);  iez = min(nz,igz+10) 
	write(0,*) 'Gz, igz, oz, dz', Gz, igz, oz, dz
	write(0,*) 'ibz, iez', ibz, iez
	WgtArray(1:iez,ibx:iex) = 0.
!	WgtArray(1:iez,:) = 0.
      call triangle2(5, 5, nz, nx, WgtArray, WgtArray)
    end if



  write(0,*) 'pass 7'


    call sep_read(inArray);

  write(0,*) 'pass 8'


  do i3=1,n3
    do ix=1,nx
      do iz=1,nz
        outArray(iz,ix,i3) = inArray(iz,ix,i3)*WgtArray(iz,ix)
      end do
    end do
  end do

  write(0,*) 'pass 9'
  write(0,*) 'Wgtfile', Wgtfile_
  write(0,*) size(WgtArray,1), size(WgtArray,2), nz, nx
  write(0,*) 'outfile', outfile_
  write(0,*) size(outArray,1), size(outArray,2), nz, nx, n3

  call sep_write(WgtArray,Wgtfile_,dim1=nz, dim2=nx,esize=4)
  call sep_write(outArray,outfile_,dim1=nz, dim2=nx, dim3=n3, esize=4)

  write(0,*) 'pass 10'


         call auxputch("n1",'i',nz, trim(outfile_));     
         call auxputch("d1",'f',dz, trim(outfile_));   
         call auxputch("o1",'f',oz,trim(outfile_));
         call auxputch("label1",'s','z',trim(outfile_));

	  call auxputch("n2",'i',nx, trim(outfile_));     
	  call auxputch("d2",'f',dx, trim(outfile_));   
	  call auxputch("o2",'f',ox,trim(outfile_));
          call auxputch("label2",'s','x',trim(outfile_));

	  call auxputch("n3",'i',n3, trim(outfile_));     
	  call auxputch("d3",'f',1., trim(outfile_));   
	  call auxputch("o3",'f',0.,trim(outfile_));
          call auxputch("label3",'s',' ',trim(outfile_));

!-----------------------------------------------------------------------------
         call auxputch("n1",'i',nz, trim(Wgtfile_));     
         call auxputch("d1",'f',dz, trim(Wgtfile_));   
         call auxputch("o1",'f',oz,trim(Wgtfile_));
         call auxputch("label1",'s','z',trim(Wgtfile_));

	  call auxputch("n2",'i',nx, trim(Wgtfile_));     
	  call auxputch("d2",'f',dx, trim(Wgtfile_));   
	  call auxputch("o2",'f',ox,trim(Wgtfile_));
          call auxputch("label2",'s','x',trim(Wgtfile_));




end program
