program PhaseShiftTest2
! Pre-compute source wavefield with specification defined in a parameter file
! Source wavefield is saved in 1 file 
! Reflecting top boundary
! And sharp velocity jump
! amount of water to remove for upgoing wavefields
!
! zw    : depth of the waterbody
! zsb   : depth at the bottom of the seabed
! 
!

! Now try to simulate a reflection of the sea furface

  use sep
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod
  use smooth_mod

  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(bc_values)     :: bc

  character(len=128)   :: wfldsou1_, wfldsou2_, wfldsou3_, sou_, dat1_, dat2_, dat3_
    
    
  logical                              :: verb
  integer                              :: nx, nz1, nz2, ii, ix, iz, nt, nst, nw, nnx, nnz
  integer                              ::  izs, izsloc, ixs, ixsloc, icheck, it, ilag
  integer                              :: nop, bndx, bndz, jsnap
  integer                              :: xbsz, zbsz, irec
  real    			       :: amp, Gx, Gz, watvel, recZ
  real                                 :: dx, ox, dz, oz, dt, dst, ot
  real                                 :: idx, idz
  real                                 :: xtapco, ztapco
  real*8                               :: time1, time2
  real, dimension(:), allocatable      :: bndz0, bndz1, bndx0, bndx1, bndxx0, bndxx1, fs
  real, dimension(:), allocatable      :: xtaper, ztaper, tempASS
  real, dimension(:,:), allocatable    :: Vo, ctaper,   amp1,  ArealSou1, ArealSou2, ArealSou3
  real, dimension(:,:), allocatable    :: uo1, um1, up1, uo2, um2, up2
  real, dimension(:,:,:), allocatable  :: tempAS
  character(len=128)                   :: out_

   !! For OMP
 ! integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer            :: inode1, inode, nodes, Totalnodes

    call sep_init()

!!======================================================
  !! Parameter reading


   ! Read source and out file
   call from_param("soufile", sou_);   
  
   ! Grid Size
   call from_param("nz", nnz, 100);  call from_param("dz",dz,10.); call from_param("oz",oz,0.);
   call from_param("nx", nnx, 100);  call from_param("dx",dx,10.); call from_param("ox",ox,0.);
  
   ! Read position of the source from command line
   call from_param('Gx',Gx, 0.5*((nnx-1)*dx+ox));     call from_param('Gz',Gz, ((nnz-2)*dz+oz));
   call from_param("recZ",recZ,oz)
   izsloc =  (Gz-oz)/dz + .5
   ixsloc =  (Gx-ox)/dx + .5
   irec = int(max((recZ-oz)/dz + .5,1.))
   
   
   call from_aux(sou_, "n1", nst);  call from_aux(sou_,"d1", dst); 
    

   
   ! Read time information
   call from_param("nt",nt,2500);  call from_param("dt",dt,0.001);   call from_param("ot",ot,0.)
   call from_param("ilag", ilag,0);

   

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop", bc%nop,8);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;


   !! Auxillary output files
   call from_param("wfldsou1", wfldsou1_, "wfldsou1.H");
   call from_param("wfldsou2", wfldsou2_, "wfldsou2.H");
   call from_param("wfldsou3", wfldsou3_, "wfldsou3.H");
   call from_param("dat1file", dat1_, "dat1.H");   call from_param("dat2file", dat2_, "dat2.H")
   call from_param("dat3file", dat3_, "dat3.H");
   call from_Param("jsnap", jsnap, 100);

 

    nop = bc%nop
    bndx = bc%bndx
    bndz = bc%bndz  
    xtapco = bc%xtapco
    ztapco = bc%ztapco
    nx = nnx + 2*bndx
        
    ! Parallel Programming
    Totalnodes = omp_get_num_procs();
    nodes = min(Totalnodes-1,4);
!    nodes = Totalnodes-1;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes
    

 !*********************************
   write(0,*) '**************************'
   write(0,*) 'nnz, dz, oz ', nnz, dz, oz
   write(0,*) 'nnx, dx, ox ', nnx, dx, ox
   write(0,*) 'Gx, Gz, recZ',           Gx, Gz, recZ
   write(0,*) 'ixsloc, izsloc',      ixsloc, izsloc
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'bc%nop, verb', bc%Op, bc%nop, verb
   write(0,*) 'bndz, bndx, nop',  bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco',   bc%xtapco, bc%ztapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'ilag',      ilag
   write(0,*) '**************************'


   !! Error checking
   if (dt .ne. dst) call erexit("dt should equal to dst")
   if ((nop .ne. 8) .and. (nop .ne. 3) .and. (nop .ne.4) ) call erexit('nop must be 3, 4, or 8')

 !******************************************************************
 ! Part Zero: Read in source wavelet



   allocate(fs(nt))
   fs = 0.;

   ! Read in the source wavelet
   if (nst .lt. nt) then
       call sep_read( fs(1:nst), sou_)
   else 
       call sep_read( fs(1:nt), sou_)
   end if



   call SecDer(fs,fs,nt,dt)

   do it=1,nt
	fs(it) = fs(it)/1.e3
   end do


!   call sep_write(fs, "fs.H");
!	   call auxputch("n1",'i',nst,'fs.H');  
! 	   call auxputch("d1",'f',0.002,'fs.H');   
!         call auxputch("o1",'f',-0.40,'fs.H');
! call erexit('Exit here for debug')




  !******************************************************************

! This is for part I
    
    nz1 = nnz+bndz+nop
    nz2 = nnz + 2*bndz 
    nx = nnx + 2*bndx

    write(0,*) 'nz1, nz2, nnz, nx, nnx', nz1, nz2, nnz, nx, nnx
    write(0,*) 'bndz, bndx', bndz, bndx

   ! Memory Allocation
   allocate(bndx0(nz2),bndx1(nz2)) 
   allocate(bndxx0(nz1),bndxx1(nz1)) 
   allocate(bndz0(nx),bndz1(nx))      
   allocate(ztaper(bndz))
   allocate(xtaper(bndx))
   allocate(ctaper(bndz,bndx))
   allocate(ArealSou1(nnx,nt))
   allocate(ArealSou2(nnx,nt))
   allocate(ArealSou3(nnx,nt))

   allocate(tempAS(nx,nt,3), tempASS(nt))
   allocate(up1(nz1,nx), um1(nz1,nx), uo1(nz1,nx))
   allocate(up2(nz2,nx), um2(nz2,nx), uo2(nz2,nx))   
   allocate(Vo(nz2,nx))
 
  watvel = 2000.
  Vo = watvel**2
  
    amp = 1.0
    
    write(0,*) 'Amp', amp
    
    
    !! Absorbing BC and Tapersetup  

    call  abcsetup(nx,nz1,Vo,dt,dx, dz, nop, bndz0,bndz1,bndxx0,bndxx1,.true.)     
    call  abcsetup(nx,nz2,Vo,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,.true.)     

    call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)
    
     idx = (dt/dx)**2;   idz = (dt/dz)**2
       ixs = bndx + ixsloc
       izs = nop + izsloc
       icheck = nop +1

       ! initializing the array
       do ix=1,nx
          do iz=1,nz1
             up1(iz,ix)=0.; um1(iz,ix)=0.; uo1(iz,ix)=0.;
          end do          
          do iz=1,nz2
             up2(iz,ix)=0.; um2(iz,ix)=0.; uo2(iz,ix)=0.;
          end do   
       end do

    xbsz = 60
    zbsz = 60

    write(0,*) 'sum Vo, idx, idz', sum(Vo), idx, idz
    write(0,*) 'dx,dz,dt', dx, dz, dt

       do it=1,nt

           if (it .eq. 101) write(0,*) 'Pass 1 it,  sum', it, sum(uo1), sum(um1), sum(up1)                                
              call laplac2d16(.true., um1, uo1, up1, Vo, nx, nz1, xbsz, zbsz, idx, idz, nop, it, nodes)
              call laplac2d16(.true., um2, uo2, up2, Vo, nx, nz2, xbsz, zbsz, idx, idz, nop, it, nodes)              

           if (it .eq. 101) write(0,*) '@Pass 2 it,  sum', it, sum(uo1), sum(um1), sum(up1)

           !! time step
           do ix=1,nx
              do iz=1,nz1
                 um1(iz,ix)=uo1(iz,ix); 
                 uo1(iz,ix)=up1(iz,ix);                 
              end do      
              do iz=1,nz2
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);    
              end do
           end do
             
           ! Injecting 
            uo1(izs,ixs) =  uo1(izs,ixs) + fs(it)  

            if (it .lt. 2000) then
            tempAS(1:nx,it,1) = uo1(icheck,1:nx);
            tempAS(1:nx,it,2) = um1(icheck,1:nx);
            tempAS(1:nx,it,3) = up1(icheck,1:nx);
	    uo2(icheck-nop+bndz,1:nx) = uo2(icheck-nop+bndz,1:nx) + uo1(icheck,1:nx) 
	    um2(icheck-nop+bndz,1:nx) = um2(icheck-nop+bndz,1:nx) + um1(icheck,1:nx)
	    up2(icheck-nop+bndz,1:nx) = up2(icheck-nop+bndz,1:nx) + up1(icheck,1:nx)	    
            end if
            
!!            if (it .gt. 300) write(0,*) '@@Pass 3 it,  sum', it, sum(uo1), sum(ArealSou1)

              !! taper		
	       call taperingCosV2(nz1,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, uo1)
	       call taperingCosV2(nz1,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, up1)
	       call taperingCosV2(nz1,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, um1)
	       call taperingCos(nz2,nx,bndz,bndx,ztaper,xtaper,ctaper, uo2)
	       call taperingCos(nz2,nx,bndz,bndx,ztaper,xtaper,ctaper, up2)
	       call taperingCos(nz2,nx,bndz,bndx,ztaper,xtaper,ctaper, um2)
	       
!!            if (it .gt. 300) write(0,*) '@@@Pass 4 it,  sum', it, sum(uo1), sum(ArealSou1)
        

             !! Absorbing boundary
!             call absorbingBC1(nz1,nx,nop,bndz0,bndz1,bndx0,bndxx1,uo1,um1)
!             call absorbingBC1(nz2,nx,nop,bndz0,bndz1,bndxx0,bndx1,uo2,um2)

           ! Store areal source to memory for every  time step
              do ix=1,nnx
                    if (it .lt. nt-1) then
			    ArealSou1(ix,it) = uo1(nop+irec,bndx+ix)
			    ArealSou2(ix,it) = uo2(bndz+irec,bndx+ix) 
	  	    end if
              end do	     


           if (verb .and. (mod(it,50) .eq. 1)) then
              write(0,*) 'Pre-Sou: it, sum', it, sum(uo1), sum(ArealSou1), sum(ArealSou2)
           end if
        
           if (mod(it, 100) .eq. 1) write(0,*) 'it', it, sum(uo1), sum(uo2)
        
      
           if(mod(it,jsnap)==0 ) then
    !          write(0,*) 'uo1, dim1', size(uo1,1), size(uo1,2)
              call sep_write(uo1,wfldsou1_)
              call sep_write(uo2,wfldsou2_)
            end if
        
       end do !(end source time loop) 
       
       ! Take first derivative of tempAS(x,t,1-3)
       do ii=1,3
       do ix=1,nnx
       	tempASS(1:nt) = tempAS(ix,1:nt,ii)
!       	call triangle(5,nst,tempAS,smtAS)
        call SecDer(tempASS,tempASS,nt,1.)
	tempAS(ix,1:nt,ii) = tempASS(1:nt)
       end do
       end do


if (1 == 1) then
       do it=1,nt

              call laplac2d16(.true., um2, uo2, up2, Vo, nx, nz2, xbsz, zbsz, idx, idz, nop, it, nodes)              

           !! time step
           do ix=1,nx
              do iz=1,nz2
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);                 
              end do          
           end do
             
           ! Injecting 
	    uo2(icheck,1:nx) = uo2(icheck,1:nx) + tempAS(1:nx,it,1)
	    um2(icheck,1:nx) = um2(icheck,1:nx) + tempAS(1:nx,it,2)
	    up2(icheck,1:nx) = up2(icheck,1:nx) + tempAS(1:nx,it,3)	    

              !! taper		
	       call taperingCos(nz2,nx,bndz,bndx,ztaper,xtaper,ctaper, uo2)
	       call taperingCos(nz2,nx,bndz,bndx,ztaper,xtaper,ctaper, up2)
	       call taperingCos(nz2,nx,bndz,bndx,ztaper,xtaper,ctaper, um2)

             !! Absorbing boundary
          !   call absorbingBC1(nz2,nx,nop,bndz0,bndz1,bndx0,bndx1,uo2,um2)

           if (mod(it, 100) .eq. 1) write(0,*) 'it', it, sum(ArealSou3), sum(uo2)

           ! Store areal source to memory for every  time step
              do ix=1,nnx
                    if (it .lt. nt-1) then
			    ArealSou3(ix,it) = uo2(bndz+irec,bndx+ix) 
	  	    end if
              end do	     

           if(mod(it,jsnap)==0 ) then
              call sep_write(uo2,wfldsou3_)
            end if
        
       end do !(end source time loop) 
end if
 
	  write(0,*) ' Writing AS file' 
       	  call sep_write(ArealSou1 , trim(dat1_), dim1=nnx, dim2=nt, esize=4)
       	  call sep_write(ArealSou2, trim(dat2_), dim1=nnx, dim2=nt, esize=4)
          call sep_write(ArealSou3, trim(dat3_), dim1=nnx, dim2=nt, esize=4)
! ------------------------------------------------------------
! ------ Output file -----------------------------------------

	write(0,*) 'writing to wfldsou1', wfldsou1_
 	   call auxputch("n1",'i',nz1,trim(wfldsou1_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou1_));   
           call auxputch("n2",'i',nx,trim(wfldsou1_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou1_));  
	   call auxputch("o2",'f',ox-(bndx)*dx, trim(wfldsou1_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou1_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou1_));
     	   
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou1_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou1_));  
           call auxputch("o3",'f', ot, trim(wfldsou1_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou1_));

  	write(0,*) 'writing to wfldsou2', wfldsou2_
 	   call auxputch("n1",'i',nz2,trim(wfldsou2_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou2_));   
           call auxputch("n2",'i',nx,trim(wfldsou2_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou2_));  
	   call auxputch("o2",'f',ox-(bndx)*dx, trim(wfldsou2_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou2_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou2_));
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou2_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou2_));  
           call auxputch("o3",'f', ot, trim(wfldsou2_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou2_));

  	write(0,*) 'writing to wfldsou3', wfldsou3_
 	   call auxputch("n1",'i',nz2,trim(wfldsou3_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou3_));   
           call auxputch("n2",'i',nx,trim(wfldsou3_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou3_));  
	   call auxputch("o2",'f',ox-(bndx)*dx, trim(wfldsou3_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou3_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou3_));
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou3_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou3_));  
           call auxputch("o3",'f', ot, trim(wfldsou3_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou3_));

        write(0,*) "writing out dat1"
 	   call auxputch("n1",'i',nnx,trim(dat1_));  
   	   call auxputch("d1",'f',dx,trim(dat1_));   
           call auxputch("n2",'i',nt,trim(dat1_));  
   	   call auxputch("d2",'f',dt,trim(dat1_));  
	   call auxputch("o2",'f',ot, trim(dat1_));
   	   call auxputch("label1", 's', 'x', trim(dat1_));
     	   call auxputch("label2", 's', 't', trim(dat1_));

        write(0,*) "writing out dat2"
 	   call auxputch("n1",'i',nnx,trim(dat2_));  
   	   call auxputch("d1",'f',dx,trim(dat2_));   
           call auxputch("n2",'i',nt,trim(dat2_));  
   	   call auxputch("d2",'f',dt,trim(dat2_));  
	   call auxputch("o2",'f',ot, trim(dat2_));
   	   call auxputch("label1", 's', 'x', trim(dat2_));
     	   call auxputch("label2", 's', 't', trim(dat2_));
     	   

        write(0,*) "writing out dat3"
 	   call auxputch("n1",'i',nnx,trim(dat3_));  
   	   call auxputch("d1",'f',dx,trim(dat3_));   
           call auxputch("n2",'i',nt,trim(dat3_));  
   	   call auxputch("d2",'f',dt,trim(dat3_));  
	   call auxputch("o2",'f',ot, trim(dat3_));
   	   call auxputch("label1", 's', 'x', trim(dat3_));
     	   call auxputch("label2", 's', 't', trim(dat3_));

write(0,*) 'pass 0'

   ! --------------------------------------------------------------

   deallocate( fs)
   write(0,*) 'dealloc 1'
   deallocate( bndz0, bndz1, bndx0, bndx1)
   write(0,*) 'dealloc 2'
   deallocate( ztaper, xtaper, ctaper)
   write(0,*) 'dealloc 3'
   deallocate( up1, um1, uo1, up2, um2, uo2)
   write(0,*) 'dealloc 6'
   deallocate( ArealSou1, ArealSou2)
   deallocate(tempAS)
   call sep_close()

! Start timer
    call cpu_time(time2)
    write(0,*) 'CPU runtime is:', time2-time1, 'sec'


write(0,*) 'program should exit'

end program
