program ReflecGen

! Read in an reflectivity model and add a seabed reflector to it
! mode = 1 : horizontal
! mode = 2 : slanted
! mode = 3 : sinosoidal

  use sep
  use waveprop_mod
  use smooth_mod
  use type3d_mod

implicit none


 type(velo) :: vel

 logical    :: smooth
 integer    :: iix, iiz
 integer    :: ix, iz,  nx, nz, width, nw, mode
 real       :: nr, bound, amp, mavel, max2, tmpval
 real       :: slope, period, pi, upperbound, A
 integer, dimension(:), allocatable :: zloc
 real, dimension(:), allocatable  ::  tempV
 real, dimension(:,:), allocatable :: m, m2

 call sep_init()

   call from_param('smooth', smooth, .false.);
   call from_param('bound',bound, 100.)
   call from_param('width', width, 3)
   call from_param('nw', nw, 0)
   call from_param('mode',mode,1);
   call from_param('slope',slope,0.02);
   call from_param('period', period,200.);
   call from_param('Amp',A,5.);

   if (width .lt. 1) then 
      call erexit('Width must be greater than 1')
   end if


  ! Read original reflectivity model param
   call from_history("n1", vel%z%n);  call from_history("d1", vel%z%d);  call from_history("o1",vel%z%o);
   call from_history("n2", vel%x%n);  call from_history("d2", vel%x%d);  call from_history("o2",vel%x%o);


   pi = 3.14159
   nz = vel%z%n;
   nx = vel%x%n;

   write(0,*) '****************'
   write(0,*) 'smooth, width', smooth, width
   write(0,*) 'dx, dz', nx, nz, nw
   write(0,*) '****************'
   
   allocate( tempV(nz))
   allocate( m(nz,nx), m2(nz-nw,nx))
   allocate(zloc(nx))

   ! Read in the background velocity
    call sep_read( m(1:nz,1:nx))

    ! Smoothing velocity model
    if (smooth) then
    write(0,*) 'Inside the smoother', nz,nx, width
      call triangle2(1, width, nz, nx, m, m)
!      call triangle2(width, width, nz, nx, m, m)
    end if

   ! take Laplacian of the model
   call Laplacian2D(.true.,m,nz,nx)
   
    ! Smoothing velocity model
 !   if (smooth) then
 !   write(0,*) 'Inside the smoother', nz,nx, width
 !     call triangle2(width, width, nz, nx, m, m)
 !   end if


   
   
   
   do ix=1, nx
      do iz=1, nz-nw
!         if ((iz .lt. 15) .and. (abs(m(iz+nw,ix)) .gt. 15)) then             
!            amp=20.
!         else
!            amp=1.
!         end if
	tmpval = m(iz+nw,ix)
         m2(iz,ix) = tmpval
      end do
   end do


!   m2(1:3,:) = 0.
   m2(:,1:5) = 0.
   m2(:,nx-5:nx) = 0.
   m2(nz-nw-5:nz-nw,:) = 0.

   max2 = maxval(m2)

   m2=m2/max2

!   do ix=1,nx/30
!      do iz=1,(nz-nw)/30
!         iix= ix*30-10
!	 iiz= iz*30
!	   m2(iiz-1:iiz+1,iix-1:iix+1) = 5000.
!      end do
!   end do 
     
  
!----------------------------------------------
! Output

  call sep_write(m2,dim1=nz-nw, dim2=nx, esize=4)
  write(0,*) "vel%z%n, nw", vel%z%n, nw
  call to_history("n1",vel%z%n-nw); call to_history("d1",vel%z%d);  call to_history("o1",vel%z%o);
  call to_history("n2",vel%x%n); call to_history("d2",vel%x%d);  call to_history("o2",vel%x%o);


end program
