program SouGen2dDn
! Pre-compute source wavefield with specification defined in a parameter file
! For down-going mirror only
! amount of water to remove for upgoing wavefields
!
! zw    : depth of the waterbody
! zsb   : depth at the bottom of the seabed
! 

  use sep
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod
  use smooth_mod

  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(velo)           :: vel_grid
  type(bc_values)     :: bc

  character(len=128)   :: wfldsou2_, wfldsou1_, sou_
    
    
  logical                              :: verb, wfldout, smooth
  integer                              :: nsave, xsnap, istore, ii
  integer                              :: nx, nz0, nz1, nz, ix, iz, nt, nw,ntas, nnx, nnz
  integer                              ::  izs, izs1,izsloc, ixs, ixsloc, ishot, it, ilag
  integer                              :: nop, bndx, bndz, jsnap, ndata
  integer                              :: nst, nsb, ndelay, xbsz, zbsz
  integer                              :: izb, ize, ixb, ixe, ixzero
  real    			       :: amp, Gx, Gz, tdelay, watvel
  real                                 :: dx, ox, dz, oz, dsx, osx, orx, dt, ot
  real                                 :: rho, hx,  wvel, tt
  real                                 :: idx, idz
  real                                 :: xtapco, ztapco
  real*8                               :: time1, time2
  integer, dimension(:), allocatable   :: ithx
  real, dimension(:), allocatable      ::  fs
  real, dimension(:), allocatable      :: xtaper, ztaper
  real, dimension(:,:), allocatable    :: Vo, Vo1, ctaper, amp2,  ArealSou, ArealSou2
  real, dimension(:,:), allocatable    :: uo1, um1, up1
  real, dimension(:,:), allocatable    :: uo2, um2, up2, uuo
  real, dimension(:,:,:), allocatable  :: us2
  character(len=128)                   :: out_, ampfile_

   !! For OMP
 ! integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer            :: inode1, inode, nodes, Totalnodes

    call sep_init()

!!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.); 
   call from_param("wfldout", wfldout,.false.);   call from_param("smooth", smooth, .false.);

   ! Read back ground velocity param      
   call from_history("n1", vel_grid%z%n);  call from_history("d1", vel_grid%z%d);  call from_history("o1",vel_grid%z%o);
   call from_history("n2", vel_grid%x%n);  call from_history("d2", vel_grid%x%d);  call from_history("o2",vel_grid%x%o);

   ! Read source and out file
   call from_param("soufile", sou_);   
   call from_param("outtag", out_);
   call from_param("ampfile", ampfile_);
      
   ! Read water depth, water vel and density
   call from_param('rho', rho);
   call from_param('zw', vel_grid%zw);               ! Depth of the water column
   call from_param('zsb', vel_grid%zsb);             ! depth of the bottom of the seabed zone
     
   ! Read position of the source from command line
   call from_param('Gx',Gx);     call from_param('Gz',Gz);
   izsloc =  (Gz-vel_grid%z%o)/vel_grid%z%d + .5
   ixsloc =  (Gx-vel_grid%x%o)/vel_grid%x%d + .5
   
   call from_aux(sou_, "n1", sougrid%t%n);  call from_aux(sou_,"d1", sougrid%t%d);  
   
   ! Read time information
   call from_param("nt",nt,2500);  call from_param("dt",dt,0.001);   call from_param("ot",ot,0.)
   call from_param("tdelay",tdelay,0.);
   call from_param("ilag", ilag,0);
   call from_param("wvel",wvel);   
      
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop", bc%nop);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;

   !! Auxillary output files
   call from_param("wfldsou1", wfldsou1_, "wfldsou1.H");   
   call from_param("wfldsou2", wfldsou2_, "wfldsou2.H");
   call from_Param("jsnap", jsnap, 100);
   call from_param("xsnap", xsnap, 8);

    nx = vel_grid%x%n
    dx = vel_grid%x%d
    ox = vel_grid%x%o

    dz = vel_grid%z%d
    oz = vel_grid%z%o

    nop = bc%nop
    bndx = bc%bndx
    bndz = bc%bndz  
    xtapco = bc%xtapco
    ztapco = bc%ztapco
    nnx = nx 
    nx = nnx + 2*bndx

    nw = (vel_grid%zw - oz)/dz + .5
    nsb = (vel_grid%zsb - vel_grid%zw - oz)/dz + 0.5
    ndelay = (tdelay - ot)/dt + 0.5
    ntas=nw-nop
    nnz = vel_grid%z%n
    izs = izsloc + nop
    izs1 = izsloc + bndz

    nz0 = nnz + nop+ bndz 
    nz1 = nnz + 2*bndz 
    
    ! Parallel Programming
    Totalnodes = omp_get_num_procs();
    nodes = min(Totalnodes-1,4);
!    nodes = Totalnodes-1;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes
    

   write(0,*) '**************************'
   write(0,*) 'vel: nnz, dz, oz ', vel_grid%z%n, vel_grid%z%d, vel_grid%z%o
   write(0,*) 'vel: nnx, dx, ox',  vel_grid%x%n, vel_grid%x%d, vel_grid%x%o
   write(0,*) 'vel: zw, nw, ntas',      vel_grid%zw, nw, ntas
   write(0,*) 'vel: zsb, nsb',    vel_grid%zsb, nsb
   write(0,*) 'nz0, nz1',          nz0, nz1
   write(0,*) 'Gx, Gz',           Gx, Gz
   write(0,*) 'tdelay, ndelay',           tdelay, ndelay
   write(0,*) 'wvel',       wvel
   write(0,*) 'ixsloc, izsloc',      ixsloc, izsloc
   write(0,*) 'izs, izs1',        izs, izs1
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'bcOp, bc%nop, verb', bc%Op, bc%nop, verb
   write(0,*) ' wfldout',         wfldout
   write(0,*) 'bndz, bndx, nop',  bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco',   bc%xtapco, bc%ztapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'smooth',           smooth
   write(0,*) 'ilag',      ilag
   write(0,*) '**************************'


   !! Error checking
   if ((nop .ne. 8) .and. (nop .ne. 3) .and. (nop .ne.4) ) call erexit('nop must be 3, 4, or 8')

 !******************************************************************
 ! Part Zero: Read in velocity model first
   ! Memory Allocation
   allocate(ztaper(bndz))
   allocate(xtaper(bndx))
   allocate(ctaper(bndz,bndx))


   allocate(up1(nz0,nx), um1(nz0,nx), uo1(nz0,nx))
   allocate(up2(nz1,nx), um2(nz1,nx), uo2(nz1,nx))
   allocate(uuo(nnz-nw,nnx))
   allocate(Vo(nz0,nx), Vo1(nz1,nx))

   ! Read in the background velocity
    call sep_read( Vo1(1+bndz:nz1-bndz,1+bndx:nx-bndx))

   ! padding
    do iz=1,bndz; Vo1(   iz  ,:) = Vo1( 1+bndz,:); end do
    do iz=1,bndz; Vo1(nz1-iz+1,:) = Vo1(nz1-bndz,:); end do !padding
    do ix=1,bndx; Vo1(:,   ix  ) = Vo1(:, 1+bndx); Vo1(:,nx-ix+1) = Vo1(:,nx-bndx); end do !padding

 
       write(0,*) 'smoothing the slowness'
      call smoothslow2d(10, 10, nz1, nx, Vo1, Vo1)
 !     call smoothslow2d(3, 3, nz1, nx, Vo1, Vo1)
 !     call smoothslow2d(3, 3, nz1, nx, Vo1, Vo1)            
!      call smoothslow2dv2(10, 10, nz1/2, nx, Vo1, Vo1)
!      call smoothslow2dv2(10, 10, nz1/2, nx, Vo1, Vo1)                   
    

 if (1 == 2) then  
   call sep_write(Vo1, "Vo.H",dim1=nz1, dim2=nx,esize=4);
	   call auxputch("n2",'i',nx,'Vo.H');  
  	   call auxputch("d2",'f',dx,'Vo.H');   
           call auxputch("o2",'f', (-1*bndx)*dx+ox,'Vo.H');
	   call auxputch("n1",'i',nz1,'Vo.H');  
  	   call auxputch("d1",'f',dz,'Vo.H');   
           call auxputch("o1",'f',(-1*bndz)*dz+oz,'Vo.H');
    call erexit('exit here after Vo')
   end if


    do ix=1,nx
       do iz=1,nz1      
          Vo1(iz,ix) = Vo1(iz,ix)*Vo1(iz,ix)      
       end do
    end do


   nst = sougrid%t%n
   allocate(fs(nt))
   fs = 0.;

   ! Read in the source wavelet
   if (nst .lt. nt) then
      call sep_read( fs(1:nst), sou_)
   else
      call sep_read( fs(1:nt), sou_)
   end if

   do it=1,nt
	fs(it) = fs(it)*1000.
   end do

  if (1==2) then
   call sep_write(fs, "fs.H");
	   call auxputch("n1",'i',nst,'fs.H');  
  	   call auxputch("d1",'f',0.002,'fs.H');   
           call auxputch("o1",'f',-0.40,'fs.H');
   call erexit('Exit here for debug')
   end if

  !******************************************************************

! This is for part I 
    nx = vel_grid%x%n + 2*bndx

    write(0,*) 'nz1, nw, nsb, nop', nz1, nw, nsb
    write(0,*) 'bndz, bndx vel_grid%x%n', bndz, bndx, vel_grid%x%n


 
 
   do iz=nop+1,nz0-bndz
      Vo(iz,:) = Vo1(iz-nop+bndz,:)
   end do
         
   do iz=1,nop;  Vo(iz,:) = Vo(nop+1,:); end do ! padding
   do iz=1,bndz; Vo(nz0-iz+1,:) = Vo(nz0-bndz,:); end do !padding




 


    amp = 1.0
        
    !! Absorbing BC and Tapersetup  
    call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.95,0.05)    
    
    idx = 1./dx**2; idz = 1./dz**2; 
    if (nop .eq. 8) then
    	idx = idx*dt**2;     idz = idz*dt**2
    end if
 

       ixs = bndx + ixsloc

       ! initializing the array
       do ix=1,nx
          do iz=1,nz0
             up1(iz,ix)=0.; um1(iz,ix)=0.; uo1(iz,ix)=0.;
          end do
          do iz=1,nz1
             up2(iz,ix)=0.; um2(iz,ix)=0.; uo2(iz,ix)=0.;
          end do          
       end do

    xbsz = 60
    zbsz = 60
    
    nsave = (nt)/xsnap
    write(0,*) 'allocating us', (nnz-nw)*nnx*nsave
    allocate(us2(nnz-nw,nnx,nsave) )
    allocate(amp2(nnz-nw,nx))
     ;

       do it=1,nt+ndelay

!           if (it .gt. 101) write(0,*) 'Pass 1 it,  sum', it, sum(uo2), sum(um2)

            if (nop .eq. 2) then
              !! 4th order laplacian
              call laplac4(.true., um1, uo1,up1, Vo, dt, nx, nz0, idx, idz, nop,it)
              call laplac4(.true., um2, uo2,up2, Vo1, dt, nx, nz1, idx, idz, nop,it)              
            else if (nop .eq. 3) then
              !! 6th order laplacian
              call laplac6(.true., um1, uo1,up1, Vo, dt, nx, nz0, idx, idz, nop,it)
              call laplac6(.true., um2, uo2,up2, Vo1, dt, nx, nz1, idx, idz, nop,it)              
            else if (nop .eq. 4) then
              !! 8th order laplacian
              call laplac8(.true., um1, uo1,up1, Vo, dt, nx, nz0, idx, idz, nop,it)     
              call laplac8(.true., um2, uo2,up2, Vo1, dt, nx, nz1, idx, idz, nop,it)                                     
            else if (nop .eq. 8) then
              call laplac2d16(.true., um1, uo1, up1, Vo, nx, nz0, xbsz, zbsz, idx, idz, nop, it, nodes)
              call laplac2d16(.true., um2, uo2, up2, Vo1, nx, nz1, xbsz, zbsz, idx, idz, nop, it, nodes)
	    end if

           !! time step
           do ix=1,nx
              do iz=1,nz0
                 um1(iz,ix)=uo1(iz,ix); 
                 uo1(iz,ix)=up1(iz,ix);
              end do  
              do iz=1,nz1
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);
              end do          
           end do

!           if (it .gt. 101) write(0,*) '@Pass 2 it,  sum', it, sum(uo2), sum(um2)
             
           ! Injecting source/rec wavelet into the finite diff grid    
            if (it .lt. nt) then
              uo1(izs,ixs) =  uo1(izs,ixs) + fs(it)  
              uo2(izs1,ixs) =  uo2(izs1,ixs) + fs(it)  
            end if  
          
	   if (mod(it,100) == 1) write(0,*) 'before tapering', sum(uo1(nz0-bndz:nz0,:)), sum(uo2(nz1-bndz:nz0,:))
              !! taper		
	       call taperingCosV2(nz0,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, uo1)
	       call taperingCosV2(nz0,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, up1)
	       call taperingCosV2(nz0,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, um1)              
	       call taperingCos(nz1,nx,bndz,bndx,ztaper,xtaper,ctaper, uo2)
	       call taperingCos(nz1,nx,bndz,bndx,ztaper,xtaper,ctaper, up2)
	       call taperingCos(nz1,nx,bndz,bndx,ztaper,xtaper,ctaper, um2)
	   if (mod(it,100) == 1) write(0,*) 'after tapering', sum(uo1(nz0-bndz:nz0,:)), sum(uo2(nz1-bndz:nz0,:))


         do ix=1,nnx
            do iz=1,nnz-nw
!            uuo(iz,ix) = uo1(nop+nw+iz,ix+bndx)       
            uuo(iz,ix) = uo1(nop+nw+iz,ix+bndx)  - uo2(bndz+nw+iz,ix+bndx)          
            end do
         end do


        ! Store source wavefield to memory for every xsnap time step
        ! Calculate source amplitude for amp balancing
	do ix=1,nnx; do iz=1,nnz-nw
           amp2(iz,ix) = amp2(iz,ix) + uuo(iz,ix)**2
        end do; end do



       if ((it .gt. ndelay).and.((mod(it-ndelay,xsnap)==0) .or. (xsnap .eq. 1)) ) then
          istore = (it-ndelay)/xsnap
           write(0,*) 'outputting it istore', it, istore              
          if (istore .le. size(us2,3)) then
             do ix=1,nnx
                do iz=1,nnz-nw
       	           us2(iz,ix,istore) = uuo(iz,ix)
                end do
             end do             
          end if
       end if

           
           if (verb .and. (mod(it,50) .eq. 1)) then
              write(0,*) 'Sou: it, sum', it-ndelay, sum(uo1), sum(uo2)
           end if
        
!             if(  (it .gt. ndelay) .and.  mod(it-ndelay,jsnap)==0 .and. wfldout) then
             if( mod(it,jsnap)==1 .and. wfldout ) then
                   write(0,*) 'writing to wfldsou2', it, ndelay
		   call sep_write(uo1,wfldsou1_)
		   call sep_write(uo2,wfldsou2_)
             end if
             
             
       end do
              


       
    
 !******************************************************************
 !******************************************************************
  
 

! ------------------------------------------------------------
! ------ Output file -----------------------------------------

 
 write(0,*) 'before us2', out_
       
           call sep_write(us2 , trim(out_), dim1=nnz-nw, dim2=nnx, dim3=nsave,esize=4)

write(0,*) 'after us2'

            call sep_write(amp2, trim(ampfile_), dim1=nnz-nw, dim2=nnx,esize=4)

write(0,*) 'nz, nnz, nw', nz, nnz, nw

   call auxputch("n1",'i',nnz-nw, trim(out_)); call auxputch("d1",'f', dz, trim(out_));  call auxputch("o1",'f', vel_grid%zw, trim(out_));
   call auxputch("n2",'i', nnx, trim(out_)); call auxputch("d2",'f', dx, trim(out_));  call auxputch("o2",'f', ox, trim(out_));
   call auxputch("n3",'i', nsave, trim(out_)); call auxputch("d3",'f', dt*xsnap, trim(out_));   
   call auxputch("o3",'f', ((xsnap-1)*dt+ot), trim(out_));
   call auxputch("n4",'i', 1, trim(out_)); call auxputch("d4",'f', 1., trim(out_));  call auxputch("o4",'f', 0., trim(out_));        
   call auxputch("n5",'i', 1, trim(out_)); call auxputch("d5",'f', 400., trim(out_));  call auxputch("o5",'f', Gx, trim(out_));       
   call auxputch("label1", 's', 'nz', trim(out_));
   call auxputch("label2", 's', 'nx', trim(out_));
   call auxputch("label3", 's', 'nt', trim(out_));
   call auxputch("label4", 's', 'UD', trim(out_));
   call auxputch("label5", 's', 'shot', trim(out_));


   call auxputch("n1",'i',nnz-nw, trim(ampfile_)); call auxputch("d1",'f', dz, trim(ampfile_));  call auxputch("o1",'f', oz, trim(ampfile_));
   call auxputch("n2",'i', nnx, trim(ampfile_)); call auxputch("d2",'f', dx, trim(ampfile_));  call auxputch("o2",'f', ox, trim(ampfile_));
   call auxputch("n3",'i', 2, trim(ampfile_)); call auxputch("d3",'f', 1., trim(ampfile_));  call auxputch("o3",'f', 0., trim(ampfile_));               
   call auxputch("label1", 's', 'nz', trim(ampfile_));
   call auxputch("label2", 's', 'nx', trim(ampfile_));
   call auxputch("label3", 's', 'UD', trim(ampfile_));



 if ((wfldout) .and. (1 == 1)) then
   	write(0,*) 'writing to wfldsou1', wfldsou1_
 	   call auxputch("n1",'i',nz0,trim(wfldsou1_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou1_));   
           call auxputch("n2",'i',nx,trim(wfldsou1_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou1_));  
	   call auxputch("o2",'f',vel_grid%x%o-(bndx)*dx, trim(wfldsou1_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou1_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou1_));
           call auxputch("n3",'i',1+(nt+ndelay)/jsnap,trim(wfldsou1_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou1_));  
           call auxputch("o3",'f', -1*tdelay, trim(wfldsou1_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou1_));
  	   
  	write(0,*) 'writing to wfldsou2', wfldsou2_
 	   call auxputch("n1",'i',nz1,trim(wfldsou2_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou2_));   
           call auxputch("n2",'i',nx,trim(wfldsou2_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou2_));  
	   call auxputch("o2",'f',vel_grid%x%o-(bndx)*dx, trim(wfldsou2_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou2_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou2_));
           call auxputch("n3",'i',1+(nt+ndelay)/jsnap,trim(wfldsou2_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou2_));  
           call auxputch("o3",'f', -1*tdelay, trim(wfldsou2_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou2_));
 end if
 
 

write(0,*) 'pass 0'

   ! --------------------------------------------------------------

   deallocate( fs)
write(0,*) 'dealloc 1'
   deallocate(Vo1)
write(0,*) 'dealloc 2'
   write(0,*) 'dealloc 1'
   write(0,*) 'dealloc 2'
!   deallocate( ztaper, xtaper, ctaper)
   write(0,*) 'dealloc 3'
      deallocate( up1, um1, uo1)
   write(0,*) 'dealloc 4'
   deallocate( us2)
   write(0,*) 'dealloc 5'
   deallocate( up2, um2, uo2)
   write(0,*) 'dealloc 6'
!   deallocate( ArealSou2)
   
   call sep_close()

! Start timer
    call cpu_time(time2)
    write(0,*) 'CPU runtime is:', time2-time1, 'sec'


write(0,*) 'program should exit'

end program
