program SouGen2dStream
! Pre-compute source wavefield with streamer geometry
! Source wavefield is saved in 1 file 
! Reflecting top boundary
! And sharp velocity jump
! amount of water to remove for upgoing wavefields
!
! 
!


  use sep
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod
  use smooth_mod

  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(velo)           :: vel_grid
  type(bc_values)     :: bc

  character(len=128)   :: wfldsou1_, wfldsou2_, sou_, ASfile_
    
    
  logical                              :: verb, wfldout, ASout, abctop, smooth
  integer                              :: bcOp, nsave, xsnap, istore, shotnum
  integer                              :: nx, nz, ix, iz, nt, seed, nw, nnx, nnz
  integer                              ::  izs, izsloc, ixs, ixsloc, it
  integer                              :: nop, bndx, bndz, jsnap, ndata
  integer                              :: nst, ndelay, xbsz, zbsz
  real    			       :: amp, Gx, Gz, tdelay, watvel
  real                                 :: dx, ox, dz, oz, dsx, osx, orx, dt, ot
  real                                 :: rho
  real                                 :: idx, idz
  real                                 :: xtapco, ztapco
  real*8                               :: time1, time2
  real, dimension(:), allocatable      :: bndz0, bndz1, bndz2, bndx0, bndx1,  fs
  real, dimension(:), allocatable      :: xtaper, ztaper, xtaper2, ztaper2
  real, dimension(:,:), allocatable    :: vp2, ctaper,  ctaper2, amp1
  real, dimension(:,:), allocatable    :: uo1, um1, up1
  real, dimension(:,:,:), allocatable  :: us1
  character(len=128)                   :: out_, ampfile_

   !! For OMP
 ! integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer            :: inode1, inode, nodes, Totalnodes

    call sep_init()

!!======================================================
  !! Parameter reading



   call from_param("verb",verb,.false.);   call from_param("bcOp",bc%Op,1);  
   call from_param("wfldout", wfldout,.false.);   call from_param("ASout", ASout,.false.);
   call from_param("abctop",abctop,.true.);       call from_param("smooth", smooth, .false.);

   ! Read back ground velocity param
   

   
   call from_history("n1", vel_grid%z%n);  call from_history("d1", vel_grid%z%d);  call from_history("o1",vel_grid%z%o);
   call from_history("n2", vel_grid%x%n);  call from_history("d2", vel_grid%x%d);  call from_history("o2",vel_grid%x%o);



   ! Read source and out file
   call from_param("soufile", sou_);   
   call from_param("outtag", out_);
   call from_param("ampfile", ampfile_);

   

   
   ! Read water depth, water vel and density
   call from_param('rho', rho);
   call from_param('zw', vel_grid%zw);               ! Depth of the water column

   


   ! Source arrangement
  
   ! Read position of the source from command line
   call from_param('Gx',Gx);     call from_param('Gz',Gz);
   izsloc =  (Gz-vel_grid%z%o)/vel_grid%z%d + .5
   ixsloc =  (Gx-vel_grid%x%o)/vel_grid%x%d + .5


   
   call from_aux(sou_, "n1", sougrid%t%n);  call from_aux(sou_,"d1", sougrid%t%d);  
   
   
   ! Read time information
   call from_param("nt",nt,2500);  call from_param("dt",dt,0.001);   call from_param("ot",ot,0.)
   call from_param("tdelay",tdelay,0.);

   

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop", bc%nop);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;


   !! Auxillary output files
   call from_param("wfldsou1", wfldsou1_, "wfldsou1.H");
   call from_param("ASfile", ASfile_, "ASfile.H");
   call from_Param("jsnap", jsnap, 100);
   call from_param("xsnap", xsnap, 8);

 
    nx = vel_grid%x%n
    dx = vel_grid%x%d
    ox = vel_grid%x%o


    bndx = bc%bndx
    bndz = bc%bndz  
    
    nnz = vel_grid%z%n
    nz  = nnz + 2*bndz
    dz = vel_grid%z%d
    oz = vel_grid%z%o




    nop = bc%nop
    bcOp = bc%Op
    xtapco = bc%xtapco
    ztapco = bc%ztapco
    nnx = nx 
    nx = nnx + 2*bndx

    nw = (vel_grid%zw - oz)/dz + .5
    ndelay = (tdelay-ot)/dt+.5

    ! Parallel Programming
    Totalnodes = omp_get_num_procs();
    nodes = min(Totalnodes-1,4);
!    nodes = Totalnodes-1;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes


   ! for blocking up
   xbsz=60
   zbsz=60
    

! ********************

 
! ********************


 !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', vel_grid%z%n, vel_grid%z%d, vel_grid%z%o
   write(0,*) 'vel: nx, dx, ox',  vel_grid%x%n, vel_grid%x%d, vel_grid%x%o
   write(0,*) 'vel: zw, nw',      vel_grid%zw, nw
   write(0,*) 'Gx, Gz',           Gx, Gz
   write(0,*) 'tdelay',           tdelay, ndelay
   write(0,*) 'ixsloc, izsloc',      ixsloc, izsloc
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'bc%nop, verb', bc%nop, verb
   write(0,*) ' wfldout',         wfldout
   write(0,*) 'bndz, bndx, nop',  bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco',   bc%xtapco, bc%ztapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'smooth, ASout',           smooth, ASout 
   write(0,*) 'abctop',           abctop
   write(0,*) 'xbsz, zbsz', xbsz, zbsz
   write(0,*) '**************************'


   !! Error checking
   if ((nop .ne. 8) .and. (nop .ne. 3) .and. (nop .ne.4) ) call erexit('nop must be 3, 4, or 8')

 !******************************************************************
 ! Part Zero: Read in velocity model first

    allocate(vp2(nnz+2*bndz,nx))
    vp2=0.;
   ! Read in the background velocity
   ! No need to pad the top, it is now reflective
    call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx))
    do iz=1,bndz; vp2(   iz  ,:) = vp2( 1+bndz,:); vp2(nz-iz+1,:) = vp2(nz-bndz,:); end do !padding
    do ix=1,bndx; vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); end do !padding

    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix) = vp2(iz,ix)*vp2(iz,ix)      
       end do
    end do



     ! Smoothing velocity model
     if (smooth) then
      write(0,*) 'smoothing the velocity'
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
     end if




   nst = nt 
   allocate(fs(nst))
   fs = 0.;



   ! Read in the source wavelet
   call sep_read( fs(1:nst), sou_)
   fs(sougrid%t%n:nst)=0.;



   do it=1,nst
	fs(it) = fs(it)*10000.
   end do


! Do we need second derivative here?
       

   call sep_write(fs, "fs.H");
	   call auxputch("n1",'i',nst,'fs.H');  
  	   call auxputch("d1",'f',0.002,'fs.H');   
           call auxputch("o1",'f',-0.40,'fs.H');
 ! call erexit('Exit here for debug')




  !******************************************************************

! This is for part I

    nx = vel_grid%x%n + 2*bndx

    write(0,*) 'nz, nw, nop', nz, nw
    write(0,*) 'bndz, bndx vel_grid%x%n', bndz, bndx, vel_grid%x%n

   ! Memory Allocation
   allocate(bndz0(nx),bndz1(nx))      
   allocate(ztaper(bndz))
   allocate(xtaper(bndx))
   allocate(ctaper(bndz,bndx))
   

    amp = 1.0
    
    write(0,*) 'Amp', amp
    
    
    !! Absorbing BC and Tapersetup  


    call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)
    
    write(0,*) 'pass 1'    

    idx = 1./dx**2; idz = 1./dz**2; 
    if (nop .eq. 8) then
    	idx = idx*dt**2;     idz = idz*dt**2
    end if
 
    write(0,*) 'pass 2'    
 
    
    
 !******************************************************************
 !******************************************************************
    ! This is part 2 
    write(0,*) ' Begin Part 2'

    allocate(bndx0(nz), bndx1(nz))
    allocate(ztaper2(bndz))
    allocate(xtaper2(bndx))
    allocate(ctaper2(bndz,bndx))            

     call  abcsetup(nx,nz,vp2,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,.false.)      
     call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)
     call  tapersetupCos(bndz,bndx,ztaper2,xtaper2,ctaper2,0.8,0.2)

    write(0,*) 'nz, nx, nw', nz, nx,nw

! call erexit('Exit to check velocity')


    nsave = (nt)/xsnap
    write(0,*) 'allocating us', (nnz-nw)*nnx*nsave
    allocate( us1(nnz-nw,nnx,nsave))

    write(0,*) 'allocating 6X', nz*nx
    allocate(up1(nz,nx), um1(nz,nx), uo1(nz,nx))

    allocate(amp1(nz-nw,nx))
     

    up1=0.;  um1=0.;  uo1=0.; 
    amp1=0.;     
 


    ! -------------------------------------------------------------
    ! Pre-compute source wavefield

    ! Start timer
    call cpu_time(time1)



if (1 == 1) then    
       
       ixs = ixsloc + bndx

       ! initializing the array
       do ix=1,nx
          do iz=1,nz
             up1(iz,ix)=0.; um1(iz,ix)=0.; uo1(iz,ix)=0.;
          end do          
       end do
       

       do it=1,nt+ndelay

       ! if ((it .gt. 950)) write(0,*) 'A: it', it, sum(uo1) 


            if (nop .eq. 2) then
               !! 4th order laplacian
               call laplac4(.true., um1, uo1,up1, vp2, dt, nx, nz, idx, idz, nop,it)
            else if (nop .eq. 3) then        
	        !! 6th order laplacian
           	call laplac6(.true., um1, uo1,up1, vp2, dt, nx, nz, idx, idz, nop,it)
             else if (nop .eq. 4) then        
	        !! 8th order laplacian
           	call laplac8(.true., um1, uo1,up1, vp2, dt, nx, nz, idx, idz, nop,it)
 	    else if (nop .eq. 8) then
 	       !! 16th order laplacian
 	       call laplac2d16(.true., um1, uo1, up1, vp2, nx, nz, xbsz, zbsz, idx, idz, nop, it, nodes)
            end if
            
   !         write(0,*) 'pass 16'



! write(0,*) 'pass 17'

       ! if ((it .gt. 950)) write(0,*) '@B: it', it, sum(uo1) 

           !! time step
           do ix=1,nx
              do iz=1,nz
                 um1(iz,ix)=uo1(iz,ix); 
                 uo1(iz,ix)=up1(iz,ix);
              end do 
           end do

!write(0,*) 'pass 18'
          
          ! if ((it .gt. 950)) write(0,*) '@@C: it', it, sum(uo1)    

      
           izs = izsloc + bndz 

           if (it .lt. nst) uo1(izs,ixs) =  uo1(izs,ixs) + fs(it); 
          

           ! if ((it .gt. 950)) write(0,*) '@@@D: it', it, sum(uo1), fs(it) 

              !! taper
		
	       call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper, uo1)
	       call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper, up1)
	       call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper, um1)


       ! if ((it .gt. 950)) write(0,*) '@@@@E: it', it, sum(uo1) 

             !! Absorbing boundary
!             call absorbingBC1(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo1,um1)


       ! if ((it .gt. 950)) write(0,*) '@@@@F: it', it, sum(uo1) 


              ! Store source wavefield to memory for every xsnap time step
              ! Calculate source amplitude for amp balancing
	      do ix=1,nnx
                 do iz=1,nnz-nw
                    amp1(iz,ix) = amp1(iz,ix) + uo1(iz+bndz+nw,ix+bndx)**2

                    if ((it .gt. ndelay).and.((mod(it-ndelay,xsnap)==0) .or. (xsnap .eq. 1)) ) then
                        istore = (it-ndelay)/xsnap
                      
                        us1(iz,ix,istore) = uo1(iz+bndz+nw,ix+bndx)
                    end if

                 end do
              end do





 
           if (verb .and. (mod(it,50) .eq. 1)) then
              write(0,*) 'Sou: it, sum', it-ndelay, sum(uo1),  abctop
           end if
        
 
        	

!             if(  (it .gt. ndelay) .and.  mod(it-ndelay,jsnap)==0 .and. wfldout ) then
             if( mod(it,jsnap)==1 .and. wfldout ) then
                   write(0,*) 'writing to wfldsou1', it, ndelay
		   call sep_write(uo1,wfldsou1_)
!	           if (it .gt. 500) call erexit('exit now (it-ndelay)/jsnap')
             end if
           


       end do !(end source time loop)


end if



  
 

! ------------------------------------------------------------
! ------ Output file -----------------------------------------

 
 write(0,*) 'before us1 us2', out_
       
           call sep_write(us1 , trim(out_), dim1=nnz-nw, dim2=nnx, dim3=nsave,esize=4)

write(0,*) 'after us1 us2'

       

write(0,*) 'after ASout'
!            do ix=1,nnx
!              do iz=1,nnz-nw
!                 amp1(iz,ix) = sqrt(amp1(iz,ix))
!              end do
!            end do
 
            call sep_write(amp1, trim(ampfile_), dim1=nnz-nw, dim2=nnx,esize=4)

write(0,*) 'nz, nnz, nw', nz, nnz, nw

   call auxputch("n1",'i',nnz-nw, trim(out_)); call auxputch("d1",'f', dz, trim(out_));  call auxputch("o1",'f', vel_grid%zw, trim(out_));
   call auxputch("n2",'i', nnx, trim(out_)); call auxputch("d2",'f', dx, trim(out_));  call auxputch("o2",'f', ox, trim(out_));
   call auxputch("n3",'i', nsave, trim(out_)); call auxputch("d3",'f', dt*xsnap, trim(out_));   
   call auxputch("o3",'f', ((xsnap-1)*dt+ot), trim(out_));
   call auxputch("n4",'i', 1, trim(out_)); call auxputch("d4",'f', 1., trim(out_));  call auxputch("o4",'f', 0., trim(out_));        
   call auxputch("n5",'i', 1, trim(out_)); call auxputch("d5",'f', 400., trim(out_));  call auxputch("o5",'f', Gx, trim(out_));       
   call auxputch("label1", 's', 'nz', trim(out_));
   call auxputch("label2", 's', 'nx', trim(out_));
   call auxputch("label3", 's', 'nt', trim(out_));
   call auxputch("label4", 's', 'UD', trim(out_));
   call auxputch("label5", 's', 'shot', trim(out_));


   call auxputch("n1",'i',nnz-nw, trim(ampfile_)); call auxputch("d1",'f', dz, trim(ampfile_));  call auxputch("o1",'f', oz, trim(ampfile_));
   call auxputch("n2",'i', nnx, trim(ampfile_)); call auxputch("d2",'f', dx, trim(ampfile_));  call auxputch("o2",'f', ox, trim(ampfile_));
   call auxputch("n3",'i', 1, trim(ampfile_)); call auxputch("d3",'f', 1., trim(ampfile_));  call auxputch("o3",'f', 0., trim(ampfile_));               
   call auxputch("label1", 's', 'nz', trim(ampfile_));
   call auxputch("label2", 's', 'nx', trim(ampfile_));
   call auxputch("label3", 's', 'UD', trim(ampfile_));



 if ((wfldout) .and. (1 == 1)) then
	write(0,*) 'writing to wfldsou1', wfldsou1_
 	   call auxputch("n1",'i',nz,trim(wfldsou1_));
   	   call auxputch("d1",'f',dz,trim(wfldsou1_));   
	   call auxputch("o1",'f',vel_grid%zw-(bndz)*dz, trim(wfldsou1_));
           call auxputch("n2",'i',nx,trim(wfldsou1_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou1_));   
	   call auxputch("o2",'f',vel_grid%x%o-(bndx)*dx, trim(wfldsou1_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou1_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou1_));
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou1_));
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou1_));   
           call auxputch("o3",'f', -tdelay+(jsnap-1)*dt, trim(wfldsou1_));            
  	   call auxputch("label3", 's', 'time', trim(wfldsou1_));

 end if
 
 

write(0,*) 'pass 0'

   ! --------------------------------------------------------------

   deallocate( fs)
   write(0,*) 'dealloc 1'
   deallocate(vp2)
   write(0,*) 'dealloc 2'
   deallocate( bndz0, bndz1, bndx0, bndx1)
   write(0,*) 'dealloc 3'
!   deallocate( ztaper, xtaper, ctaper)
   write(0,*) 'dealloc 4'
   deallocate( ztaper2, xtaper2, ctaper2)
   write(0,*) 'dealloc 5'
   deallocate( us1)
   write(0,*) 'dealloc 6'
   deallocate( up1, um1, uo1)
   write(0,*) 'dealloc 7'
   
   call sep_close()

! Start timer
    call cpu_time(time2)
    write(0,*) 'CPU runtime is:', time2-time1, 'sec'


write(0,*) 'program should exit'

end program
