program SouGenJustMul
! Calculate the multiple part of the source wavefield
! This is for down-going only
! Basically, inject the down-going data from the sea-surface
! Input:
! Pdatfile_ = filename for input down-going data
! dim1 = space, dim2=time
! Output:    

  use sep
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod
  use smooth_mod

  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(velo)           :: vel_grid
  type(bc_values)     :: bc

  character(len=128)   :: Pdatfile_, wfldsou2_, ASfile_
    
    
  logical                              :: verb, wfldout, ASout, smooth
  integer                              :: nsave, xsnap, istore
  integer                              :: nx, nz, ix, iz, nt, nw, nnx, nnz
  integer                              ::  izs, izsloc, ixs, ixsloc, it, ilag
  integer                              :: nop, bndx, bndz, jsnap
  integer                              :: nsb, ndelay, xbsz, zbsz
  real    			       :: amp, Gx, Gz, tdelay
  real                                 :: dx, ox, dz, oz, dsx, osx, orx, dt, ot
  real                                 :: rho, watvel
  real                                 :: idx, idz, idt2
  real                                 :: xtapco, ztapco
  real*8                               :: time1, time2
  real, dimension(:), allocatable      :: bndz0, bndz1, bndz2, bndx0, bndx1
  real, dimension(:), allocatable      :: xtaper, ztaper, tempAS, smtAS
  real, dimension(:,:), allocatable    :: vp2, ctaper, amp1, amp2,  ArealSou
  real, dimension(:,:), allocatable    :: uo2, um2, up2, us2Der
  real, dimension(:,:,:), allocatable  :: us2
  character(len=128)                   :: out_, ampfile_

 !! For OMP
!  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer            :: inode1, inode, nodes

  ! Initial timer
  integer*4                 :: now2(3), now1(3)
  call itime(now1)


    call sep_init()

!!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.);  
   call from_param("wfldout", wfldout,.false.);   call from_param("ASout", ASout,.false.);
   call from_param("smooth", smooth, .false.);

   ! Read back ground velocity param
   call from_history("n1", vel_grid%z%n);  call from_history("d1", vel_grid%z%d);  call from_history("o1",vel_grid%z%o);
   call from_history("n2", vel_grid%x%n);  call from_history("d2", vel_grid%x%d);  call from_history("o2",vel_grid%x%o);

   ! Read source and out file
   call from_param("outtag", out_);
   call from_param("ampfile", ampfile_);
   call from_param("Pdatfile", Pdatfile_);
   
   ! Read water depth, water vel and density
   call from_param('rho', rho);
   call from_param('zw', vel_grid%zw);               ! Depth of the water column
   call from_param('zsb', vel_grid%zsb);             ! depth of the bottom of the seabed zone
   

   ! Source arrangement
  
   ! Read position of the source from command line
   call from_param('Gx',Gx);     call from_param('Gz',Gz);
   izsloc =  (Gz-vel_grid%z%o)/vel_grid%z%d + .5
   ixsloc =  (Gx-vel_grid%x%o)/vel_grid%x%d + .5
   
   call from_aux(Pdatfile_, "n1", sougrid%t%n);  call from_aux(Pdatfile_,"d1", sougrid%t%d);  

   ! Read time information
   call from_param("nt",nt,2500);  call from_param("dt",dt,0.001);   call from_param("ot",ot,0.)
   call from_param("tdelay",tdelay,0.);
   call from_param("ilag", ilag,0);

   
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop", bc%nop);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;


   !! Auxillary output files
   call from_param("wfldsou2", wfldsou2_, "wfldsou2.H");
   call from_param("ASfile", ASfile_, "ASfile.H");
   call from_Param("jsnap", jsnap, 100);
   call from_param("xsnap", xsnap, 8);

 
    nx = vel_grid%x%n
    dx = vel_grid%x%d
    ox = vel_grid%x%o

    
    dz = vel_grid%z%d
    oz = vel_grid%z%o




    nop = bc%nop
    bndx = bc%bndx
    bndz = bc%bndz  
    xtapco = bc%xtapco
    ztapco = bc%ztapco
    nnx = nx 
    nx = nnx + 2*bndx

    nw = (vel_grid%zw - oz)/dz + .5
    nsb = (vel_grid%zsb - vel_grid%zw - oz)/dz + 0.5
    
    if (tdelay .gt. 0.) then
      ndelay = (tdelay - ot)/dt + 1.5
    else 
      ndelay = 0
    end if

    
    ! Parallel Programming
!    nodes = omp_get_num_procs()-1;
!    nodes = 1;
 !   call omp_set_num_threads(nodes);
 !   inode1=0


! ********************

 
! ********************


 !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', vel_grid%z%n, vel_grid%z%d, vel_grid%z%o
   write(0,*) 'vel: nx, dx, ox',  vel_grid%x%n, vel_grid%x%d, vel_grid%x%o
   write(0,*) 'vel: zw, nw',      vel_grid%zw, nw
   write(0,*) 'vel: zsb, nsb',    vel_grid%zsb, nsb
   write(0,*) 'Gx, Gz',           Gx, Gz
   write(0,*) 'tdelay',           tdelay, ndelay
   write(0,*) 'ixsloc, izsloc',      ixsloc, izsloc
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'bc%nop, verb',     bc%nop, verb
   write(0,*) ' wfldout',         wfldout
   write(0,*) 'bndz, bndx, nop',  bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco',   bc%xtapco, bc%ztapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'smooth, ASout',           smooth, ASout 
   write(0,*) 'ilag',             ilag
   write(0,*) '**************************'


   !! Error checking
   if ((nop .ne. 8) .and. (nop .ne. 3) .and. (nop .ne.4)) call erexit('nop must be 3, 4, or 8')

 !******************************************************************
 ! Part Zero: Read in velocity model first

    nnz = vel_grid%z%n
    nz  = nnz + 2*bndz
    allocate(vp2(nnz+2*bndz,nx))
    vp2=0.;
   ! Read in the background velocity
   ! No need to pad the top, it is now reflective
    call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx))
    do iz=1,bndz; vp2(   iz  ,:) = vp2( 1+bndz,:); vp2(nz-iz+1,:) = vp2(nz-bndz,:); end do !padding
    do ix=1,bndx; vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); end do !padding

    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix) = vp2(iz,ix)*vp2(iz,ix)      
       end do
    end do


     ! Smoothing velocity model
     if (smooth) then
      write(0,*) 'smoothing the velocity'
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
     end if





  !******************************************************************

! This is for part I

    nx = vel_grid%x%n + 2*bndx

 
   ! Memory Allocation
   allocate(ArealSou(nnx,nt))
   allocate(ztaper(bndz))
   allocate(xtaper(bndx))
   allocate(ctaper(bndz,bndx))
   allocate(tempAS(nt),smtAS(nt))


    amp=0.005
    
    write(0,*) 'Amp', amp
    
    
    !! Absorbing BC and Tapersetup  

    idx = 1./dx**2; idz = 1./dz**2; idt2= 1./dt**2
 
    write(0,*) 'pass 1'

    ! Read the ArealSou (down-going data) from a file
    call sep_read(ArealSou, Pdatfile_, dim1=nnx, dim2=nt, esize=4)

    ! Taperoff the end of ArealSou to avoid sharp cutoff      
    write(0,*) '@A bndx it', bndx, bndz, it 
    write(0,*) 'size ctaper 1 2', size(ctaper,1), size(ctaper,2)
      call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.5,0.5)
      call  taperingCos(nnx,nt,bndx,bndz,xtaper,ztaper,ctaper, ArealSou)    

     watvel = sqrt(vp2(bndz+nw/2,bndx/2))
 
! ! Is this still needed?
       ! need to take a vertical derivatives of ArealSou 
       ! to simulate reflection at sea surface
       ilag = int(dz/watvel/dt/2.)
       write(0,*) 'take derivatives of ArealSou, ilag, watvel', ilag, watvel
       do ix=1,nnx
       	tempAS(1:nt) = ArealSou(ix,1:nt)
       	call FirstDer(.true.,tempAS,tempAS,nt,dt)
       
       	do it=1+ilag,nt
         	ArealSou(ix,it) = tempAS(it-ilag)
       	end do
       end do


    !   if (ASout) then
!	  write(0,*) ' Writing AS file' 
!       	  call sep_write(ArealSou , trim(ASfile_), dim1=nnx, dim2=nt, esize=4)
!       	  call erexit('Exit after looking as ArealSou')
!       end if

    
    
 !******************************************************************
 !******************************************************************
    ! This is part 2 
    write(0,*) ' Begin Part 2'
    nz = nnz + 2*bndz       !  pad the top
    
    allocate(bndz0(nx), bndz1(nx))
    allocate(bndx0(nz), bndx1(nz))

     call  abcsetup(nx,nz,vp2,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,.false.)      
     call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)

    idx = 1./dx**2; idz = 1./dz**2; 
    if (nop .eq. 8) then
    	idx = idx*dt**2;     idz = idz*dt**2
    end if
 
 
    write(0,*) 'nz, nx, nw', nz, nx,nw


    !! Note: vp2 is not squared from the input file, but squared in Part I !!!


    call sep_write(vp2,"vp2.H",dim1=nz,dim2=nx,esize=4)


     	   call auxputch("n1",'i',nz,'vp2.H');  
   	   call auxputch("d1",'f',dz,'vp2.H');   
           call auxputch("n2",'i',nx,'vp2.H');  
   	   call auxputch("d2",'f',dx,'vp2.H');  
   	   call auxputch("label1", 's', 'X', 'vp2.H');
     	   call auxputch("label2", 's', 't', 'vp2.H');
! call erexit('Exit to check velocity')


    nsave = (nt)/xsnap
    write(0,*) 'allocating us', (nnz-nw)*nnx*nsave
    allocate(  us2(nnz-nw,nnx,nsave) )

    write(0,*) 'allocating 6X', nz*nx
    allocate(up2(nz,nx), um2(nz,nx), uo2(nz,nx))
    allocate( us2Der(nz,nx))

    allocate(amp2(nz-nw,nx))
     


    up2=0.;  um2=0.;  uo2=0.;
    us2Der=0.;
    amp2=0.;
 
    xbsz = 60
    zbsz = 60

    ! -------------------------------------------------------------
    ! Pre-compute source wavefield


if (1 == 1) then    
       
       ixs = ixsloc + bndx

       ! initializing the array
       do ix=1,nx
          do iz=1,nz
             up2(iz,ix)=0.; um2(iz,ix)=0.; uo2(iz,ix)=0.;
          end do          
       end do
       

       do it=1,nt+ndelay

       ! if ((it .gt. 950)) write(0,*) 'A: it', it,  sum(uo2) 


            if (nop .eq. 2) then
               !! 4th order laplacian 
                  call laplac4(.true., um2, uo2,up2, vp2, dt, nx, nz, idx, idz, nop,it)
            else if (nop .eq. 3) then        
	        !! 6th order laplacian        
              	  call laplac6(.true., um2, uo2,up2, vp2, dt, nx, nz, idx, idz, nop,it)
             else if (nop .eq. 4) then        
	        !! 8th order laplacian        
              	call laplac8(.true., um2, uo2,up2, vp2, dt, nx, nz, idx, idz, nop,it) 
            else if (nop .eq. 8) then
              call laplac2d16(.true., um2, uo2, up2, vp2, nx, nz, xbsz, zbsz, idx, idz, nop, it, nodes)
	    end if


	  ! Take second derivative (in time) of source wavefield
	  if (it .gt. 1) then
		  do ix=1,nx
		     do iz=1,nz
       		        us2Der(iz,ix) = (up2(iz,ix) - 2*uo2(iz,ix) + um2(iz,ix))
		     end do		   
	  	  end do
	  else
	  	  do ix=1,nx
		     do iz=1,nz
       		        us2Der(iz,ix) = 0.
		     end do
	  	  end do   
	  end if


       ! if ((it .gt. 950)) write(0,*) '@B: it', it,  sum(uo2) 

           !! time step
           do ix=1,nx
              do iz=1,nz
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);
              end do    
           end do

          
          ! if ((it .gt. 950)) write(0,*) '@@C: it', it,  sum(uo2)    

           ! Injecting source/rec wavelet into the finite diff grid
           izs = bndz+1

	   if (it .lt. nt) then 
       		do ix=1,nnx 
      		  uo2(izs,ix+bndx) = uo2(izs,ix+bndx) + -1*amp*ArealSou(ix,it)
       	      	end do
           end if


           ! if ((it .gt. 950)) write(0,*) '@@@D: it', it,  sum(uo2)

              !! taper
  	          call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper, uo2)
	          call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper, up2)
	          call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper, um2)


       ! if ((it .gt. 950)) write(0,*) '@@@@E: it', it,  sum(uo2) 

             !! Absorbing boundary
                call absorbingBC1(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo2,um2)             


       ! if ((it .gt. 950)) write(0,*) '@@@@F: it', it,  sum(uo2) 


              ! Store source wavefield to memory for every xsnap time step
              ! Calculate source amplitude for amp balancing
	      do ix=1,nnx
                 do iz=1,nnz-nw
			amp2(iz,ix) = amp2(iz,ix) + us2Der(iz+bndz+nw,ix+bndx)**2

                    if ((it .gt. ndelay).and.((mod(it-ndelay,xsnap)==0) .or. (xsnap .eq. 1)) ) then
                        istore = (it-ndelay)/xsnap
!  		            us2(iz,ix,istore) = uo2(iz+bndz+nw,ix+bndx)		
    		            us2(iz,ix,istore) = us2Der(iz+bndz+nw,ix+bndx)		        
                    end if
                 end do
              end do





 
           if (verb .and. (mod(it,50) .eq. 1)) then
              write(0,*) 'Sou: it, sum', it-ndelay, sum(uo2)
           end if
        
 
        	

!             if(  (it .gt. ndelay) .and.  mod(it-ndelay,jsnap)==0 .and. wfldout) then
             if( mod(it,jsnap)==1 .and. wfldout ) then
                   write(0,*) 'writing to wfldsou1', it, ndelay
		   call sep_write(us2Der,wfldsou2_)
!	           if (it .gt. 500) call erexit('exit now (it-ndelay)/jsnap')
             end if
           


       end do !(end source time loop)


end if

 

! ------------------------------------------------------------
! ------ Output file -----------------------------------------

!maxus2 = 0.
!do isave=1,nsave
!   do ix=1,nnx
!      do iz=1,nnz-nw
!      	 val = abs(us2(iz,ix,isave))
!      	 if (val .gt. maxus2) val = maxus2
!      end do
!   end do
!end do
 
!do isave=1,nsave
!   do ix=1,nnx
!      do iz=1,nnz-nw
!      	 val = abs(us2(iz,ix,isave))
!	 diff = (maxus2 - val)/2
!      end do
!  end do
!end do



           call sep_write(us2 , trim(out_), dim1=nnz-nw, dim2=nnx, dim3=nsave,esize=4)



write(0,*) 'after ASout'
!            do ix=1,nnx
!              do iz=1,nnz-nw
!                 amp2(iz,ix) = sqrt(amp2(iz,ix))
!              end do
!            end do
 
            call sep_write(amp2, trim(ampfile_), dim1=nnz-nw, dim2=nnx,esize=4)

write(0,*) 'nz, nnz, nw', nz, nnz, nw

   call auxputch("n1",'i',nnz-nw, trim(out_)); call auxputch("d1",'f', dz, trim(out_));  call auxputch("o1",'f', vel_grid%zw, trim(out_));
   call auxputch("n2",'i', nnx, trim(out_)); call auxputch("d2",'f', dx, trim(out_));  call auxputch("o2",'f', ox, trim(out_));
   call auxputch("n3",'i', nsave, trim(out_)); call auxputch("d3",'f', dt*xsnap, trim(out_));   
   call auxputch("o3",'f', ((ndelay+xsnap-1)*dt-tdelay), trim(out_));
   call auxputch("n4",'i', 1, trim(out_)); call auxputch("d4",'f', 400., trim(out_));  call auxputch("o4",'f', Gx, trim(out_));       
   call auxputch("label1", 's', 'nz', trim(out_));
   call auxputch("label2", 's', 'nx', trim(out_));
   call auxputch("label3", 's', 'nt', trim(out_));
   call auxputch("label4", 's', 'shot', trim(out_));


   call auxputch("n1",'i',nnz-nw, trim(ampfile_)); call auxputch("d1",'f', dz, trim(ampfile_));  call auxputch("o1",'f', oz, trim(ampfile_));
   call auxputch("n2",'i', nnx, trim(ampfile_)); call auxputch("d2",'f', dx, trim(ampfile_));  call auxputch("o2",'f', ox, trim(ampfile_));

   call auxputch("label1", 's', 'nz', trim(ampfile_));
   call auxputch("label2", 's', 'nx', trim(ampfile_));




 if ((wfldout) .and. (1 == 1)) then

  	write(0,*) 'writing to wfldsou2', wfldsou2_
 	   call auxputch("n1",'i',nz,trim(wfldsou2_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou2_));   
           call auxputch("n2",'i',nx,trim(wfldsou2_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou2_));  
	   call auxputch("o2",'f',vel_grid%x%o-(bndx)*dx, trim(wfldsou2_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou2_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou2_));
           call auxputch("n3",'i',nt/50,trim(wfldsou2_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou2_));  
           call auxputch("o3",'f', -tdelay+(jsnap-1)*dt, trim(wfldsou2_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou2_));
 end if
 
 
 if (ASout) then

 	   call auxputch("n1",'i',nnx,trim(ASfile_));  
   	   call auxputch("d1",'f',dx,trim(ASfile_));   
           call auxputch("n2",'i',nt,trim(ASfile_));  
   	   call auxputch("d2",'f',dt,trim(ASfile_));  
   	   call auxputch("label1", 's', 'X', trim(ASfile_));
     	   call auxputch("label2", 's', 't', trim(ASfile_));

  	   
 end if


write(0,*) 'pass 0'

   ! --------------------------------------------------------------



   deallocate(vp2)
   write(0,*) 'dealloc 1'
   deallocate( bndz0, bndz1, bndx0, bndx1)
   write(0,*) 'dealloc 2'
!   deallocate( ztaper, xtaper, ctaper)
   write(0,*) 'dealloc 3'
   write(0,*) 'dealloc 4'
   deallocate( us2)
   write(0,*) 'dealloc 5'
   deallocate(  up2, um2, uo2)
   write(0,*) 'dealloc 6'
   deallocate( ArealSou)
   
   call sep_close()

   ! Stop timer
   call itime(now2)
   write(0,*) 'compute time Forward3D (s)', 3600*(now2(1)-now1(1))+60.*(now2(2)-now1(2))+(now2(3)-now1(3))



write(0,*) 'program should exit'

end program
