!! Input a parameter file make a 5-reflectors velocity model
!! 5 reflectorss model. The first and last layer is horizontal
!!                      Second layer is a slope
!!                      3rd and 4th layer is a bump
!! Parameter Explanations:
!! h1 = depth for first refl; h2= depth for 2nd refl (center of the slope)
!! h3,h4,h5 = depth for 3rd, 4th, and 5th reflector
!! v1= vel above 1st refl, v2-v5: vel above 2nd-5th refl; v6: vel below 5th refl


!! t1=depth value of first layer;     t2=depth value of 2nd layer
!! a = (km/m) height of Exp bump;  b= width of Exp bump (x=c +/- b when z=amp*0.37);  c=center of Exp bump
!! scat: (1=place scatter, 0 = no scatter)
!! xscat, zscat:  x and z coordinate of the scatter
!! velscat: velocity of the scatterer
!! n1, d1, o1, label1 - for z direction;      n2, d2, o2, label2 - for x direction

program Vel2

  use sep
  use smooth_mod
  implicit none

  logical :: verb, smooth
  integer :: nz, nx, iz, ix
  real :: h1, h2, h3, h4, h5
  real :: oz, ox, dz, dx
  real :: v1, v2, v3, v4, v5, v6

  real :: z, x
  real :: deg, slope2       !! Degree value for slope
  real :: a3, b3, c3, zdome3
  real :: a4, b4, c4, zdome4
  real :: pi, degrad, xmid

  character(len=128)  :: label1
  character(len=128)  :: label2

!! velocity values
  real, dimension(:,:), pointer :: vel



!! Open I/O
  call sep_init()

  call from_param("verb",verb,.false.);
  call from_param("smooth",smooth,.false.);


!! Read file dimensions
  call from_param("n1",nz,300);    call from_param("o1",oz,0.);   call from_param("d1",dz,10.);
  call from_param("n2",nx,1000);   call from_param("o2",ox,0.);   call from_param("d2",dx,10.);
  call from_param("label1",label1,'z');    call from_param("label2",label2,'x');

  if (verb .eq. 1) write(0,*) 'nz, oz, dz',  nz, oz, dz
  if (verb .eq. 1) write(0,*) 'nx, ox, dx',  nx, ox, dx


!! Read velocity  and depth
  call from_param("v1",v1,1500.);    call from_param("v2",v2,2000.);   call from_param("v3",v3,3000.);
  call from_param("v4",v4,3500.);    call from_param("v5",v5,4000.);   call from_param("v6",v6,5000.);

  call from_param("h1",h1,500.);    call from_param("h2",h2,800.);   call from_param("h3",h3,1500.);
  call from_param("h4",h4,2000.);    call from_param("h5",h5,2500.);   


  if (verb .eq. 1) write(0,*) 'v1, v2, v3, v4, v5, v6', v1, v2, v3,v4, v5, v6
  if (verb .eq. 1) write(0,*) 'h1, h2, h3, h4, h5', h1, h2, h3, h4, h5


!! Read the parameter for the Exponential bump
  call from_param("deg", deg, 5.);

  call from_param("a3",a3);      call from_param("b3",b3,(nx/8)*dx);    call from_param("c3",c3,(nx/4)*dx+ox);
  call from_param("a4",a4, 150.);      call from_param("b4",b4,(nx/8)*dx);    call from_param("c4",c4,(nx/4)*dx+ox);

  if (verb .eq. 1) write(0,*) 'a3, b3, c3', a3, b3, c3
  if (verb .eq. 1) write(0,*) 'a4, b4, c4', a4, b4, c4
  if (verb .eq. 1) write(0,*) 'deg', deg


!! Error Check  zmax> h5>h4>h3>h2>h1 > zmin
   if ((h1 .ge. h2) .or. (h2 .ge. h3) .or. (h3 .ge. h4) .or. (h4 .ge. h5)) call erexit('h5>h4>h3>h2>h1 unsatisfied')
   if ((h1 .lt. oz) .or. (h5 .gt. (oz+(nz-1)*dz))) call erexit('h1<oz or h5>zmax')
!! Error Check that degree falls within +/- 90 degree
   if ((deg .ge. 90.) .or. (deg .le. -90.)) call erexit('Deg must be within +/- 90');
   


!! Output files
  call to_history("n1",nz);     call to_history("o1",oz);    call to_history("d1",dz);
  call to_history("n2",nx);     call to_history("o2",ox);    call to_history("d2",dx);
  call to_history("label1",label1);   call to_history("label2",label2);


  call sep_close()



  allocate(vel(nz,nx))

  pi = 3.14159265;
  vel=0.
  xmid = ox + (nx/2-1)*dx
  degrad = deg*pi/180.;

write(0,*) '*** c3, c4 ***', c3, c4

   !! Main loop (with an Exponential bump)
   do iz=1,nz     
      z=oz+dz*(iz-1)
      do ix=1,nx    
         x=ox+dx*(ix-1)

         slope2 = h2 - tan(degrad)*(x-xmid);
         zdome3 = -a3*exp(-1*((x-c3)/b3)**2)+h3;
         zdome4 = -a4*exp(-1*((x-c4)/b4)**2)+h4;


         if (z .ge. h5) then
            vel(iz,ix)=v6;
         else if (z .ge. zdome4) then
            vel(iz,ix)=v5;
         else if (z .ge. zdome3) then
            vel(iz,ix)=v4;
         else if (z .ge. slope2) then
            vel(iz,ix)=v3;
         else if (z .ge. h1) then 
            vel(iz,ix)=v2;
         else
            vel(iz,ix)=v1;
         end if

      end do
   end do

! if (smooth) then
!      call triangle2(5, 5, nz, nx, vel, vel)
! end if



   


!! Write new vel values to output
  call sep_write(vel(1:nz,1:nx))


  call exit(0)
   
end program 
