program VelSmooth
! Read in a velocity model, smooth the slowness with specified width
! Output the smoothed velocity model.
! width  : smoothing width


  use sep
  use smooth_mod
  use type_mod

implicit none


 type(velo) :: vel

 logical    :: smooth
 integer    :: ix, iz,  nx, nz, width
 real, dimension(:,:), allocatable :: vvel, slow

 call sep_init()

   
   call from_param('width', width, 0)


   if (width .gt. 0) then 
      smooth = .true.
   else 
      smooth = .false.   
   end if

   if (width .gt. 10) call erexit('width must be < 10')

  ! Read original velocity model param
   call from_history("n1", vel%z%n);  call from_history("d1", vel%z%d);  call from_history("o1",vel%z%o);
   call from_history("n2", vel%x%n);  call from_history("d2", vel%x%d);  call from_history("o2",vel%x%o);


   nz = vel%z%n;
   nx = vel%x%n;

   write(0,*) '****************'
   write(0,*) 'smooth, width', smooth, width
   write(0,*) 'dx, dz', vel%x%d, vel%z%d
   write(0,*) '****************'
   
   allocate(vvel(nz,nx), slow(nz,nx))


   ! Read in the background velocity
    call sep_read( vvel(1:nz,1:nx))

    do ix=1,nx
       do iz=1,nz
         if (vvel(iz,ix) .gt. 0) then
            slow(iz,ix) = 1/vvel(iz,ix)
         else
           call erexit('Error: there are zeros in the velocity file')
         end if
       end do
    end do
           
     ! Smoothing the slowness
     if (smooth) then
      write(0,*) 'smoothing the velocity'
      call triangle2(width, width, nz, nx, slow, slow)
      call triangle2(width, width, nz, nx, slow, slow)
      call triangle2(width, width, nz, nx, slow, slow)
      call triangle2(width, width, nz, nx, slow, slow)
      call triangle2(width, width, nz, nx, slow, slow)
      call triangle2(width, width, nz, nx, slow, slow)
     end if


    do ix=1,nx
       do iz=1,nz
         if (slow(iz,ix) .gt. 0) then
            vvel(iz,ix) = 1/slow(iz,ix)
         else
           call erexit('Error: there are zeros in the slowness file')
         end if
       end do
    end do
   
!----------------------------------------------
! Output

  call sep_write(vvel,dim1=nz, dim2=nx, esize=4)
  write(0,*) "vel%z%n", vel%z%n
  call to_history("n1",vel%z%n); call to_history("d1",vel%z%d);  call to_history("o1",vel%z%o);
  call to_history("n2",vel%x%n); call to_history("d2",vel%x%d);  call to_history("o2",vel%x%o);
  


end program
