program SouGen2dDnMul
! Pre-compute source wavefield with specification defined in a parameter file
! Source wavefield is saved in 1 file 
! Reflecting top boundary
! And sharp velocity jump
! amount of water to remove for upgoing wavefields
!
! zw    : depth of the waterbody
! zsb   : depth at the bottom of the seabed
! 
!


  use sep
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod


  implicit none
  
  
  type(sou_grid)       :: sougrid
  type(bc_values)     :: bc

  character(len=128)   :: wfldsou1_, wfldsou2_, wfldsou3_, sou_, dat1_, dat2_, dat3_, ampfile_
    
    
  logical                              :: verb, ASout, wfldout
  integer                              :: nx, nz1, nz2, ii, ix, iz, nt, nst, nw,nsb, ntas,  nnx, nnz
  integer                              ::  izs1, izs2, izsloc, ixs, ixsloc, icheck, it, ilag
  integer                              :: nop, bndx, bndz, jsnap, xsnap, nsave
  integer                              :: xbsz, zbsz, irec, ndelay, istore
  real    			       :: amp, Gx, Gz, watvel, recZ
  real                                 :: dx, ox, dz, oz, dt, dst, ot
  real                                 :: idx, idz, zw, zsb, tdelay
  real                                 :: xtapco, ztapco, rho
  real*8                               :: time1, time2
  real, dimension(:), allocatable      :: bndz0, bndz1, bndx0, bndx1,  fs
  real, dimension(:), allocatable      :: xtaper, ztaper
  real, dimension(:,:), allocatable    :: Vo, Vo2, ctaper,   amp1,  ArealSou1, ArealSou2, ArealSou3
  real, dimension(:,:), allocatable    :: uo1, um1, up1, uo2, um2, up2, uuo, amp2
  real, dimension(:,:,:), allocatable  :: us2
  character(len=128)                   :: out_

   !! For OMP
 ! integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer            :: inode1, inode, nodes, Totalnodes

    call sep_init()

!!======================================================
  !! Parameter reading

   call from_param("verb",verb,.false.); call from_param("ASout", ASout,.false.);
   call from_param("wfldout", wfldout, .false.);   

   ! Read source and out file
   call from_param("soufile", sou_);   
   call from_param("outtag", out_);
   call from_param("ampfile", ampfile_);

   ! Read back ground velocity param
   call from_history("n1",nnz);  call from_history("d1",dz);  call from_history("o1",oz);
   call from_history("n2",nnx);  call from_history("d2",dx);  call from_history("o2",ox);
  
   ! Read position of the source from command line
   call from_param('Gx',Gx, 0.5*((nnx-1)*dx+ox));     call from_param('Gz',Gz, ((nnz-2)*dz+oz));
   izsloc =  (Gz-oz)/dz + .5
   ixsloc =  (Gx-ox)/dx + .5

   call from_param("recZ",recZ,oz)
   irec = int(max((recZ-oz)/dz + .5,1.))

   ! Read water depth, water vel and density
   call from_param('rho',rho);
   call from_param("zw",zw,500.);          ! depth of water
   nw = (zw-oz)/dz + 0.5   
   call from_param("zsb",zsb);
   
   call from_aux(sou_, "n1", nst);  call from_aux(sou_,"d1", dst); 
    

   
   ! Read time information
   call from_param("nt",nt,2500);  call from_param("dt",dt,0.001);   call from_param("ot",ot,0.)
   call from_param("tdelay", tdelay, 0.);
   call from_param("ilag", ilag,0);

     
   !! taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)
   call from_param("nop", bc%nop,8);
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;


   !! Auxillary output files
   call from_param("wfldsou1", wfldsou1_, "wfldtest1.H");
   call from_param("wfldsou2", wfldsou2_, "wfldtest2.H");
   call from_param("wfldsou3", wfldsou3_, "wfldtest3.H");
   call from_param("dat1file", dat1_, "dat1.H");   call from_param("dat2file", dat2_, "dat2.H")
   call from_param("dat3file", dat3_, "dat3.H");
   call from_param("jsnap", jsnap, 100);   call from_param("xsnap",xsnap,8)

 

    nop = bc%nop
    bndx = bc%bndx
    bndz = bc%bndz  
    xtapco = bc%xtapco
    ztapco = bc%ztapco
    nx = nnx + 2*bndx
        
    nw = (zw-oz)/dz + .5
    nsb = (zsb-zw-oz)/dz + 0.5
    ndelay = (tdelay-ot)/dt+.5
   
    ntas = nw


    ! Parallel Programming
    Totalnodes = omp_get_num_procs();
    nodes = min(Totalnodes-1,4);
!    nodes = Totalnodes-1;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes

     
    

 !*********************************
   write(0,*) '**************************'
   write(0,*) 'nnz, dz, oz ', nnz, dz, oz
   write(0,*) 'nnx, dx, ox ', nnx, dx, ox
   write(0,*) 'Gx, Gz',           Gx, Gz
   write(0,*) 'zw, zsb, nw, nsb', zw, zsb, nw, nsb
   write(0,*) 'ixsloc, izsloc,irec',      ixsloc, izsloc, irec
   write(0,*) 'nt, ot, dt',       nt, ot, dt
   write(0,*) 'bc%nop, verb', bc%Op, bc%nop, verb
   write(0,*) 'bndz, bndx, nop',  bc%bndz, bc%bndx, bc%nop
   write(0,*) 'xtapco, ztapco',   bc%xtapco, bc%ztapco
   write(0,*) '# of nodes used',  nodes
   write(0,*) 'tdelay, ndelay',   tdelay, ndelay
   write(0,*) '**************************'


   !! Error checking
   if (dt .ne. dst) call erexit("dt should equal to dst")
   if ((nop .ne. 8) .and. (nop .ne. 3) .and. (nop .ne.4) ) call erexit('nop must be 3, 4, or 8')

 !******************************************************************
 ! Part Zero: Read in source wavelet



   allocate(fs(nt))
   fs = 0.;

   ! Read in the source wavelet
   if (nst .lt. nt) then
       call sep_read( fs(1:nst), sou_)
   else 
       call sep_read( fs(1:nt), sou_)
   end if


   do it=1,nt
	fs(it) = fs(it)*1000.
   end do


  ! call sep_write(fs, "fs.H");
  !	   call auxputch("n1",'i',nt,'fs.H');  
  !	   call auxputch("d1",'f',0.0015,'fs.H');   
  !       call auxputch("o1",'f',0,'fs.H');
  !  call erexit('Exit here for debug')




  !******************************************************************

    nz1 = nnz + 2*bndz
    nz2 = nnz + nop + bndz 
    nx = nnx + 2*bndx
    nsave = nt/xsnap


    write(0,*) 'nz1, nz2, nnz, nx, nnx', nz1, nz2, nnz, nx, nnx
    write(0,*) 'bndz, bndx', bndz, bndx

   ! Memory Allocation
   allocate(ztaper(bndz))
   allocate(xtaper(bndx))
   allocate(ctaper(bndz,bndx))
   if (ASout) then
      allocate(ArealSou1(nnx,nt))
      allocate(ArealSou2(nnx,nt))
      allocate(ArealSou3(nnx,nt))
   end if

   allocate(up1(nz1,nx), um1(nz1,nx), uo1(nz1,nx))
   allocate(up2(nz2,nx), um2(nz2,nx), uo2(nz2,nx))   
   allocate(uuo(nz2,nx))
   allocate(Vo(nz1,nx), Vo2(nz2,nx))
   allocate(us2(nnz-nw,nnx,nsave))
   allocate(amp2(nnz-nw,nnx))
 
!  watvel = 2000.

   ! NO seabed for Vo  
!   do ix=1,nx
!      Vo(1:(bndz+nw),ix)=1500.**2
!      Vo((bndz+nw+1):,ix)=2000.**2
!   end do   
!   do ix=1,nx
!      Vo2(1:(nop+nw),ix)=1500.**2
!      Vo2((nop+nw+1):,ix)=2000.**2
!   end do


   ! read the background velocity
   call sep_read(Vo2(1+nop:nz2-bndz,1+bndx:nx-bndx))
 
   ! padding
   do iz=1,nop; Vo2(iz,:) = Vo2(1+nop,:); end do
   do iz=1,bndz; Vo2(nz2-iz+1,:) = Vo2(nz2-bndz,:); end do
   do ix=1,bndx; Vo2(:,ix)=Vo2(:,1+bndx); Vo2(:,nx-ix+1)=Vo2(:,nx-bndx); end do

   do ix=1,nx
      do iz=1,nz2
         Vo2(iz,ix) = Vo2(iz,ix)*Vo2(iz,ix)
      end do
   end do

  ! Now we defined Vo
    do iz=1,nnz+nsb
      Vo(bndz+iz,:) = Vo2(nop+iz,:)
    end do


    do iz=1,bndz; Vo(nz1-iz+1,:) = Vo(nz1-bndz,:); Vo(iz,:) = Vo(1+bndz,:); end do;
    do ix=1,bndx; Vo(:,ix)=Vo(:,1+bndx); Vo(:,nx-ix+1)=Vo(:,nx-bndx); end do;
   
    !! Absorbing BC and Tapersetup  


    call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)
    
     idx = (dt/dx)**2;   idz = (dt/dz)**2
       ixs = bndx + ixsloc
       izs1 = bndz + izsloc
       izs2 = nop + izsloc

       ! initializing the array
       do ix=1,nx
             up1(:,ix)=0.; um1(:,ix)=0.; uo1(:,ix)=0.;
             up2(:,ix)=0.; um2(:,ix)=0.; uo2(:,ix)=0.;
             uuo(:,ix)=0.;
       end do

    xbsz = 60
    zbsz = 60

    write(0,*) 'sum Vo, idx, idz', sum(Vo), idx, idz
    write(0,*) 'dx,dz,dt', dx, dz, dt

       do it=1,nt

           if (it .eq. 101) write(0,*) 'Pass 1 it,  sum', it, sum(uo1), sum(um1), sum(up1)                                
              call laplac2d16(.true., um1, uo1, up1, Vo, nx, nz1, xbsz, zbsz, idx, idz, nop, it, nodes)
              call laplac2d16(.true., um2, uo2, up2, Vo2, nx, nz2, xbsz, zbsz, idx, idz, nop, it, nodes)              

           if (it .eq. 101) write(0,*) '@Pass 2 it,  sum', it, sum(uo1), sum(um1), sum(up1)

           !! time step
           do ix=1,nx
                 um1(:,ix)=uo1(:,ix); 
                 uo1(:,ix)=up1(:,ix);
                 um2(:,ix)=uo2(:,ix); 
                 uo2(:,ix)=up2(:,ix);                 
           end do
             
           ! Injecting 
            uo1(izs1,ixs) =  uo1(izs1,ixs) + fs(it)  
            uo2(izs2,ixs) =  uo2(izs2,ixs) + fs(it)  

            
            
!!            if (it .gt. 300) write(0,*) '@@Pass 3 it,  sum', it, sum(uo1), sum(ArealSou1)

              !! taper		
	       call taperingCos(nz1,nx,bndz,bndx,ztaper,xtaper,ctaper, uo1)
	       call taperingCos(nz1,nx,bndz,bndx,ztaper,xtaper,ctaper, up1)
	       call taperingCos(nz1,nx,bndz,bndx,ztaper,xtaper,ctaper, um1)
	       call taperingCosV2(nz2,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, uo2)
	       call taperingCosV2(nz2,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, up2)
	       call taperingCosV2(nz2,nx,nop,bndz,bndx,ztaper,xtaper,ctaper, um2)
	       
!!            if (it .gt. 300) write(0,*) '@@@Pass 4 it,  sum', it, sum(uo1), sum(ArealSou1)
         do ix=1,nx
            uuo(nop+1:,ix) = uo2(nop+1:,ix)-uo1(bndz+1:,ix)
         end do



        ! Store source wavefield to memory for every xsnap time step
        ! Calculate source amplitude for amp balancing
	do ix=1,nnx; do iz=1,nnz-nw
           amp2(iz,ix) = amp2(iz,ix) + uuo(iz+nop+nw,ix+bndx)**2
        end do; end do


       if ((it .gt. ndelay).and.((mod(it-ndelay-1,xsnap)==0) .or. (xsnap .eq. 1)) ) then
          istore = (it-ndelay)/xsnap+1
         ! write(0,*) 'outputting it istore', it, istore              
          if (istore .lt. size(us2,3)) then
             do ix=1,nnx
                do iz=1,nnz-nw
       	           us2(iz,ix,istore) = uuo(iz+nop+nw,ix+bndx)
                end do
             end do             
          end if
       end if


       ! Store areal source to memory for every  time step
       if (ASout .and. (it .lt. nt-1)) then
          do ix=1,nnx
             ArealSou1(ix,it) = uo1(nop+irec,bndx+ix)
             ArealSou2(ix,it) = uo2(nop+irec,bndx+ix)
             ArealSou3(ix,it) = uuo(nop+irec,bndx+ix) 
	  end do
       end if     

        
       if (mod(it, 100) .eq. 1) write(0,*) 'it', it, sum(uo1), sum(uo2)

           if(wfldout .and. (mod(it,jsnap)==0 )) then
              call sep_write(uo1,wfldsou1_)
              call sep_write(uo2,wfldsou2_)
              call sep_write(uuo,wfldsou3_)
            end if
        
       end do !(end source time loop) 
       
! ------------------------------------------------------------
! ------ Output file ----------------------------------------- 
! ------------------------------------------------------------
    
       write(0,*) 'before us2', out_
       call sep_write(us2 , trim(out_), dim1=nnz-nw, dim2=nnx, dim3=nsave,esize=4)
       write(0,*) 'after us2'

       write(0,*) 'writing to out_',out_
       call auxputch("n1",'i',nnz-nw, trim(out_)); call auxputch("d1",'f', dz, trim(out_));  
       call auxputch("o1",'f', zw, trim(out_));    call auxputch("label1", 's', 'nz', trim(out_));
       call auxputch("n2",'i', nnx, trim(out_));   call auxputch("d2",'f', dx, trim(out_));  
       call auxputch("o2",'f', ox, trim(out_));    call auxputch("label2", 's', 'nx', trim(out_));
       call auxputch("n3",'i', nsave, trim(out_)); call auxputch("d3",'f', dt*xsnap, trim(out_));   
       call auxputch("o3",'f', ot, trim(out_)); call auxputch("label3", 's', 'nt', trim(out_));
       call auxputch("n4",'i', 1, trim(out_));     call auxputch("d4",'f', 1., trim(out_));  
       call auxputch("o4",'f', 0., trim(out_));    call auxputch("label4", 's', 'UD', trim(out_));   
       call auxputch("n5",'i', 1, trim(out_));     call auxputch("d5",'f', 400., trim(out_));  
       call auxputch("o5",'f', Gx, trim(out_));    call auxputch("label5", 's', 'shot', trim(out_));
      
      write(0,*) 'outputting ampfile'
      call sep_write(amp2, trim(ampfile_), dim1=nnz-nw, dim2=nnx,esize=4)
      
      call auxputch("n1",'i',nnz-nw, trim(ampfile_)); call auxputch("d1",'f', dz, trim(ampfile_));  
      call auxputch("o1",'f', oz, trim(ampfile_));    call auxputch("label1", 's', 'nz', trim(ampfile_));
      call auxputch("n2",'i', nnx, trim(ampfile_));   call auxputch("d2",'f', dx, trim(ampfile_));  
      call auxputch("o2",'f', ox, trim(ampfile_));    call auxputch("label2", 's', 'nx', trim(ampfile_));
      call auxputch("n3",'i', 1, trim(ampfile_));     call auxputch("d3",'f', 1., trim(ampfile_));  
      call auxputch("o3",'f', 0., trim(ampfile_));    call auxputch("label3", 's', 'UD', trim(ampfile_));



     if (wfldout) then
	write(0,*) 'writing to wfldsou1', wfldsou1_
 	   call auxputch("n1",'i',nz1,trim(wfldsou1_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou1_));   
           call auxputch("n2",'i',nx,trim(wfldsou1_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou1_));  
	   call auxputch("o2",'f',ox-(bndx)*dx, trim(wfldsou1_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou1_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou1_));
     	   
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou1_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou1_));  
           call auxputch("o3",'f', ot, trim(wfldsou1_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou1_));

  	write(0,*) 'writing to wfldsou2', wfldsou2_
 	   call auxputch("n1",'i',nz2,trim(wfldsou2_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou2_));   
           call auxputch("n2",'i',nx,trim(wfldsou2_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou2_));  
	   call auxputch("o2",'f',ox-(bndx)*dx, trim(wfldsou2_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou2_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou2_));
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou2_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou2_));  
           call auxputch("o3",'f', ot, trim(wfldsou2_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou2_));

  	write(0,*) 'writing to wfldsou3', wfldsou3_
 	   call auxputch("n1",'i',nz2,trim(wfldsou3_));  
   	   call auxputch("d1",'f',dz,trim(wfldsou3_));   
           call auxputch("n2",'i',nx,trim(wfldsou3_));  
   	   call auxputch("d2",'f',dx,trim(wfldsou3_));  
	   call auxputch("o2",'f',ox-(bndx)*dx, trim(wfldsou3_));
   	   call auxputch("label1", 's', 'Z', trim(wfldsou3_));
     	   call auxputch("label2", 's', 'X', trim(wfldsou3_));
           call auxputch("n3",'i',nt/jsnap,trim(wfldsou3_));   
           call auxputch("d3",'f', dt*jsnap, trim(wfldsou3_));  
           call auxputch("o3",'f', ot, trim(wfldsou3_));              
  	   call auxputch("label3", 's', 'time', trim(wfldsou3_));
    end if

      
        if (ASout) then
      
	  write(0,*) ' Writing AS file' 
!	  call sep_write(ArealSou1 , trim(dat1_), dim1=nnx, dim2=nt, esize=4)
       	  call sep_write(ArealSou2, trim(dat2_), dim1=nnx, dim2=nt, esize=4)
          call sep_write(ArealSou3, trim(dat3_), dim1=nnx, dim2=nt, esize=4)
        
        write(0,*) "writing out dat1"
 	   call auxputch("n1",'i',nnx,trim(dat1_));  
   	   call auxputch("d1",'f',dx,trim(dat1_));   
           call auxputch("n2",'i',nt,trim(dat1_));  
   	   call auxputch("d2",'f',dt,trim(dat1_));  
	   call auxputch("o2",'f',ot, trim(dat1_));
   	   call auxputch("label1", 's', 'x', trim(dat1_));
     	   call auxputch("label2", 's', 't', trim(dat1_));
        end if

        write(0,*) "writing out dat2"
 	   call auxputch("n1",'i',nnx,trim(dat2_));  
   	   call auxputch("d1",'f',dx,trim(dat2_));   
           call auxputch("n2",'i',nt,trim(dat2_));  
   	   call auxputch("d2",'f',dt,trim(dat2_));  
	   call auxputch("o2",'f',ot, trim(dat2_));
   	   call auxputch("label1", 's', 'x', trim(dat2_));
     	   call auxputch("label2", 's', 't', trim(dat2_));
     	   
        write(0,*) "writing out dat3"
 	   call auxputch("n1",'i',nnx,trim(dat3_));  
   	   call auxputch("d1",'f',dx,trim(dat3_));   
           call auxputch("n2",'i',nt,trim(dat3_));  
   	   call auxputch("d2",'f',dt,trim(dat3_));  
	   call auxputch("o2",'f',ot, trim(dat3_));
   	   call auxputch("label1", 's', 'x', trim(dat3_));
     	   call auxputch("label2", 's', 't', trim(dat3_));

write(0,*) 'pass 0'

   ! --------------------------------------------------------------

   deallocate( fs)
   write(0,*) 'dealloc 1'
   write(0,*) 'dealloc 2'
   deallocate( ztaper, xtaper, ctaper)
   write(0,*) 'dealloc 3'
   deallocate( up1, um1, uo1, up2, um2, uo2)
   write(0,*) 'dealloc 6'
   deallocate( ArealSou1, ArealSou2)
   call sep_close()

! Start timer
    call cpu_time(time2)
    write(0,*) 'CPU runtime is:', time2-time1, 'sec'


write(0,*) 'program should exit'

end program
