program redatumCMP
  ! redatum CRG for OBS for simple velocity scanning analysis
  ! Each shot (at sea surface) are redatum thru the water column according to the depth of OBS node
  ! Redatuming is achieved by subtracting the travel time thru the water column 
  ! Input: Data volume by (t, xs)
  ! Input: water vel, Gx (x position of node), Gz (z position)
  ! Output: Data volume (t, mx, hx)
  

  use sep
  
  implicit none

  
  
  logical            :: verb 
  integer            :: ixs, ixa, imx, ihx, itd, it
  integer            :: nxs, nxa, nmx, nhx, ntd, nt
  real               :: dxs, dxa, dmx, dhx, dtd, dt
  real               :: oxs, oxa, omx, ohx, otd, ot  
  real               :: xs, xa, mx, hx, td, t
  real               :: vw, Gz, Gx
  character(len=128) :: out_
  real, dimension(:,:,:), allocatable    :: Inarray 
  real, dimension(:,:,:,:), allocatable  :: Outarray
  !! For OMP
  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer, external  :: srite_window, sreed_window, getch
  integer            :: inode1, inode, nodes
  
   call sep_init()


  
  
  !!======================================================
  !! Parameter reading
  
  call from_param("verb",verb,.false.);
  
  ! water vel, and node positions
  call from_param("vw",vw,1500.);
  call from_param("Gx", Gx); call from_param("Gz",Gz);

  ! Input file parameter
  call from_history("n1", nxs); call from_history("d1",dxs); call from_history("o1",oxs);
  call from_history("n2", nt); call from_history("d2",dt); call from_history("o2",ot);
  
  call from_param("nmx",nmx, nxs);  call from_param("dmx",dmx, dxs);   call from_param("omx",omx, oxs);
    call from_param("nodes",nodes, 4);
   
  nxa = nmx
  dxa = dmx
  oxa = omx  
  nhx = nmx
  dhx = dmx
  ohx = 0.0
  
  dtd = dt
  otd = ot
 
   ! Parallel Programming

  if (nodes .gt. omp_get_num_procs()) nodes =omp_get_num_procs()-3;
    call omp_set_num_threads(nodes);
    inode1=0
    
    !*********************************
   write(0,*) '**************************'
   write(0,*) 'verb', verb
   write(0,*) 'vw, Gx, Gz', vw, Gx, Gz
   write(0,*) 'nxs, dxs, oxs', nxs, dxs, oxs
   write(0,*) 'nt, dt, ot', nt, dt, ot
   write(0,*) 'ntd, dtd, otd', ntd, dtd, otd   
   write(0,*) 'nmx, dmx, omx', nmx, dmx, omx
   write(0,*) 'nhx, dhx, ohx', nhx, dhx, ohx
   write(0,*) '**************************'
 
! call erexit('exit here')
 
  ! allocation 
  allocate(Inarray(nxs,nt,2), Outarray(nt,nmx,nhx,1))
  Inarray = 0.
  Outarray = 0.
  
  !$OMP parallel do ordered private(inode, ixs, xs, ixa, xa, td, itd, mx, imx, hx, ihx) schedule(dynamic,1)
  do ixs=1,nxs
  
  
      inode=omp_get_thread_num(); 
    if(nodes==0.) inode=0; 
    if(inode>=inode1) inode1=inode;
    write(0,*) 'ixs, inode', ixs, inode
  
       xs = (ixs-1)*dxs + oxs
    do ixa=1,nxa  
       xa = (ixa-1)*dxa + oxa
       td = sqrt(Gz*Gz + (xa-xs)**2)/vw
       itd = (td-otd)/dtd + 1.5
       mx = 0.5*(xa + Gx)
       imx = (mx-omx)/dmx + 1.5
       hx = abs(0.5*(Gx-xa))
       ihx = (hx-ohx)/dhx + 1.5
       
       !$OMP CRITICAL
       ! Up going event redatuming
       Outarray(1:nt-itd,imx,ihx,1) = Outarray(1:nt-itd,imx,ihx,1) + Inarray(ixs,itd+1:nt,1)
       
       ! down-going event redatuming??
       !$OMP END CRITICAL
    end do
  end do
  !$OMP end parallel do     
  
  ! Output file
  
  call sep_write(Outarray, out_, esize=4)
  call auxputch("n1",'i', nt, trim(out_));   call auxputch("label1",'s', 't', trim(out_)); 
  call auxputch("d1",'f', dt, trim(out_));  call auxputch("o1",'f', ot, trim(out_));
  call auxputch("n2",'i', nmx, trim(out_));      call auxputch("label2",'s', 'mx', trim(out_)); 
  call auxputch("d2",'f', dmx, trim(out_));  call auxputch("o2",'f', omx, trim(out_));
  call auxputch("n3",'i', nhx, trim(out_));      call auxputch("label3",'s', 'offset', trim(out_)); 
  call auxputch("d3",'f', dhx, trim(out_));  call auxputch("o3",'f', ohx, trim(out_));   
  call auxputch("n4",'i', 2, trim(out_));      call auxputch("label4",'s', 'UD', trim(out_)); 
  call auxputch("d4",'f', 1.0, trim(out_));  call auxputch("o3",'f', 0., trim(out_)); 
  
  
end program
