program rtmOff2d_Main
  ! Shot profile Reverse Time Migration for up and down going data
  ! Input: Expect two set of data one for up going waves and down going waves
  !      : axis 1 = nshot, 2=x, 3=t, 4=UD
  !      : Expect Sou wavefield to be pre-computed
  !      : axis 1=nz, 2=nx, 3=nt, 4=nshot, 5=UD    
  ! input : assume all source wavelet is the same
  !              source file: n1=nt, source location from param: osx, dsx, zs\
  ! input:  inv =0 (Adjoint RTM), 
  ! Propagation Grid: use the mesh defined by velocity model plus bndx and bndz are padding. 
  !              therefore, drx should equal dx
  ! Time step : use the time step from receiver. 
  !             therefore nst .le. nrt
  !                       dst .eq. drt
  

  use sep
  use rtmOff2d_oper
  use waveprop_mod
  use waveprop2d_mod
  use type3d_mod
  use smooth_mod
  
  

  implicit none

  type(velo) :: vel
  type(sou_grid)  :: sougrid
  type(rec_grid)  :: recgrid
  type(bc_values) :: bc


  
  logical                   :: smooth, verb, rbc
  integer                   ::  it, id, stat, niter, itercount
  integer                   :: iz, ix, nz, nx, nt, jsnap, inv, shotnum
  integer                   :: bndx, bndz, nnz, nnx, nw, nmx, nmz, nhx
  character(len=128)        :: sou_, rec_, velout_, wfldsou_, wfldrec_, wfldmod_
  character(len=128)        :: vp_, out_, MWfile_, DWfile_
  real                      :: xs, xr
  real                      :: rho
  integer                   :: inputmode
  real, dimension(:,:), allocatable  :: tmpdata
  real, dimension(:,:,:), allocatable  :: tmpmodel


  
 


  call sep_init()


  
  
  !!======================================================
  !! Parameter reading
  
  write(0,*) ' This is the beginning of rtmOff_main'



   call from_param("verb",verb,.false.);     call from_param("smooth",smooth,.false.);   
   call from_param("wfrout",bc%wfrout,.false.);       call from_param("wfmout",bc%wfmout,.false.);
   call from_param("DW", bc%DW, .false.);
   
   !! preconditioning
   call from_param("lpfil", bc%lpfil, .false.);     call from_param("VDer", bc%VDer, .false.);
   call from_param("gauss", bc%gauss, .false.);


write(0,*) 'Input 1'
if (1 == 1) then   

   ! Read file name
   call from_param("velin", vp_);

  
   
   call from_param("wfldsou", wfldsou_);      call from_param("recfile",rec_);
   call from_param("MWfile", MWfile_);        
   if (bc%DW)  call from_param("DWfile", DWfile_);





   ! Read back ground velocity param
   call from_aux(vp_, "n1", vel%z%n);  call from_aux( vp_, "d1", vel%z%d);  call from_aux( vp_, "o1",vel%z%o);
   call from_aux( vp_, "n2", vel%x%n);  call from_aux( vp_, "d2", vel%x%d);  call from_aux( vp_, "o2",vel%x%o);
   call from_param('zw',vel%zw);
   call from_param('nhx', vel%h%n, 0);   
    nw = (vel%zw-vel%z%o)/vel%z%d + 0.5



   !All shots has the same source wavelet
   call from_aux(wfldsou_, "n3", sougrid%t%n);  call from_aux(wfldsou_,"d3", sougrid%t%d);   call from_aux(wfldsou_,"o3", sougrid%t%o);
   sougrid%ntup = sougrid%t%n; 
   sougrid%dtup = sougrid%t%d;
   sougrid%otup = sougrid%t%o;


   sougrid%x%d = 1.
   call from_param('Gx',sougrid%x%o);     call from_param('Gz',sougrid%z%o);


   call from_aux(rec_, "n1", recgrid%x%n);  call from_aux(rec_,"d1", recgrid%x%d);   call from_aux(rec_,"o1", recgrid%x%o);
   call from_aux(rec_, "n2", recgrid%t%n);  call from_aux(rec_,"d2", recgrid%t%d);   call from_aux(rec_,"o2", recgrid%t%o);
   call from_param('zr', recgrid%z%o);
   
   ! Data weighting and Masking
   call from_param("bmask",bc%bmask,recgrid%x%o);  call from_param("emask",bc%emask,recgrid%x%o+(recgrid%x%n-1)*recgrid%x%d);

   ! Read in Density
   call from_param('rho', rho);
  
   
   !!  taper coefficients and padding
   call from_param("xtapco",bc%xtapco,0.0002);  call from_param("ztapco",bc%ztapco,0.0002)

   bc%nop = 2
   call from_param("bndz",bc%bndz,bc%nop); if(bc%bndz<bc%nop) bc%bndz=bc%nop; 
   call from_param("bndx",bc%bndx,bc%nop); if(bc%bndx<bc%nop) bc%bndx=bc%nop;

write(0,*) 'Input 5'


   !! Auxillary output files
   call from_param("outtag", out_);
   call from_param("velfile", velout_, "velfile.H");
   call from_param("wfldrec", wfldrec_, "wfldrec.H");
   call from_param("wfldmod", wfldmod_, "wfldmod.H");
   call from_Param("jsnap", jsnap, 100);


write(0,*) 'Input 6' 
  ! Error check
   if ((sougrid%x%o .lt. vel%x%o) .or. (sougrid%x%o .gt. (vel%x%o+(vel%x%n-1)*vel%x%d))) then
           write(0,*) 'Gx, vel%x%o, vel-xmax', sougrid%x%o, vel%x%o, (vel%x%o+(vel%x%n-1)*vel%x%d)
      	   call erexit('shotnum value is invalid. must be >0 and <velgrid%x%n')
   end if

   if (jsnap .le. 0) jsnap=50

!   if (recgrid%x%d .ne. vel%x%d) call erexit('rec dx does not match vel dx')
   
   write(0,*) 'nst, nrt', sougrid%t%n, recgrid%t%n
   if (sougrid%t%n .gt. recgrid%t%n) call erexit('nst should be less than nrt') 
   
   if (recgrid%t%n .gt. 8001) call erexit('too many time steps, nt < 8001')
   
!   if (sougrid%t%d .ne. recgrid%t%d) call erexit('rec dt does not match sou dt')

   if ((recgrid%z%n .eq. 2) .and. (recgrid%z%d .eq. 0)) call erexit('Expect over/under but found dz=0')


   !! Definition
   vel%smooth = smooth
   sougrid%x%n = 1
   nnz = vel%z%n
   nnx = vel%x%n
   bndx = bc%bndx
   bndz = bc%bndz
   vel%z%n = vel%z%n + 2*bndz; 
   vel%x%n = vel%x%n+2*bndx
   nz = vel%z%n;
   nx = vel%x%n;
   nt=recgrid%t%n
   nw = (vel%zw-vel%z%o)/vel%z%d + .5
   nhx = vel%h%n
   

  
   !*********************************
   write(0,*) '**************************'
   write(0,*) 'vel: nz, dz, oz ', nz, vel%z%d, vel%z%o
   write(0,*) 'vel: nx, dx, ox',  nx, vel%x%d, vel%x%o
   write(0,*) 'vel: zw, nw'         ,  vel%zw, nw
   write(0,*) 'sou: nx, dx, ox',  sougrid%x%n, sougrid%x%d, sougrid%x%o  
   write(0,*)  'sou: nt, dt',     sougrid%t%n, sougrid%t%d
   write(0,*) 'rec: nx, dx, ox', recgrid%x%n, recgrid%x%d, recgrid%x%o
   write(0,*) 'nt, dt',          recgrid%t%n, recgrid%t%d
   write(0,*) ' nrz', recgrid%z%n
   write(0,*) 'bcOp, verb', bc%Op, verb
   write(0,*) 'smooth, wfrout, wfmout', smooth, bc%wfrout, bc%wfmout
   write(0,*) 'UDout', bc%UDout
   write(0,*) ' zr', recgrid%z%o
   write(0,*) 'bndz, bndx', bc%bndz, bc%bndx
   write(0,*) '**************************'



   !----------------------------------------------------------------

   
  !! output files
   call auxputch("n1",'i', nnz-nw, trim(out_));   call auxputch("label1",'s', 'z', trim(out_)); 
   call auxputch("d1",'f', vel%z%d, trim(out_));  call auxputch("o1",'f', vel%zw, trim(out_));
   call auxputch("n2",'i', nnx, trim(out_));      call auxputch("label2",'s', 'x', trim(out_)); 
   call auxputch("d2",'f', vel%x%d, trim(out_));  call auxputch("o2",'f', vel%x%o, trim(out_));
   call auxputch("n3",'i', 2*nhx+1, trim(out_));      call auxputch("label3",'s', 'offset', trim(out_)); 
   call auxputch("d3",'f', vel%x%d, trim(out_));  call auxputch("o3",'f', -1.*vel%x%d*nhx, trim(out_));   





   ! Allocation
   write(0,*) 'allocating', recgrid%x%n,nt
   write(0,*) 'allocating size', nnz-nw,nnx,2*nhx+1
   allocate(tmpdata(recgrid%x%n,nt), tmpmodel((nnz-nw),nnx,(2*nhx+1)))



   ! initialize
   tmpdata = 0.;
   tmpmodel =0.;



   write(0,*) 'before reading rec data'
   ! Read in the rec data
   call sep_read(tmpdata, rec_)


   
   if (bc%wfrout) then

         call auxputch("n1",'i',nz,trim(wfldrec_));   
         call auxputch("d1",'f',vel%z%d,trim(wfldrec_));   call auxputch("o1",'f',vel%z%o,trim(wfldrec_));
	 call auxputch("label1", 's', 'Z', trim(wfldrec_));

         call auxputch("n2",'i',nx,trim(wfldrec_)); 
         call auxputch("d2",'f',vel%x%d,trim(wfldrec_));   call auxputch("o2",'f',vel%x%o,trim(wfldrec_));
	 call auxputch("label2", 's', 'Z', trim(wfldrec_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldrec_));   
         call auxputch("d3",'f',-1*recgrid%t%d*jsnap,trim(wfldrec_));   
	 call auxputch("o3",'f',recgrid%t%o+(nt-jsnap)*recgrid%t%d,trim(wfldrec_));
 	 call auxputch("label3", 's', 'time', trim(wfldrec_));

   end if

   if (bc%wfmout) then

         call auxputch("n1",'i',nnz-nw,trim(wfldmod_));    call auxputch("label1", 's', 'Z', trim(wfldmod_));
         call auxputch("d1",'f',vel%z%d,trim(wfldmod_));   call auxputch("o1",'f',vel%z%o,trim(wfldmod_));

         call auxputch("n2",'i',nnx,trim(wfldmod_));   	 call auxputch("label2", 's', 'X', trim(wfldmod_));
         call auxputch("d2",'f',vel%x%d,trim(wfldmod_));   call auxputch("o2",'f',vel%x%o,trim(wfldmod_));

!        call auxputch("n3",'i', 5, trim(wfldmod_));
         call auxputch("n3",'i',nt/jsnap,trim(wfldmod_));   
         call auxputch("d3",'f',-1.*recgrid%t%d*jsnap,trim(wfldmod_));   
         call auxputch("o3",'f',recgrid%t%o,trim(wfldmod_));
 	 call auxputch("label3", 's', 'time', trim(wfldmod_));         

   end if





   ! Call L_initiate
     stat=L_init(sougrid, recgrid, vel, wfldsou_,vp_,rho, bc, wfldrec_in=wfldrec_, wfldmod_in=wfldmod_, &
                 jsnap_in=jsnap, verb_in=verb, velout_in=velout_,MWfile_in=MWfile_, DWfile_in=DWfile_)


   ! ***************************   
   !  Main Body of the program
   ! ***************************

      ! Adjoint only
      ! RTM: data ---> reflectivity 
      stat=L_oper(.true.,.false.,tmpmodel,tmpdata)

 
 
 

 
   write(0,*) 'pass 7'


   call sep_write(tmpmodel,out_,esize=4)



      call sep_close()
     
     deallocate(tmpdata, tmpmodel);

end if
  
 end program
 
