
module rtmOff2d_oper


  use sep
  use type3d_mod
  use waveprop_mod
  use waveprop2d_mod
  use smooth_mod

  implicit none
  
  type(velo)       :: vel_grid

  logical, private                             :: verb, wfrout, wfmout
  logical, private                             :: smooth, VDer, lpfil, gauss, DW
  integer, private                             :: nx, nz, nrx, nt, nw, nnx, nnz, nhx
  integer, private                             :: i, ii, it, ix, iz, itmin
  integer, private                             ::  izr, ixr, irx, ixs, ibmask, iemask
  integer, private                             :: nop, bndx, bndz, jsnap, ndata
  integer, private                             :: xbsz, zbsz
  real, private                                :: dx, ox, dz, oz, dsx, osx, drx, orx, dt, ot
  real, private                                :: zr, orz, drz, rho, maxv, pi=3.14159265359
  real, private                                :: idx, idz, t
  real, private                                :: xtapco, ztapco
  real*8, private                              :: time1, time2
  real, dimension(:), allocatable,private      :: bndz0, bndz1, bndz2, bndx0, bndx1, xtaper, ztaper, fs, tempfr
  real, dimension(:), allocatable, private     :: xtaper2, ztaper2
  real, dimension(:,:), allocatable,private    :: vp2
  real, dimension(:,:), allocatable, private   :: ctaper, ctaper2, WgtDataUp
  real, dimension(:,:), allocatable,private    :: uo, um, up
  real, dimension(:,:), allocatable, private   ::  DWArray
  real, dimension(:,:,:), allocatable,private  :: us,MWArray, Wup
  character(len=128), private                  ::  wfldrec_, wfldmod_
  integer, dimension(:), allocatable, private  :: UPmarker
  real                                         :: dtup, otup
  integer                                      :: ntup, imark, nzmax
  
   !! For OMP
!  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer            :: inode1, inode, nodes, Totalnodes
  
  contains 
  
   

  integer function L_init( sougrid_in, recgrid_in, velgrid_in, us_in, vp_in, rho_in, bc_in, &
  		  wfldrec_in, wfldmod_in, jsnap_in, verb_in, velout_in, MWfile_in, DWfile_in) result(stat)


    optional ::  wfldrec_in, wfldmod_in, jsnap_in, verb_in, DWfile_in

    type(sou_grid)       :: sougrid_in
    type(rec_grid)       :: recgrid_in
    type(velo)           :: velgrid_in
    type(bc_values)     :: bc_in

    logical              :: verb_in
    integer              ::  ix, iz, ihx, jsnap_in
    integer              :: igx, igz, ibx, ibz, iex, iez
    real                 :: rho_in
    character(len=128)   :: wfldrec_in, wfldmod_in, velout_in, us_in, vp_in
    character(len=128)   :: MWfile_in, DWfile_in


    nx = velgrid_in%x%n
    dx = velgrid_in%x%d
    ox = velgrid_in%x%o
    nz = velgrid_in%z%n
    dz = velgrid_in%z%d
    oz = velgrid_in%z%o
    smooth = velgrid_in%smooth
    nhx = velgrid_in%h%n

    dsx = sougrid_in%x%d
    osx = sougrid_in%x%o
    
    ntup = sougrid_in%ntup
    otup = sougrid_in%otup
    dtup = sougrid_in%dtup

  

    nrx = recgrid_in%x%n
    orx = recgrid_in%x%o
    drx = recgrid_in%x%d
    zr = recgrid_in%z%o

    nt = recgrid_in%t%n
    dt = recgrid_in%t%d
    ot = recgrid_in%t%o
   
    bndx = bc_in%bndx
    bndz = bc_in%bndz
    
    nop = bc_in%nop
    nnx = nx - 2*bndx
    nnz = nz - 2*bndz

    nw = (velgrid_in%zw - oz)/dz + .5
    wfrout  = bc_in%wfrout
    wfmout   = bc_in%wfmout
    DW       = bc_in%DW
    lpfil    = bc_in%lpfil
    VDer     = bc_in%VDer
    gauss    = bc_in%gauss
    ibmask   = (bc_in%bmask - orx)/drx + 1.5-1.
    iemask   = (bc_in%emask - orx)/drx + 1.5+1.    

    write(0,*) '@@@', bc_in%emask, orx, drx, iemask

    jsnap = jsnap_in


    vel_grid = velgrid_in

    xbsz = 60
    zbsz = 60

! ********************
    write(0,*) 'jsnap', jsnap
! ********************
    allocate(UPmarker(ntup))
    allocate(tempfr(nt), vp2(nz,nx)) 

	
write(0,*) 'pass 4'
    ndata = nrx*nt

    
    tempfr = 0.;
    vp2=0.;


write(0,*) 'pass 5'

    if (wfrout) then
       wfldrec_ = wfldrec_in
    end if
    if (wfmout) then
       wfldmod_ = wfldmod_in
    end if


    if (present(verb_in)) verb = verb_in

     ! Load up the UPmarker
     write(0,*) 'nt, dt, ot', nt, dt, ot
     write(0,*) 'ntup, dtup, otup', ntup, dtup, otup
     do i=1,ntup
          t = otup + (i-1)*dtup
          ii = (t-ot)/dt + .5
          UPmarker(i) = ii
          write(0,*) 'UPmarker', i, ii
     end do
     
     imark = ntup
  
   ! Read in the background velocity
   call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx), vp_in, esize=4)
!!@   call sep_read( vp2(1+nop:nz-bndz,1+bndx:nx-bndx), vp_in, esize=4)
   
  
   ! Padding
   do iz=1,bndz; vp2(iz,:) = vp2(bndz+1,:); end do ! padding the top with absorbing boundary
   do iz=1,bndz; vp2(nz-iz+1,:) = vp2(nz-bndz,:); end do !padding the bottom
   do ix=1,bndx; vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); end do !padding


    write(0,*) 'allocating 3X', nz*nx
    allocate(up(nz,nx))
    allocate(um(nz,nx))
    allocate(uo(nz,nx))
    
    write(0,*) 'nz, nx'
    
    !----------------------------------------------------------------
    !! one-way absorbing BC and Tapersetup  
    allocate(bndz0(nx),bndz1(nx))      
    allocate(bndx0(nz),bndx1(nz))
    allocate(ztaper(bndz))
    allocate(xtaper(bndx))
    allocate(ctaper(bndz,bndx))
    allocate(ztaper2(bndz))
    allocate(xtaper2(bndx))
    allocate(ctaper2(bndz,bndx))
    


    bndz0 =0.; bndz1=0.; 
    bndx0=0.; bndx1=0.;
    ztaper=0.; xtaper=0.;
    ctaper=0.; ztaper2=0.;
    xtaper2=0.; ctaper2=0.;

    !----------------------------------------------------------------

   
    idx = 1./dx**2; idz = 1./dz**2;
    if (nop .eq. 8) then
    	idx = idx*dt**2;     idz = idz*dt**2
    end if
    
    call  abcsetup(nx,nz,vp2,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,.false.)      
     call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)
     call  tapersetupCos(bndz,bndx,ztaper2,xtaper2,ctaper2,0.8,0.2)

   !  waterv=1500.


    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix) = vp2(iz,ix)*vp2(iz,ix) 
       end do
    end do


  ! call taperingCosV2(nz,nx,nop,bndz,bndx,ztaper2,xtaper2,ctaper2, vp2)


     ! Smoothing velocity model
     if (smooth) then
      write(0,*) 'smoothing the velocity'
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
     end if



  

     
   !! Write out the velocity model
    if (exist_file("velout") ) then

  velout_in = "velfile.H"
         write(0,*) '%%% writing to velout', trim(velout_in)
         call auxputch("n1",'i',nz,trim(velout_in));   
         call auxputch("d1",'f',dz,trim(velout_in));   call auxputch("o1",'f',oz,trim(velout_in));

         call auxputch("n2",'i',nx,trim(velout_in));   
         call auxputch("d2",'f',dx,trim(velout_in));   call auxputch("o2",'f',ox,trim(velout_in));

         write(0,*) 'size vp2', size(vp2) 
         call sep_write(vp2, trim(velout_in), dim1=nz, dim2=nx, esize=4)
 ! call erexit('Exit now')
    end if


 write(0,*) 'pass 8'
    ! Allocation
    write(0,*) 'allocating us', (nnz-nw)*nnx*ntup
    allocate(us(nnz-nw,nnx,ntup))
    us=0.;

    write(0,*) 'allocating 2X', nz*nx


   ! allocating data weighting and model weighting matrix
    allocate(Wup(nnz-nw,nnx,2*nhx+1))
    allocate(WgtDataUp(nrx,nt))

    write(0,*) 'nrx, recgrid_in%x%n', nrx, recgrid_in%x%n
    write(0,*) size(WgtDataUp,1)

    do it=1,nt
       do irx=1,nrx
          WgtDataUp(irx,it) =1.
       end do
    end do    


  

    ! Data muting for unsampled offset
    WgtDataUp(1:ibmask,:) = 0.
    WgtDataUp(iemask:nrx,:) =0.

    write(0,*) 'pass 6', ibmask, iemask

    if (DW) then        
        allocate(DWArray(nrx,nt))
        write(0,*) 'DWfile', DWfile_in
        call sep_read(DWArray,DWfile_in)
	WgtDataUp(:,:) = WgtDataUp(:,:)*DWArray(:,:)
    end if


    ! Model Weighting
    allocate(MWArray(nnz-nw,nnx,2*nhx+1))
    write(0,*) 'MWfile', MWfile_in
    write(0,*) 'mwfile size', nnz-nw, nnx, 2*nhx+1
    call sep_read(MWArray,MWfile_in, dim1=nnz-nw, dim2=nnx, dim3=2*nhx+1, esize=4)

    write(0,*) 'pass 7', sum(MWarray)
!    call erexit('exit to look at MWarray')
 
    do ihx=1,2*nhx+1
    do ix=1,nnx
    	do iz=1,nnz-nw
		Wup(iz,ix,ihx) = 1.;
	end do
    end do   	
    end do
    



    do ihx=1,2*nhx+1
      do ix=1,nnx
         do iz=1,nnz-nw
	    Wup(iz,ix,ihx) = Wup(iz,ix,ihx)*MWArray(iz,ix,ihx)
         end do
      end do
    end do


  if (1 == 2) then

         call sep_write(Wup,'Wup.H',dim1=nnz-nw, dim2=nnx, dim3=2*nhx+1,esize=4)
  
 	 call auxputch("n1",'i',nnz-nw,'Wup.H');   
         call auxputch("d1",'f',dz,'Wup.H');   call auxputch("o1",'f',oz,'Wup.H');
	 call auxputch("label1", 's', 'Z', 'Wup.H');

         call auxputch("n2",'i',nnx,'Wup.H');   
         call auxputch("d2",'f',dx,'Wup.H');   call auxputch("o2",'f',ox,'Wup.H');
	 call auxputch("label2", 's', 'Z', 'Wup.H');

         call auxputch("n3",'i',2*nhx+1,'Wup.H');   
         call auxputch("d3",'f',dx,'Wup.H');   call auxputch("o3",'f',ox,'Wup.H');
	 call auxputch("label3", 's', 'Z', 'Wup.H');
	 
	 call erexit('exit now')
	 
   end if




   ! Read in the sou wavefield
   write(0,*) 'before reading sou wavefield', trim(us_in)
   call sep_read( us, us_in, esize=4)

    
    uo=0.; um=0.; up=0.;


   ! --------------------------------------------------------------
 
    stat=0

  end function L_init

  !---------------------------------------------------------------------------
  !***************************************************************************
  !---------------------------------------------------------------------------
  integer function L_oper(adj, add, model, data) result(stat)


    logical, intent(in)         :: adj, add
    real, dimension(:,:,:)  :: model
    real, dimension(:,:)    :: data 

    
    integer                   :: ix, ihx, iz, iop, id, i
    integer                   :: i1,i2,i3,i4



    stat = 1

    ! Start timer
    call cpu_time(time1)

    
    
    if (.not. add) then
       if (adj) then
          model = 0.
       else
          data=0.
       end if
    end if

    ! Parallel Programming

    Totalnodes = omp_get_num_procs();
    nodes = min(Totalnodes-1,4);
!    nodes = Totalnodes-1;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes
    write(0,*) 'what is model now?', sum(model)
    
    if (adj) then      !!! ### RTM ### !!!



     ! Copy rec wavefield from 1D array back to a 4D array

        imark = ntup

         ! Apply weighting matrix to data space
         do irx=1,nrx
            do it=1,nt
               data(irx,it) = data(irx,it)*WgtDataUp(irx,it)
            end do
         end do

          
       ! initializing the array
       !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1,nx
          do iz=1,nz
             up(iz,ix)=0.; um(iz,ix)=0.; uo(iz,ix)=0.;
          end do          
       end do
       !$OMP end parallel do

    

          do it=nt,1,-1
         
               nzmax = nz

               if (nop .eq. 2) then  !! 4th order
	                   call laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop, it)
	       else if (nop .eq. 3) then  !! 6th order
		           call laplac6(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop,it)
	       else if (nop .eq. 4) then  !! 8th order
		           call laplac8(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop,it)
               else if (nop .eq. 8) then !! 16yh ordder
                           call laplac2d16(adj, um, uo, up, vp2, nx, nzmax, xbsz, zbsz, idx, idz, nop, it, nodes)
	       end if


           !! time step
           !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
           do ix=1,nx
              do iz=1,nz
                 um(iz,ix)=uo(iz,ix); 
                 uo(iz,ix)=up(iz,ix);
              end do          
           end do
          !$OMP end parallel do
          
          

               !! taper
               call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo)
               call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up)
               call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um)


             ! Inject receiver signal  
                izr = (zr-oz)/dz + 1.5 + bndz
               !$OMP parallel do ordered schedule(dynamic,1) private(irx,ixr)
               do irx=1,nrx
                   ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx
                   uo(izr,ixr) = uo(izr,ixr) + data(irx, it)  
               end do 
               !$OMP end parallel do
               if ((mod(it,jsnap) .eq. 0)) write(0,*) 'adjoint it', it, sum(data(:,it)), sum(uo), imark




            !!! RTM: IMAGING CONDITION
            ! Multiply the source and receiver wavefield at each time step (zero lag)
            if (imark .gt. 0) then
       
               if ( (it .eq. UPmarker(imark))) then

	 !       write(0,*) 'imark', imark, it, sum(us(:,:,imark)), sum(model)
                do ix=1,nnx
                   do iz=1,(nnz-nw)
	              model(iz,ix,nhx+1) = model(iz,ix,nhx+1) + us(iz,ix,imark)*uo(iz+bndz+nw,ix+bndx)
                   enddo
                enddo
                
                ! Calculate other subsurface offset image
                do ihx=1,nhx
                   do ix=1+ihx,nnx-ihx
                      do iz=1,(nnz-nw)
   	               model(iz,ix,nhx+1-ihx) = model(iz,ix,nhx+1-ihx) + us(iz,ix-ihx,imark)*uo(iz+bndz+nw,ix+bndx+ihx)
   	               model(iz,ix,nhx+1+ihx) = model(iz,ix,nhx+1+ihx) + us(iz,ix+ihx,imark)*uo(iz+bndz+nw,ix+bndx-ihx)
                      enddo
                   enddo
                end do  
                
                
                imark = imark-1            
              end if
            end if
             

             ! output receiver movie if optioned                 
             if(mod(it,jsnap)==0 .and. wfrout ) then
                write(0,*) 'outputting rec movie,it', it, nz, nx         
                call sep_write(uo(1:nz,1:nx),wfldrec_,dim1=nz,dim2=nx,esize=4)
             end if
             

             ! output model movie if optioned  
             if(mod(it,jsnap)==0 .and. wfmout  ) then
		write(0,*) 'outputting model movie dngoing ,it'
                call sep_write(model(1:nnz-nw,1:nnx,nhx+1), wfldmod_,dim1=nnz-nw,dim2=nnx,esize=4)       

             end if



        enddo ! (end rec time loop)
     


!   ! Apply filter to model
    if (VDer) then
       call Ver2ndDerOff(adj,model,nnz-nw,nnx,nhx)
    end if

!   if (gauss) then
!	call Gaussian2D(adj,model,nnz-nw,nnx)
!   else if (lpfil) then
!	call  Laplacian2D(adj,model,nnz-nw,nnx)
!   else if (VDer) then
!	call Ver2ndDer(adj,model,nnz-nw,nnx)
!   end if


  ! Apply weighting to model
   do ihx=1,2*nhx+1
   do ix=1,nnx
      do iz=1,(nnz-nw)
         model(iz,ix,ihx) = model(iz,ix,ihx)*Wup(iz,ix,ihx)
      end do
   end do
   end do




    else  !!! ### FORWARD Modelling ### !!!

      imark = 1

    write(0,*) 'For 1', sum(model)

       ! Apply weighting to model
   do ihx=1,2*nhx+1
   do ix=1,nnx
      do iz=1,(nnz-nw)
         model(iz,ix,ihx) = model(iz,ix,ihx)*Wup(iz,ix,ihx)
      end do
   end do
   end do
       
       write(0,*) 'For 2', sum(model)
         
!       ! Apply adjoint filtering to model
    if (VDer) then
       call Ver2ndDerOff(adj,model,nnz-nw,nnx,nhx)
    end if

        write(0,*) 'For 3', sum(model)

!       if (gauss) then
!	  call Gaussian2D(adj,model,nnz-nw,nnx)
!       else if (lpfil) then
!	  call  Laplacian2D(adj,model,nnz-nw,nnx)
!       else if (VDer) then
!	  call Ver2ndDer(adj,model,nnz-nw,nnx)
!       end if
  

      !! Rec Loop
       data=0.; 

        
        ! initializing the array
        !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
         do ix=1,nx
            do iz=1,nz
               up(iz,ix)=0.; um(iz,ix)=0.; uo(iz,ix)=0.;
            end do          
         end do
         !$OMP end parallel do
        
           
               do it=1,nt

                    nzmax = nz

                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um)

                  
                   if (nop .eq. 2) then  !! 4th order
        	           call laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop,it)
		   else if (nop .eq. 3) then  !! 6th order
		           call laplac6(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop,it)
		   else if (nop .eq. 4) then  !! 8th order		   	   
		           call laplac8(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop,it)
                   else if (nop .eq. 8) then !! 16yh ordder
                           call laplac2d16(adj, um, uo, up, vp2, nx, nzmax, xbsz, zbsz, idx, idz, nop, it, nodes)
		   end if

                   !! time step
                   !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
                   do ix=1,nx
                      do iz=1,nz
                         um(iz,ix)=uo(iz,ix); 
                         uo(iz,ix)=up(iz,ix);
                       end do          
                   end do
                   !$OMP end parallel do
 
 !                  if  (it .lt. 66) then
 !                  uo(nz/2,nx/2) = uo(nz/2,nx/2) + cos(2*3.14156*(it-33)/100.) 
 !                  end if

                    ! Forward modelling Injection
                     if ((imark .le. ntup)) then
                        if ( (it .eq. UPmarker(imark)  )) then
		          
                       do ix=1,nnx
                          do iz=1,(nnz-nw)
                              uo(iz+nw+bndz,ix+bndx) = uo(iz+nw+bndz,ix+bndx) + us(iz,ix,imark)*model(iz,ix,nhx+1) 
                          enddo
                       enddo
                      
!                       write(0,*) 'imark, sum(uo), sum(us), sum(model)', imark, sum(uo), sum(us(:,:,imark)),sum(model)
                      
                       do ihx=1,nhx
                       do ix=1,nnx
                          do iz=1,(nnz-nw)
                              uo(iz+nw+bndz,ix+bndx) = uo(iz+nw+bndz,ix+bndx) + us(iz,ix,imark)*model(iz,ix,nhx+1-ihx) 
                              uo(iz+nw+bndz,ix+bndx) = uo(iz+nw+bndz,ix+bndx) + us(iz,ix,imark)*model(iz,ix,nhx+1+ihx) 
                          enddo
                       enddo       
                       end do      
                 
			imark = imark+1
                     end if
                    end if

                   !! Collect data
                    izr = (zr-oz)/dz + 1.5 + bndz
                     do irx=1,nrx
                        ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx		
                        data(irx, it) = data(irx, it)  + uo(izr,ixr)
                     end do

                     if ((mod(it,100) .eq. 1)) write(0,*) 'forward sum, it, nop', it, sum(uo), sum(data)

            
                  if(mod(it,jsnap)==0 .and. wfrout .and. (it .ne. nt) ) then
      	 	      write(0,*) 'output ur', it, sum(uo)
                        call sep_write(uo(1:nz,1:nx),wfldrec_,dim1=nz,dim2=nx,esize=4)
                  end if
                  
                  
                  


               end do ! (rec time loop)
    
    do irx=1,nrx
      write(0,*) 'data', irx, sum(data(irx,:))        
    end do

   end if




   ! ------------------------------------------------------

   
    

    if (.not. adj) then
    
    ! Apply weighting matrix to data space
       do irx=1,nrx
         do it=1,nt
            data(irx,it) = data(irx,it)*WgtDataUp(irx,it)
         end do
       end do   
    end if

    do irx=1,nrx
      write(0,*) 'data', irx, sum(data(irx,:))        
    end do
    

    ! end timer
    call cpu_time(time2)
    write(0,*) 'time for L_oper', time2-time1, 'sec'

    stat =0

    end function

    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    integer function L_close() result(stat)

      stat=1

	write(0,*) 'dalloc 1'
      deallocate( tempfr, vp2)

      	write(0,*) 'dalloc 2'
      deallocate(bndz0, bndz1)
	write(0,*) 'dalloc 3'
      deallocate(bndx0, bndx1)
	write(0,*) 'dalloc 4'
      deallocate(xtaper, ztaper, ctaper)
      deallocate(xtaper2, ztaper2, ctaper2)
	write(0,*) 'dalloc 5'
      deallocate(um, up, uo)

	write(0,*) 'dalloc 6'


        write(0,*) 'dealloc 7'
      deallocate(us)

      stat=0

    end function L_close

end module
