module rtm_oper


  use sep
  use type_mod
  use waveprop_mod
  use smooth_mod

  implicit none
  
  type(velo)       :: vel_grid

  logical, private                             :: verb, wfrout, wfmout, residout, UDout
  logical, private                             :: smooth, VDer, lpfil, gauss, DW
  integer, private                             :: bcOp, nsave, xsnap, istore
  integer, private                             :: nshot, nx, nz, nrx, nt, seed, nw, nnx, nnz
  integer, private                             :: i, ishot, it, ix, iz, itmin
  integer, private                             ::  izr, ixr, irx, ixs, ibmask, iemask
  integer, private                             :: nop, bndx, bndz, jsnap, ndata, itercount
  real, private                                :: dx, ox, dz, oz, dsx, osx, drx, orx, dt, ot
  real, private                                ::  zr, orz, drz, rho, maxv, pi=3.14159265359
  real, private                                :: c0, c1, c2, idx, idz, idt2, t , Epsup, Epsdn
  real, private                                :: xtapco, ztapco, delp, amp1, amp2, maxamp1, maxamp2
  real*8, private                              :: time1, time2
  real, dimension(:), allocatable,private      :: bndz0, bndz1, bndz2, bndx0, bndx1, xtaper, ztaper, fs, tempfr, tmpdata
  real, dimension(:), allocatable, private     :: xtaper2, ztaper2
  real, dimension(:,:), allocatable,private    :: vs1sq, vs2sq, vp2, upmodel, dnmodel, tmpmodel
  real, dimension(:,:), allocatable, private   :: ctaper, ctaper2, Wup, Wdn, WgtDataUp, WgtDataDn
  real, dimension(:,:), allocatable,private    :: uo, um, up, uo2, um2, up2, MWArray
  real, dimension(:,:,:), allocatable, private :: fr1, fr2,  DWArray
  real, dimension(:,:,:,:), allocatable,private  :: us
  character(len=128), private                  ::  wfldrec_, wfldmod_, ssave_, up_, dn_, rmov_

  
  
   !! For OMP
  integer, external  :: omp_get_num_procs, omp_get_thread_num
  integer, external  :: srite_window, sreed_window, getch
  integer            :: inode1, inode, nodes, Totalnodes
  
  contains 

  integer function L_init(nshot_in, sougrid_in, recgrid_in, velgrid_in, us_in, vp_in, rho_in, bc_in, &
  		 up_in, dn_in, rmov_in, wfldrec_in, wfldmod_in, jsnap_in, verb_in, velout_in, MWfile_in, DWfile_in) result(stat)


    optional ::  up_in, dn_in, rmov_in, wfldrec_in, wfldmod_in, jsnap_in, verb_in, MWfile_in, DWfile_in

    type(sou_grid)       :: sougrid_in
    type(rec_grid)       :: recgrid_in
    type(velo)           :: velgrid_in
    type(bc_values)     :: bc_in

    logical              :: verb_in
    integer              :: nshot_in, ix, iz, jsnap_in, seed_in
    integer              :: igx, igz, ibx, ibz, iex, iez
    real                 :: rho_in
    character(len=128)   :: wfldrec_in, wfldmod_in, velout_in, up_in, dn_in, rmov_in, us_in, vp_in
    character(len=128)   :: MWfile_in, DWfile_in

    itercount = 0
    nshot = nshot_in

    nx = velgrid_in%x%n
    dx = velgrid_in%x%d
    ox = velgrid_in%x%o
    nz = velgrid_in%z%n
    dz = velgrid_in%z%d
    oz = velgrid_in%z%o
    smooth = velgrid_in%smooth


    dsx = sougrid_in%x%d
    osx = sougrid_in%x%o
    nrx = recgrid_in%x%n
    orx = recgrid_in%x%o
    drx = recgrid_in%x%d
    zr = recgrid_in%z%o

    nt = recgrid_in%t%n
    dt = recgrid_in%t%d
    ot = recgrid_in%t%o


    bcOp = bc_in%Op
    bndx = bc_in%bndx
    bndz = bc_in%bndz
    nop = bc_in%nop
    nnx = nx - 2*bndx
    nnz = nz - 2*bndz
!!    nnz = nz - bndz - nop
    nw = (velgrid_in%zw - oz)/dz + .5
    residout = bc_in%residout
    UDout    = bc_in%UDout
    wfrout  = bc_in%wfrout
    wfmout   = bc_in%wfmout
    DW       = bc_in%DW
    lpfil    = bc_in%lpfil
    VDer     = bc_in%VDer
    gauss    = bc_in%gauss
    ibmask   = (bc_in%bmask - orx)/drx + 0.5
    iemask   = (bc_in%emask - orx)/drx + 0.5    

    xsnap = sougrid_in%t%d/recgrid_in%t%d + 0.5
    jsnap = jsnap_in
    nsave = sougrid_in%t%n 

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    
    vel_grid = velgrid_in



! ********************
    write(0,*) 'xsnap', xsnap, sougrid_in%t%d, recgrid_in%t%d
    write(0,*) 'jsnap', jsnap
 ! call erexit('Exis here')   
! ********************
    allocate(tempfr(nt), vp2(nz,nx)) 
    allocate(fr1(nshot,nrx,nt), fr2(nshot,nrx,nt))
	
write(0,*) 'pass 4'
    ndata = nshot*nrx*nt
    allocate(tmpdata(ndata))
    
    tempfr = 0.;
    vp2=0.;
    fr1=0.; fr2=0.;
    tmpdata=0.;





write(0,*) 'pass 5'
    if (present(up_in)) then
       up_ = up_in
       dn_ = dn_in
    end if

    if (residout .and. present(rmov_in)) then
       rmov_ = rmov_in
    end if

    if (wfrout) then
       wfldrec_ = wfldrec_in
    end if
    write(0,*) 'wfmout', wfmout
    write(0,*) 'wfldmod_in'
    write(0,*) trim(wfldmod_in)
    if (wfmout) then
       wfldmod_ = wfldmod_in
    end if


    if (present(verb_in)) verb = verb_in


 
  
   ! Read in the background velocity
   call sep_read( vp2(1+bndz:nz-bndz,1+bndx:nx-bndx), vp_in, esize=4)
!!@   call sep_read( vp2(1+nop:nz-bndz,1+bndx:nx-bndx), vp_in, esize=4)
   
  

   do iz=1,bndz; vp2(iz,:) = vp2(bndz+1,:); end do ! padding the top with absorbing boundary
!!@   do iz=1,nop; vp2(iz,:) = vp2(nop+1,:);  end do  ! padding at the top reflecting boundary
   do iz=1,bndz; vp2(nz-iz+1,:) = vp2(nz-bndz,:); end do !padding the bottom
   do ix=1,bndx; vp2(:,   ix  ) = vp2(:, 1+bndx); vp2(:,nx-ix+1) = vp2(:,nx-bndx); end do !padding


    write(0,*) 'allocating 6X', nz*nx
    allocate(up2(nz,nx))
    allocate(um2(nz,nx))
    allocate(uo2(nz,nx))
    allocate(up(nz,nx))
    allocate(um(nz,nx))
    allocate(uo(nz,nx))
    
    write(0,*) 'nz, nx'
    
    !----------------------------------------------------------------
    !! one-way absorbing BC and Tapersetup  
    allocate(bndz0(nx),bndz1(nx))      
    allocate(bndx0(nz),bndx1(nz))
    allocate(ztaper(bndz))
    allocate(xtaper(bndx))
    allocate(ctaper(bndz,bndx))
    allocate(ztaper2(bndz))
    allocate(xtaper2(bndx))
    allocate(ctaper2(bndz,bndx))
    


    bndz0 =0.; bndz1=0.; 
    bndx0=0.; bndx1=0.;
    ztaper=0.; xtaper=0.;
    ctaper=0.; ztaper2=0.;
    xtaper2=0.; ctaper2=0.;

    !----------------------------------------------------------------

   
    idx = 1./dx**2; idz = 1./dz**2; idt2= 1./dt**2
    
    call  abcsetup(nx,nz,vp2,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,.false.)      
     call  tapersetupCos(bndz,bndx,ztaper,xtaper,ctaper,0.98,0.02)
     call  tapersetupCos(bndz,bndx,ztaper2,xtaper2,ctaper2,0.6,0.4)

   !  waterv=1500.


    do ix=1,nx
       do iz=1,nz      
          vp2(iz,ix) = vp2(iz,ix)*vp2(iz,ix) 
       end do
    end do


   call taperingCosV2(nz,nx,bndz,bndx,ztaper2,xtaper2,ctaper2, vp2)


     ! Smoothing velocity model
     if (smooth) then
      write(0,*) 'smoothing the velocity'
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
      call triangle2(8, 8, nz, nx, vp2, vp2)
     end if



  

     
   !! Write out the velocity model
    if (exist_file("velout") ) then

         write(0,*) '%%% writing to velout', trim(velout_in)
         call auxputch("n1",'i',nz,trim(velout_in));   
         call auxputch("d1",'f',dz,trim(velout_in));   call auxputch("o1",'f',oz,trim(velout_in));

         call auxputch("n2",'i',nx,trim(velout_in));   
         call auxputch("d2",'f',dx,trim(velout_in));   call auxputch("o2",'f',ox,trim(velout_in));

         write(0,*) 'size vp2', size(vp2) 
         call sep_write(vp2, trim(velout_in), dim1=nz, dim2=nx, esize=4)

    end if


 write(0,*) 'pass 8'
    ! Allocation
    write(0,*) 'allocating us', (nnz-nw)*nnx*nsave*2
    allocate(us(nnz-nw,nnx,nsave,2))
    us=0.;

    write(0,*) 'allocating 2X', nz*nx
    allocate(tmpmodel(nnz-nw,nnx), upmodel(nnz-nw,nnx), dnmodel(nnz-nw,nnx))


   ! allocating data weighting and model weighting matrix
    allocate(Wup(nnz-nw,nnx), Wdn(nnz-nw,nnx))
    allocate(WgtDataUp(nrx,nt), WgtDataDn(nrx,nt))

    do it=1,nt
       do irx=1,nrx
          WgtDataUp(irx,it) =1.
          WgtDataDn(irx,it) =1.
       end do
    end do    




    ! Data muting for unsampled offset
!    WgtDataUp(1:ibmask,:) = 0.
!    WgtDataDn(1:ibmask,:) = 0.
!    WgtDataUp(iemask:nrx,:) =0.
!    WgtDataDn(iemask:nrx,:) =0.

    write(0,*) 'pass 6'

!    if (DW) then        
!        allocate(DWArray(nrx,nt,2))
!        call sep_read(DWArray,DWfile_in)
!	WgtDataUp(:,:) = WgtDataUp(:,:)*DWArray(:,:,1)
!	WgtDataDn(:,:) = WgtDataDn(:,:)*DWArray(:,:,2)
!    end if

!    write(0,*) 'pass 7 , MWfile', MWfile_in, nnz-nw, nnx

    ! Model Weighting
    allocate(MWArray(nnz-nw,nnx))
    MWArray = 1.0
!    call sep_read(MWArray,MWfile_in, dim1=nnz-nw, dim2=nnx, esize=4)

    write(0,*) 'pass 8'
 

       do ix=1,nnx
       	  do iz=1,nnz-nw
		tmpmodel(iz,ix) = 0.;
		upmodel(iz,ix) = 0.;
		dnmodel(iz,ix) = 0.;
		Wup(iz,ix) = 1.;
                Wdn(iz,ix) = 1.;
           end do      
	end do
  	

    

!    do ix=1,nnx
!       tapebd(ix) = 150 - 50*sin(pi*(ix-1)/(nnx-1))**2
!    end do
 
!     write(0,*) 'setting up Wup matrix'
!     call WmodeltaperSetup(tapebd,100,Wup,nnz-nw,nnx,0.3 ,0.2)
!     do ix=1,nnx
!        do iz=1,nnz-nw
!           Wdn(iz,ix) = 1. - Wup(iz,ix)
!        end do
!    end do
    


      do ix=1,nnx
         do iz=1,nnz-nw
	    Wup(iz,ix) = Wup(iz,ix)*MWArray(iz,ix)
            Wdn(iz,ix) = Wdn(iz,ix)*MWArray(iz,ix)
         end do
      end do






  if (1 == 2) then


         call sep_write(Wup,'Wup.H',dim1=nnz-nw, dim2=nnx,esize=4)
         call sep_write(Wdn,'Wup.H',dim1=nnz-nw, dim2=nnx,esize=4) 
  
 	 call auxputch("n1",'i',nnz-nw,'Wup.H');   
         call auxputch("d1",'f',dz,'Wup.H');   call auxputch("o1",'f',oz,'Wup.H');
	 call auxputch("label1", 's', 'Z', 'Wup.H');

         call auxputch("n2",'i',nnx,'Wup.H');   
         call auxputch("d2",'f',dx,'Wup.H');   call auxputch("o2",'f',ox,'Wup.H');
	 call auxputch("label2", 's', 'Z', 'Wup.H');

         call auxputch("n3",'i',2,'Wup.H');

   call erexit('Exit here Check your Wdn')

   end if




   ! Read in the sou wavefield
   write(0,*) 'before reading sou wavefield', trim(us_in)
   call sep_read( us, us_in, esize=4)

    
    uo=0.; um=0.; up=0.;
    uo2=0.; um2=0.; up2=0.;


    ! Initialization
    tmpmodel=0


   ! --------------------------------------------------------------
 
    stat=0

  end function L_init

  !---------------------------------------------------------------------------
  !***************************************************************************
  !---------------------------------------------------------------------------
  integer function L_oper(adj, add, model, data) result(stat)


    logical, intent(in)         :: adj, add
    real, dimension(:)  :: model, data 

    
    integer                   :: ix, iz, iop, id, i
    integer                   :: i1,i2,i3,i4
    real                      :: datasum, frsum


    stat = 1

    ! Start timer
    call cpu_time(time1)

    
    
    if (.not. add) then
       if (adj) then
          model = 0.
       else
          data=0.
       end if
    end if

    ! Parallel Programming

    Totalnodes = omp_get_num_procs();
    nodes =1;
!    nodes = Totalnodes-1;
    call omp_set_num_threads(nodes);
    inode1=0
    write(0,*) 'OMP: using', nodes, 'out of', Totalnodes
    
    
    if (adj) then      !!! ### RTM ### !!!



     ! Copy rec wavefield from 1D array back to a 4D array

     
        tmpdata = data(1:ndata)      
	fr1 = RESHAPE(tmpdata, (/nshot,nrx,nt/))
        tmpdata = data((ndata+1):(2*ndata))  
        fr2 = RESHAPE(tmpdata, (/nshot,nrx,nt/))


         ! Apply weighting matrix to data space
         do irx=1,nrx
            do it=1,nt
               fr1(1,irx,it) = fr1(1,irx,it)*WgtDataUp(irx,it)
               fr2(1,irx,it) = fr2(1,irx,it)*WgtDataDn(irx,it)
            end do
         end do


      !! Initizalize the array
         tmpmodel=0.; upmodel=0.; dnmodel=0.;
         
         
     ! Shot loop
       do ishot=1,nshot
                     
       ! initializing the array

       do ix=1,nx
          do iz=1,nz
             up(iz,ix)=0.; um(iz,ix)=0.; uo(iz,ix)=0.;
             up2(iz,ix)=0.; um2(iz,ix)=0.; uo2(iz,ix)=0.;
          end do          
       end do
 



          do it=nt,1,-1
         
       
       		
              !! 4th order laplacian
               call laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)
               call laplac4(adj, um2, uo2,up2, vp2, dt, nx, nz, idx, idz, nop)


 !          call absorbingBCHale(adj, vel_grid,dt, vp2, um,uo,up)

           !! time step
     
           do ix=1,nx
              do iz=1,nz
                 um(iz,ix)=uo(iz,ix); 
                 uo(iz,ix)=up(iz,ix);
                 um2(iz,ix)=uo2(iz,ix); 
                 uo2(iz,ix)=up2(iz,ix);
              end do          
           end do

          
          

               !! taper
               call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo)
               call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up)
               call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um)
	       call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo2)
	       call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up2)
	       call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um2)


             ! Inject receiver signal  
                izr = (zr-oz)/dz + 1.5 + bndz

               do irx=1,nrx
                   ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx
                   uo(izr,ixr) = uo(izr,ixr) + fr1(ishot, irx, it)  
                   uo2(izr,ixr) = uo2(izr,ixr) + fr2(ishot, irx, it)  
               end do 

               if ((mod(it,100) .eq. 0)) write(0,*) 'adjoint it', it





            !!! RTM: IMAGING CONDITION
            ! Multiply the source and receiver wavefield at each time step (zero lag)
 
       
               if ((it .lt. nsave*xsnap) .and. ((xsnap .eq. 1) .or. (mod(it,2*xsnap) .eq. 0))) then
	        istore = it/xsnap
	        
		
                do ix=1,nnx
                   do iz=1,(nnz-nw)
	              upmodel(iz,ix) = upmodel(iz,ix) + us(iz,ix,istore,1)*uo(iz+bndz+nw,ix+bndx)
		      dnmodel(iz,ix) = dnmodel(iz,ix) + us(iz,ix,istore,2)*uo2(iz+bndz+nw,ix+bndx)
                   enddo
                enddo

             end if
       
             ! output receiver movie if optioned                 
             if(mod(it,jsnap)==0 .and. wfrout .and. ishot==1 ) then
                write(0,*) 'outputting rec movie dngoing ,it', it, sum(fr1(1,1:nrx,it)), sum(fr2(1,1:nrx,it))               
                call sep_write(uo(1:nz,1:nx),wfldrec_,dim1=nz,dim2=nx,esize=4)
             end if
             
	     jsnap = xsnap*2
             ! output model movie if optioned  
             if(mod(it,jsnap)==0 .and. wfmout .and. ishot==1 ) then
		istore = it/jsnap*2
		write(0,*) 'outputting model movie dngoing ,it'
                call sep_write(upmodel(1:nnz-nw,1:nnx), wfldmod_,dim1=nnz-nw,dim2=nnx,esize=4)
                call sep_write(8500000.*us(1:nnz-nw,1:nnx,istore,1), "16shots/L2O/dm/Smov.H",dim1=nnz-nw,dim2=nnx,esize=4)
                call sep_write(100*uo(1+nop+nw:nnz+nop,1+bndx:nnx+bndx),"16shots/L2O/dm/Rmov.H",dim1=nnz-nw,dim2=nnx,esize=4)                  

             end if



        enddo ! (end rec time loop)
     enddo ! (shot loop)
     
write(0,*) 'at least I finish the rtm time loop'

   ! Apply filter to model
!   if (gauss) then
!	call Gaussian2D(adj,upmodel,nnz-nw,nnx)
! 	call Gaussian2D(adj,dnmodel,nnz-nw,nnx)       
!   else if (lpfil) then
!	call  Laplacian2D(adj,upmodel,nnz-nw,nnx)
!	call  Laplacian2D(adj,dnmodel,nnz-nw,nnx)
!   else if (VDer) then
!	call Ver2ndDer(adj,upmodel,nnz-nw,nnx)
!	call Ver2ndDer(adj,dnmodel,nnz-nw,nnx)
!   end if


!   ! Apply weighting to model
!   do ix=1,nnx
!      do iz=1,(nnz-nw)
!         upmodel(iz,ix) = upmodel(iz,ix)*Wup(iz,ix)
!         dnmodel(iz,ix) = dnmodel(iz,ix)*Wdn(iz,ix)
!      end do
!   end do


!   do ix=1,nnx
!     do iz=1,(nnz-nw)
!        tmpmodel(iz,ix) = (upmodel(iz,ix) + 2*dnmodel(iz,ix))
!     end do
!   end do

   if (UDout) then
     write(0,*) 'writing out to upmodel, dnmodel'
     call sep_write(upmodel(1:nnz-nw,1:nnx),up_, dim1=nnz-nw,dim2=nnx,esize=4)
     call sep_write(dnmodel(1:nnz-nw,1:nnx),dn_, dim1=nnz-nw,dim2=nnx,esize=4)
     
          call auxputch("n1",'i',nnz-nw,up_); call auxputch("d1", 'f', dz, up_);  call auxputch("o1", 'f', 500., up_); 
     call auxputch("n2",'i',nnx,up_);  call auxputch("d2", 'f', dx, up_);  call auxputch("o2", 'f', ox, up_); 
     call auxputch("label1", 's', 'Z',up_);
     call auxputch("label2", 's', 'X', up_);


     call auxputch("n1",'i',nnz-nw,dn_);    call auxputch("d1", 'f', dz, dn_);  call auxputch("o1", 'f', 500., dn_); 
     call auxputch("n2",'i',nnx,dn_);  call auxputch("d2", 'f', dx, dn_);  call auxputch("o2", 'f', ox, dn_); 
     call auxputch("label1", 's', 'Z', dn_);
     call auxputch("label2", 's', 'X', dn_);
     
   end if

    else  !!! ### FORWARD Modelling ### !!!


      ! Copy reflectivity model from 1D array back to 2D array
       tmpmodel = RESHAPE( model, (/nnz-nw,nnx/))
       do ix=1,nnx
          do iz=1,(nnz-nw)
		upmodel(iz,ix) = tmpmodel(iz,ix)
		dnmodel(iz,ix) = 2*tmpmodel(iz,ix)
          end do
       end do


       ! Apply weighting to model
       do ix=1,nnx
         do iz=1,(nnz-nw)
            upmodel(iz,ix) = upmodel(iz,ix)*Wup(iz,ix)
            dnmodel(iz,ix) = dnmodel(iz,ix)*Wdn(iz,ix)
         end do
       end do
        
       ! Apply adjoint filtering to model
       if (gauss) then
	  call Gaussian2D(adj,upmodel,nnz-nw,nnx)
 	  call Gaussian2D(adj,dnmodel,nnz-nw,nnx)       
       else if (lpfil) then
	  call  Laplacian2D(adj,upmodel,nnz-nw,nnx)
	  call  Laplacian2D(adj,dnmodel,nnz-nw,nnx)
       else if (VDer) then
	  call Ver2ndDer(adj,upmodel,nnz-nw,nnx)
	  call Ver2ndDer(adj,dnmodel,nnz-nw,nnx)
       end if
  

      !! Rec Loop
       fr1=0.; fr2=0.;

      ! Shot loop
      do ishot=1,nshot
        
        ! initializing the array
         do ix=1,nx
            do iz=1,nz
               up(iz,ix)=0.; um(iz,ix)=0.; uo(iz,ix)=0.;
               up2(iz,ix)=0.; um2(iz,ix)=0.; uo2(iz,ix)=0.;
            end do          
         end do

        
           
               do it=1,nt

                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up)
                    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um)
		    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, uo2)
		    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, up2)
		    call taperingCos(nz,nx,bndz,bndx,ztaper,xtaper, ctaper, um2)


                   !! 4th order laplacian
                   call laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)
                   call laplac4(adj, um2, uo2,up2, vp2, dt, nx, nz, idx, idz, nop)

                   !! time step
                   do ix=1,nx
                      do iz=1,nz
                         um(iz,ix)=uo(iz,ix); 
                         uo(iz,ix)=up(iz,ix);
                         um2(iz,ix)=uo2(iz,ix); 
                         uo2(iz,ix)=up2(iz,ix);
                       end do          
                   end do

 
                    ! Forward modelling Injection
                   if ( (xsnap .eq. 1) .or. ((mod(it,xsnap) .eq. 0) .and. (it .lt. nsave*xsnap)) )then
		      istore = it/xsnap   

                      do ix=1,nnx
                         do iz=1,(nnz-nw)
                             uo(iz+nw+bndz,ix+bndx) = uo(iz+nw+bndz,ix+bndx) + us(iz,ix,istore,1)*upmodel(iz,ix) 
                 	     uo2(iz+nw+bndz,ix+bndx) = uo2(iz+nw+bndz,ix+bndx) + us(iz,ix,istore,2)*dnmodel(iz,ix)
                         enddo
                      enddo

                   end if


                   !! Collect data
                    izr = (zr-oz)/dz + 1.5 + bndz
                     do irx=1,nrx
                        ixr = (orx+(irx-1)*drx -ox)/dx + 1.5 + bndx
 !                       if ((irx .eq. nrx/2) .and. (mod(it,100) .eq. 1)) write(0,*) izr, ixr
                        fr1(ishot, irx, it) = fr1(ishot, irx, it)  + uo(izr,ixr)
                        fr2(ishot, irx, it) = fr2(ishot, irx, it)  + uo2(izr,ixr)
                     end do
                     if ((mod(it,100) .eq. 1)) write(0,*) 'forward sum, it', it, sum(fr2), sum(fr1)


            
                  if(mod(it,jsnap)==0 .and. wfrout .and. ishot==1 .and. (it .ne. nt) ) then
      	 	      write(0,*) 'output ur2', it, sum(uo2)
                        call sep_write(uo2(1:nz,1:nx),wfldrec_,dim1=nz,dim2=nx,esize=4)
                  end if
                  
                  
                  


               end do ! (rec time loop)
        enddo ! (shot loop)

        if (residout) then
           call sep_write(fr1(1,1:nrx,1:nt),rmov_,  dim1=nrx,dim2=nt,esize=4)
           call sep_write(fr2(1,1:nrx,1:nt),rmov_,  dim1=nrx,dim2=nt,esize=4)
        end if
        
        
   end if



if (wfrout) then
         call auxputch("n1",'i',nz,trim(wfldrec_));   
         call auxputch("d1",'f',dz,trim(wfldrec_));   call auxputch("o1",'f',oz,trim(wfldrec_));
	 call auxputch("label1", 's', 'Z', trim(wfldrec_));

         call auxputch("n2",'i',nx,trim(wfldrec_));   
         call auxputch("d2",'f',dx,trim(wfldrec_));   call auxputch("o2",'f',ox,trim(wfldrec_));
	 call auxputch("label2", 's', 'Z', trim(wfldrec_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldrec_));   
         call auxputch("d3",'f',dt*jsnap,trim(wfldrec_));   
	 call auxputch("o3",'f',ot+(jsnap-1)*dt,trim(wfldrec_));
 	 call auxputch("label3", 's', 'time', trim(wfldrec_));
end if

if (wfmout) then 	 
 	 
 	 call auxputch("n1",'i',nnz-nw,trim(wfldmod_));   
         call auxputch("d1",'f',dz,trim(wfldmod_));   call auxputch("o1",'f',oz,trim(wfldmod_));
	 call auxputch("label1", 's', 'Z', trim(wfldmod_));

         call auxputch("n2",'i',nnx,trim(wfldmod_));   
         call auxputch("d2",'f',dx,trim(wfldmod_));   call auxputch("o2",'f',ox,trim(wfldmod_));
	 call auxputch("label2", 's', 'Z', trim(wfldmod_));

         call auxputch("n3",'i',nt/jsnap,trim(wfldmod_));   
         call auxputch("d3",'f',-1*dt*jsnap,trim(wfldmod_));   
	 call auxputch("o3",'f',ot+(nt-jsnap)*dt,trim(wfldmod_));
 	 call auxputch("label3", 's', 'time', trim(wfldmod_));
 	 
end if
   ! ------------------------------------------------------

   
    

    if (.not. adj) then
    
    ! Apply weighting matrix to data space
       do irx=1,nrx
         do it=1,nt
            fr1(1,irx,it) = fr1(1,irx,it)*WgtDataUp(irx,it)
            fr2(1,irx,it) = fr2(1,irx,it)*WgtDataDn(irx,it)
         end do
       end do
    
	ndata = nshot*nt*nrx
       write(0,*) 'cp data back', ndata
       data(1:ndata) = RESHAPE(fr1,SHAPE=(/nshot*nt*nrx/))
       data((ndata+1):(2*ndata)) = RESHAPE(fr2,SHAPE=(/nshot*nt*nrx/))
    else
    
    
       model = RESHAPE(tmpmodel,SHAPE=(/ (nnz-nw)*nnx /) )

   
    end if

    

    ! end timer
    call cpu_time(time2)
    write(0,*) 'time for L_oper', time2-time1, 'sec'

    stat =0

    end function

    !---------------------------------------------------------------------------
    !---------------------------------------------------------------------------
    integer function L_close() result(stat)

      stat=1

	write(0,*) 'dalloc 1'
      deallocate(fr1, fr2, tempfr, vp2)

      	write(0,*) 'dalloc 2'
      deallocate(bndz0, bndz1)
	write(0,*) 'dalloc 3'
      deallocate(bndx0, bndx1)
	write(0,*) 'dalloc 4'
      deallocate(xtaper, ztaper, ctaper)
      deallocate(xtaper2, ztaper2, ctaper2)
	write(0,*) 'dalloc 5'
      deallocate(um, up, uo, um2, up2, uo2)

	write(0,*) 'dalloc 6'
      deallocate(tmpmodel, upmodel, dnmodel)

        write(0,*) 'dealloc 7'
      deallocate(us)
      deallocate(tmpdata)
      
      stat=0

    end function L_close

end module
