module waveprop2d_mod

  use type3d_mod
  use omp_lib

  implicit none

      real                    :: W0_2d12, W1_2d12, W2_2d12, W3_2d12
      real                    :: W4_2d12, W5_2d12, W6_2d12, FN_2d12
      real                    :: W0_2d16, W1_2d16, W2_2d16, W3_2d16, W4_2d16
      real                    :: W5_2d16, W6_2d16, W7_2d16, W8_2d16

!     ------------------------------------------------------
!     16th order optimized FD weights (2nd):
!     ------------------------------------------------------
      PARAMETER  ( W0_2D16 = -3.22511718414079 )  
      PARAMETER  ( W1_2D16 =  1.93606382829618 )  
      PARAMETER  ( W2_2D16 = -0.43844416993301 )
      PARAMETER  ( W3_2D16 =  0.16442503430510 )
      PARAMETER  ( W4_2D16 = -0.07205642556714 )
      PARAMETER  ( W5_2D16 =  0.03264972066062 )
      PARAMETER  ( W6_2D16 = -0.01416701049701 )
      PARAMETER  ( W7_2D16 =  0.00516908814658 )
      PARAMETER  ( W8_2D16 = -0.00108147334093 )
!      PARAMETER  ( FN_2D16 = 1.0 / 1.46229534342707 )
  
!
!     ------------------------------------------------------
!     12th order optimized FD weights, extreme (2nd):
!     ------------------------------------------------------
      PARAMETER  ( W0_2D12  = -3.16513614080052 )  
      PARAMETER  ( W1_2D12  =  1.87834860513654 )  
      PARAMETER  ( W2_2D12  = -0.38710535345218 )
      PARAMETER  ( W3_2D12  =  0.12241455362069 )
      PARAMETER  ( W4_2D12  = -0.04073045512005 )
      PARAMETER  ( W5_2D12  =  0.01159369937042 )
      PARAMETER  ( W6_2D12  = -0.00195297915516 )
      PARAMETER  ( FN_2D12  = 1.0 / 1.41857564413310 )



contains

!!--------------------------------------------------------------

  subroutine Ver2ndDerOff(adj,m,nz,nx,nhx)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, nhx, ix,iz, ihx 
  real, dimension(:), allocatable :: temp
  real, dimension(:,:,:)      :: m


  allocate(temp(nz))

  do ihx=1,2*nhx+1
    do ix=1,nx
      temp(:) = m(:,ix,ihx)  
      m(:,ix,ihx) = 0.
      do iz=2,nz-1
         if (adj) then
            m(iz,ix,ihx) = m(iz,ix,ihx) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
         else
            m(iz,ix,ihx) = m(iz,ix,ihx) + -2*temp(iz)
            m(iz,ix,ihx) = m(iz,ix,ihx) + temp(iz+1)
            m(iz,ix,ihx) = m(iz,ix,ihx) + temp(iz-1)
         end if
       end do
    end do
  end do

  end subroutine 

    
!==========================================================
    subroutine calcBksz(n, nblocks, grid_size,block_size,nnop)
	! Calculate block size 
        
       integer          :: n, nblocks, grid_size, block_size, nnop

       if (mod(n,nblocks) .ne. 0) then
          write(0,*) 'size n=', n, 'nblocks=', nblocks
          call erexit('Error: make sure nblocks divide n')

       end if

  
       grid_size=(n)/nblocks
       block_size = grid_size
 
       if (1.5*nnop .gt. grid_size) call erexit('Error: block size < 1.5*nop')
       if (grid_size .lt. 1) call erexit("Error: model size is too small")

    end subroutine


!==========================================================
    subroutine calcBkNum(nnop,xx)
	! Calculate number of blocks based on block size and grid_size
        

       type(blcIndex)  :: xx
       integer          :: nnop

  
       xx%effn = xx%n - 2*nnop
       xx%effbsz = xx%bsz -2*nnop
       xx%nblocks = ceiling(real(xx%effn)/real(xx%effbsz)) 
       xx%effbszlast = mod(xx%effn,xx%effbsz)
       if (xx%effbszlast .eq. 0) xx%effbszlast = xx%effbsz
       xx%bszlast = xx%effbszlast+2*nnop
  
!       write(0,*) 'n, bsz, nop', xx%n, xx%bsz, nnop
!       write(0,*) 'effn,effbsz,nblocks', xx%effn, xx%effbsz, xx%nblocks
 !      write(0,*) 'effbszlast, bszlast', xx%effbszlast, xx%bszlast

       if (xx%effbsz .le. 5) call erexit('calcBkNum error: effbsz is too small')
       if (xx%effn .lt. 1) call erexit('calcBkNum error: effn is too small')

       
 

    end subroutine

!==========================================================
    subroutine calcBkNum2(effn,effbsz,nblocks,effbszlast,bszlast,n,bsz,nnop)
        ! This is the same as calcBkNum but without the use of type variable
        

       integer          :: nnop, effn, effbsz, nblocks
       integer          :: effbszlast, bszlast,n,bsz

  
       effn = n - 2*nnop
       effbsz = bsz -2*nnop
       nblocks = ceiling(real(effn)/real(effbsz)) 
       effbszlast = mod(effn,effbsz)
       if (effbszlast .eq. 0) effbszlast = effbsz
       bszlast = effbszlast+2*nnop
  
!       write(0,*) 'n, bsz, nop', n, bsz, nnop
!       write(0,*) 'effn,effbsz,nblocks', effn, effbsz, nblocks
 !      write(0,*) 'effbszlast, bszlast', effbszlast, bszlast

       if (effbsz .le. 5) call erexit('calcBkNum error: effbsz is too small')
       if (effn .lt. 1) call erexit('calcBkNum error: effn is too small')

       
 

    end subroutine
!==========================================================
    subroutine calcIndex(x,iblock, nnop)

       type(blcIndex)   :: x
       integer          :: iblock, nnop

       x%iib = (iblock-1)*x%grid_size
       x%padb = min((iblock-1)*2*nnop,nnop)
       x%pade = min((x%nblocks-iblock)*2*nnop,nnop)
       x%ib = x%iib+1-x%padb
       x%ie = x%iib+x%block_size+x%pade
       x%dbloc=x%block_size+x%pade+x%padb-2*nnop


    end subroutine

!==========================================================
    subroutine calcIndex2(x,iblock, nnop)
    ! calculate ib and ie for the block
    ! This version input as fixed block size and then sub-divide the domain
    ! it is called after calcBlkNum function

       type(blcIndex)   :: x
       integer          :: iblock, nnop, curreffbsz

       if (iblock .eq. x%nblocks) then
          x%ib = 1+nnop+(x%nblocks-1)*x%effbsz
          x%ie = x%ib+x%effbszlast-1
          x%dbloc = x%effbszlast
       else
          x%ib = 1+(iblock-1)*x%effbsz+nnop
          x%ie = x%ib + x%effbsz -1
          x%dbloc = x%effbsz 
       end if
       

    end subroutine
!==========================================================
    subroutine calcIndex3(nblocks,ib,ie,effbsz,effbszlast,dbloc,iblock, nnop)
    ! This is the same as calcIndex2, but do not use the type structure
    ! calculate ib and ie for the block
    ! This version input as fixed block size and then sub-divide the domain
    ! it is called after calcBlkNum function

       type(blcIndex)   :: x
       integer          :: iblock, nnop, curreffbsz
       integer          :: nblocks,ib,ie,effbsz,effbszlast,dbloc

       if (iblock .eq. nblocks) then
          ib = 1+nnop+(nblocks-1)*effbsz
          ie = ib+effbszlast-1
          dbloc = effbszlast
       else
          ib = 1+(iblock-1)*effbsz+nnop
          ie = ib + effbsz -1
          dbloc = effbsz 
       end if
       

    end subroutine
!===========================================================
!!--------------------------------------------------------------------------
!! 16th Order approximation to the 2D Laplacian
!! Coefficients are optimized to minimize dispersion
!! This code uses domain decomposition with secified block size
!! Expanding domain in z


  subroutine laplac2d16(adj, um, uo,up, vp2, nx,nzmax, xbsz,zbsz, idx,idz,nop, iit,nodes)

    logical    :: adj
    integer    :: nop, iit, inode, nodes
    integer    :: iblockx, ix, jx, nx
    integer    :: iblockz, iz, jz, nzmax
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: c0, idx, idz
    real       :: c1x, c2x, c3x, c4x, c5x, c6x, c7x, c8x
    real       :: c1z, c2z, c3z, c4z, c5z, c6z, c7z, c8z


    integer    :: xdbloc, xib, xie, nbkx, xeffbszlast, xeffbsz
    integer    :: zdbloc, zib, zie, nbkz, zeffbszlast, zeffbsz
    integer    :: xeffn, xbszlast, xbsz
    integer    :: zeffn, zbszlast, zbsz

    real, dimension(:,:), allocatable   :: Buo, Bvp2

    ! Define cofficients 
  c0 = W0_2d16*(idx+idz)
  c1x = W1_2d16*idx;   c1z =W1_2d16*idz; 
  c2x = W2_2d16*idx;   c2z = W2_2d16*idz;  
  c3x = W3_2d16*idx;   c3z = W3_2d16*idz;
  c4x = W4_2d16*idx;   c4z = W4_2d16*idz; 
  c5x = W5_2d16*idx;   c5z = W5_2d16*idz; 
  c6x = W6_2d16*idx;   c6z = W6_2d16*idz; 
  c7x = W7_2d16*idx;   c7z = W7_2d16*idz; 
  c8x = W8_2d16*idx;   c8z = W8_2d16*idz; 

  ! Calculate Number of Blocks
  call calcBkNum2(xeffn,xeffbsz,nbkx,xeffbszlast,xbszlast,nx,xbsz,nop)
  call calcBkNum2(zeffn,zeffbsz,nbkz,zeffbszlast,zbszlast,nzmax,zbsz,nop)

!  if ((iit .eq. 1) ) write(0,*) 'nbkz,nbkx',nbkz,nbkx
!  if ((iit .eq. 1) ) write(0,*) 'zbsz,xbsz',zbsz,xbsz
!  if ((iit .eq. 1) ) write(0,*) zeffn, zeffbsz, nbkz, zeffbszlast, zbszlast, nzmax, zbsz, nop
!  if ((iit .eq. 1) ) call erexit('exit for now to check blocky stuff')

    if (adj) then    !! RTM

!        write(0,*) 'in the Adjoint direction', iit, nodes

	!$OMP PARALLEL DO DEFAULT(shared) schedule(dynamic,1) num_threads(nodes) PRIVATE(inode, iblockx, xib,zib,xie,zie,xdbloc,zdbloc,iz, jz,ix,jx,Buo) 
	  do iblockx=1,nbkx
 	     call calcIndex3(nbkx,xib,xie,xeffbsz,xeffbszlast,xdbloc,iblockx,nop)

!	   inode = omp_get_thread_num()
!	   write(0,*) 'block #:', iblockx, 'out of', nbkx, 'thread:', inode

           if (allocated(Buo)) deallocate(Buo); allocate(Buo(zbsz,xbsz)) 

            do iblockz=1,nbkz
	       call calcIndex3(nbkz,zib,zie,zeffbsz,zeffbszlast,zdbloc,iblockz,nop)
               !write(0,*) 'iblockz, zib, zie, nop', iblockz, zib, zie, nop
	       Buo(1:zdbloc+2*nop,1:xdbloc+2*nop) =  &
                             uo(zib-nop:zie+nop,xib-nop:xie+nop)

                 do ix = 1+nop,xdbloc+nop
                    jx = ix+xib-1-nop
                   do iz = 1+nop,zdbloc+nop
                    jz = iz+zib-1-nop

                   up(jz,jx) =  2*Buo(iz, ix) - um(jz,jx) + &
                                   vp2(jz, jx)*(  &
                   c0* Buo(iz,  ix   )                              + &
                   c1z*(Buo(iz-1,ix  ) + Buo(iz+1,ix  )) + &
                   c1x*(Buo(iz  ,ix-1) + Buo(iz  ,ix+1)) + &
                   c2z*(Buo(iz-2,ix  ) + Buo(iz+2,ix  )) + &
                   c2x*(Buo(iz  ,ix-2) + Buo(iz  ,ix+2)) + &
                   c3z*(Buo(iz-3,ix  ) + Buo(iz+3,ix  )) + &
                   c3x*(Buo(iz  ,ix-3) + Buo(iz  ,ix+3)) + &
                   c4z*(Buo(iz-4,ix  ) + Buo(iz+4,ix  )) + &
                   c4x*(Buo(iz  ,ix-4) + Buo(iz  ,ix+4)) + &
                   c5z*(Buo(iz-5,ix  ) + Buo(iz+5,ix  )) + &
                   c5x*(Buo(iz  ,ix-5) + Buo(iz  ,ix+5)) + &
                   c6z*(Buo(iz-6,ix  ) + Buo(iz+6,ix  )) + &
                   c6x*(Buo(iz  ,ix-6) + Buo(iz  ,ix+6)) + &
                   c7z*(Buo(iz-7,ix  ) + Buo(iz+7,ix  )) + &
                   c7x*(Buo(iz  ,ix-7) + Buo(iz  ,ix+7)) + &
                   c8z*(Buo(iz-8,ix  ) + Buo(iz+8,ix  )) + &
                   c8x*(Buo(iz  ,ix-8) + Buo(iz  ,ix+8)) )

                  end do
                end do

	   end do
           if (allocated(Buo)) deallocate(Buo);
	 end do
	!$OMP END PARALLEL DO

    else 



	!$OMP PARALLEL DO DEFAULT(shared) schedule(dynamic,1) num_threads(nodes) PRIVATE(inode, iblockx, xib,zib,xie,zie,xdbloc,zdbloc,iz, jz,ix,jx,Buo,Bvp2) 
           
	  do iblockx=1,nbkx
 	     call calcIndex3(nbkx,xib,xie,xeffbsz,xeffbszlast,xdbloc,iblockx,nop)

!	   inode = omp_get_thread_num()
!	   write(0,*) 'block #:', iblockx, 'out of', nbky, 'thread:', inode

           if (allocated(Buo)) deallocate(Buo); allocate(Buo(zbsz,xbsz))
	   if (allocated(Bvp2)) deallocate(Bvp2); allocate(Bvp2(zbsz,xbsz))


            do iblockz=1,nbkz
	       call calcIndex3(nbkz,zib,zie,zeffbsz,zeffbszlast,zdbloc,iblockz,nop)

	       Buo(1:zdbloc+2*nop,1:xdbloc+2*nop) =  &
                             uo(zib-nop:zie+nop,xib-nop:xie+nop)
	       Bvp2(1:zdbloc+2*nop,1:xdbloc+2*nop) = & 
                             vp2(zib-nop:zie+nop,xib-nop:xie+nop)

                 do ix = 1+nop,xdbloc+nop
                    jx = ix+xib-1-nop
                   do iz = 1+nop,zdbloc+nop
                    jz = iz+zib-1-nop

                    up(jz,jx) =  2*Buo(iz, ix) - um(jz,jx) +  &
                    c0* Buo(iz,  ix  )*Bvp2(iz,ix)                              + &
              c1z*(Buo(iz-1,ix)*Bvp2(iz-1,ix) + Buo(iz+1,ix)*Bvp2(iz+1,ix)) + &
              c1x*(Buo(iz,ix-1)*Bvp2(iz,ix-1) + Buo(iz,ix+1)*Bvp2(iz,ix+1)) + &
	      c2z*(Buo(iz-2,ix)*Bvp2(iz-2,ix) + Buo(iz+2,ix)*Bvp2(iz+2,ix)) + &
              c2x*(Buo(iz,ix-2)*Bvp2(iz,ix-2) + Buo(iz,ix+2)*Bvp2(iz,ix+2)) + &
              c3z*(Buo(iz-3,ix)*Bvp2(iz-3,ix) + Buo(iz+3,ix)*Bvp2(iz+3,ix)) + &
              c3x*(Buo(iz,ix-3)*Bvp2(iz,ix-3) + Buo(iz,ix+3)*Bvp2(iz,ix+3)) + &
              c4z*(Buo(iz-4,ix)*Bvp2(iz-4,ix) + Buo(iz+4,ix)*Bvp2(iz+4,ix)) + &
              c4x*(Buo(iz,ix-4)*Bvp2(iz,ix-4) + Buo(iz,ix+4)*Bvp2(iz,ix+4)) + &
              c5z*(Buo(iz-5,ix)*Bvp2(iz-5,ix) + Buo(iz+5,ix)*Bvp2(iz+5,ix)) + &
              c5x*(Buo(iz,ix-5)*Bvp2(iz,ix-5) + Buo(iz,ix+5)*Bvp2(iz,ix+5)) + &
              c6z*(Buo(iz-6,ix)*Bvp2(iz-6,ix) + Buo(iz+6,ix)*Bvp2(iz+6,ix)) + &
              c6x*(Buo(iz,ix-6)*Bvp2(iz,ix-6) + Buo(iz,ix+6)*Bvp2(iz,ix+6)) + &
              c7z*(Buo(iz-7,ix)*Bvp2(iz-7,ix) + Buo(iz+7,ix)*Bvp2(iz+7,ix)) + &
              c7x*(Buo(iz,ix-7)*Bvp2(iz,ix-7) + Buo(iz,ix+7)*Bvp2(iz,ix+7)) + &
              c8z*(Buo(iz-8,ix)*Bvp2(iz-8,ix) + Buo(iz+8,ix)*Bvp2(iz+8,ix)) + &
              c8x*(Buo(iz,ix-8)*Bvp2(iz,ix-8) + Buo(iz,ix+8)*Bvp2(iz,ix+8) )

                end do
              end do


	   end do
           if (allocated(Buo)) deallocate(Buo); 
	   if (allocated(Bvp2)) deallocate(Bvp2);
	 end do
	!$OMP END PARALLEL DO

    end if



  end subroutine

!====================================================================================
end module
