module waveprop_mod

  use sep
  use type3d_mod
	

  implicit none

  
contains

!!--------------------------------------------------------------

  subroutine Hor1stDer(ArrayIn, ArrayOut,nz,nx,dx)
  ! Take the 1d=st derivative horizontally and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, iz, ix, noop
  real                      :: a0, a1, a2, a3, idx, dx
  real, dimension(:,:)      :: ArrayIn, ArrayOut


  noop=3

  a0 = 0.
  a1 = 45./60.*idx
  a2 = -9./60.*idx
  a3 = 1./60.*idx
    
  do ix=1+noop,nx-noop
  
     ArrayOut(:,ix) = (a1*(ArrayIn(:,ix+1)-ArrayIn(:,ix-1)) + &
                       a2*(ArrayIn(:,ix+2)-ArrayIn(:,ix-2)) + &
                       a3*(ArrayIn(:,ix+3)-ArrayIn(:,ix-3)))
  
  end do

  end subroutine 

!!--------------------------------------------------------------

  subroutine Ver1stDer(ArrayIn, ArrayOut,nz,nx,dz)
  ! Take the 1d=st derivative horizontally and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, iz, ix, noop
  real                      :: a0, a1, a2, a3, idz, dz
  real, dimension(:,:)      :: ArrayIn, ArrayOut


  noop=3
  idz=1./dz
  a0 = 0.
  a1 = 45./60.*idz
  a2 = -9./60.*idz
  a3 = 1./60.*idz
    
  do iz=1+noop,nz-noop
  
     ArrayOut(iz,:) = (a1*(ArrayIn(iz+1,:)-ArrayIn(iz-1,:)) + &
                       a2*(ArrayIn(iz+2,:)-ArrayIn(iz-2,:)) + &
                       a3*(ArrayIn(iz+3,:)-ArrayIn(iz-3,:)))

  end do

  end subroutine 

!!--------------------------------------------------------------

  subroutine Hori2ndDer(adj,m,nz,nx)
  ! Take the second derivative horizontally and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, iz
  real, dimension(:), allocatable :: temp
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  allocate(temp(nx))

  do iz=1,nz
    temp(:) = m(iz,:)
    call SecDer(temp, temp, nx, 1.0)
    m(iz,:) = temp(:)
  end do

  end subroutine 


!!--------------------------------------------------------------

  subroutine Ver2ndDer(adj,m,nz,nx)
  ! Take the second derivative vertically and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix,iz 
  real, dimension(:), allocatable :: temp
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d

  allocate(temp(nz))

  do ix=1,nx
    temp(:) = m(:,ix)  
    m(:,ix) = 0.
    do iz=2,nz-1
       if (adj) then
          m(iz,ix) = m(iz,ix) + (-2.*temp(iz) + temp(iz+1) + temp(iz-1))
       else
          m(iz,ix) = m(iz,ix) + -2*temp(iz)
          m(iz,ix) = m(iz,ix) + temp(iz+1)
          m(iz,ix) = m(iz,ix) + temp(iz-1)
       end if
     end do
  end do


  end subroutine 


!!--------------------------------------------------------------

  subroutine Gaussian1D(adj,m,nz,nx)
  ! Take a 1D Gaussian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix, iz, nop,wd, i, j, k
  real                      :: dx, dz, width, val, norm , kk, width2, a
  real, dimension(:), allocatable :: gauss
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  wd = 2
  norm = 0.0
  width = 0.0
  val = 0.0

  allocate(d(nz,nx))
  allocate(gauss(wd+1))  

  do ix=1,nx
     do iz=1,nz
        d=0.
     end do
  end do

   write(0,*) 'Constructing filters'
  ! filter is only half of the entire coverage because it is symmetric about zero
  ! the center of the filter is gauss(1) (largest amp), and the smallest value is gauss(wd+1)
  a = 6.
  width2 = 2.0*real(wd+1)**2
  kk = a*a/width2


    do i=1,wd+1
       val = real(i-1)
       gauss(i) = exp(-kk*val*val)
       norm = norm + gauss(i)
    end do

  gauss = gauss/norm
  gauss = gauss/norm/2.0
        write(0,*) 'gauss filter',  gauss


  write(0,*) 'Applying filter'

  if (.not. adj) then
   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
          d(iz,ix) = d(iz,ix) + gauss(k)*(m(iz+k,ix) + m(iz-k,ix))      
       end do
     enddo
   end do


  else

   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
           d(iz+k,ix) = d(iz+k,ix) + gauss(k)*m(iz,ix)
           d(iz-k,ix) = d(iz-k,ix) + gauss(k)*m(iz,ix)    
       end do
     enddo
   end do
  end if

    write(0,*) 'm, d', m(iz/6,ix/2), d(iz/6,ix/2)


    do ix=1,nx
     do iz=1,nz
        m(iz,ix)=m(iz,ix) - d(iz,ix)
     end do
    end do


  write(0,*) 'outputting Gaussian 1D'

  end subroutine

!!--------------------------------------------------------------
!!--------------------------------------------------------------

  subroutine Gaussian2D(adj,m,nz,nx)
  ! Take a 1D Gaussian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix, iz, nop,wd, i, j, k, l
  real                      :: dx, dz, width2, val, val2, norm, a, kk 
  real, dimension(:,:), allocatable :: gauss
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  wd = 2
  norm = 0.0
  width2 = 0.0
  val = 0.0

  allocate(d(nz,nx))
  allocate(gauss(wd+1,wd+1))  

  do ix=1,nx
     do iz=1,nz
        d=0.
     end do
  end do
 
  write(0,*) 'Constructing filters'
  ! filter is only a quarter of the entire coverage because it is symmetric about zero
  ! the center of the filter is gauss(1,1) (largest amp), and the smallest value is gauss(wd+1,wd+1)
  a = 5.
  width2 = 2.0*real(wd+1)**2
  kk = a*a/width2

  do j=1,wd+1
    do i=1,wd+1
       val = real(i-1)
       val2 = real(j-1)
       gauss(i,j) = exp(-kk*val*val-kk*val2*val2)
       norm = norm + gauss(i,j)
    end do
  end do

  gauss = gauss/norm/4.0
        write(0,*) 'gauss filter',  gauss


  write(0,*) 'Applying filter'

  if (.not. adj) then
   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
         do l=0,wd
          d(iz,ix) = d(iz,ix) + gauss(k,l)*(m(iz+k,ix+l) + m(iz-k,ix+l) + m(iz+k,ix-l) + m(iz-k,ix-l))      
         end do
       end do
     enddo
   end do


  else

   do ix=1+wd, nx-wd  
     do iz=1+wd, nz-wd
       do k=0,wd
         do l=0,wd
           d(iz+k,ix+l) = d(iz+k,ix+l) + gauss(k,l)*m(iz,ix)
           d(iz-k,ix+l) = d(iz-k,ix+l) + gauss(k,l)*m(iz,ix)
           d(iz+k,ix-l) = d(iz+k,ix-l) + gauss(k,l)*m(iz,ix)
           d(iz-k,ix-l) = d(iz-k,ix-l) + gauss(k,l)*m(iz,ix)     
         end do
       end do
     enddo
   end do
  end if

    write(0,*) 'm, d', m(iz/6,ix/2), d(iz/6,ix/2)


    do ix=1,nx
     do iz=1,nz
        m(iz,ix)=m(iz,ix) - d(iz,ix)
     end do
    end do



  write(0,*) 'outputting Gaussian', sum(m)

  end subroutine
!!--------------------------------------------------------------


!!--------------------------------------------------------------

  subroutine Laplacian2D(adj,m,nz,nx)
  ! Take a 2D laplacian and its adjoint

  logical                   :: adj
  integer                   :: nz, nx, ix, iz, nop
  real, dimension(:,:)      :: m
  real, dimension(:,:), allocatable  :: d


  allocate(d(nz,nx))
  
  do ix=1,nx
     do iz=1,nz
        d=0.
     end do
  end do
 
  nop=1 

  if (.not. adj) then
    do ix=1+nop,nx-nop
       do iz=1+nop, nz-nop

          d(iz,ix) = d(iz,ix) + 8.*m(iz,ix) - m(iz+1,ix) - m(iz-1,ix) - &
                     m(iz,ix+1) - m(iz,ix-1) - m(iz+1,ix+1) -&
                     m(iz+1,ix-1) - m(iz-1,ix+1) -m(iz-1,ix-1)

       enddo
    enddo


  else


    do ix=1+nop,nx-nop
       do iz=1+nop, nz-nop
          d(iz,ix) = d(iz,ix) + 8.*m(iz,ix)
          d(iz+1,ix) = d(iz+1,ix) + -1.*m(iz,ix)
	  d(iz-1,ix) = d(iz-1,ix) + -1.*m(iz,ix)
          d(iz,ix+1) = d(iz,ix+1) + -1.*m(iz,ix)
          d(iz,ix-1) = d(iz,ix-1) + -1.*m(iz,ix)
          d(iz+1,ix+1) = d(iz+1,ix+1) + -1.*m(iz,ix)
          d(iz+1,ix-1) = d(iz+1,ix-1) + -1.*m(iz,ix)
          d(iz-1,ix+1) = d(iz-1,ix+1) + -1.*m(iz,ix)
          d(iz-1,ix-1) = d(iz-1,ix-1) + -1.*m(iz,ix)
       enddo
    enddo


  end if

    do ix=1,nx
     do iz=1,nz
        m(iz,ix)=d(iz,ix)
     end do
    end do

  write(0,*) 'Laplace v2', adj

  end subroutine

!!--------------------------------------------------------------
!! Weighting matrix setup for Wmodel

  subroutine WmodeltaperSetup(tapebd,wd,Wup,nz,nx, cent, A)
  ! tapebd = center of tapering zone
  ! wd = width of taper zone
  
    integer                     :: iz, ix, nz, nx, dist, iztp, wd
    integer                     :: izb, ize
    integer, dimension(:)          :: tapebd
    real, dimension(:,:)        :: Wup
    real                        :: cent,A,pi=3.14159265359
    real                        :: ztapco, amp

    ztapco = 1.

   write(0,*) 'nx,nz',nx,nz
   ! Make wd even integer
   if (mod(wd,2) .ne. 0) then
      wd = wd+1
   end if
  
  
   write(0,*) 'hello world 1', wd, tapebd(nx/2)    

    
     do ix=1,nx

        iztp = tapebd(ix)
        izb = max(1,iztp-wd/2)
        ize = izb + wd
!        if (izb .lt. 1) then
!            izb=1
!            ize= izb+wd
!        else if (ize .gt. nz) then
!            ize= nz
!            izb= ize-wd
!        end if

!        write(0,*) 'ix, izb, ize', ix, izb, ize
               
        do iz=max(izb,1),min(ize,nz)
            dist = wd-(izb-iz)
           ! Cosine Taper
            Wup(iz,ix) = (cent + A*cos(pi*(dist)/wd))
!            write(0,*) Wup(iz,ix), iz, ix
 
        end do
           amp = Wup(izb,ix)
	   izb = max(2,izb)
           Wup(1:izb-1,ix) = amp
           amp = Wup(ize,ix)
           Wup(ize+1:nz,ix) = amp
!	   Wup(ize+1:nz,ix) = 0.5
!	   write(0,*) 'test 2', ize, izb
     end do
           


  end subroutine

!! ---------------------------------------------------------
!! Random seed Initialize

   subroutine init_random_seed(seed_in)
      
     integer               :: i, n, clock, seed_in
     integer, dimension(:), allocatable    :: seed

     n=1
     call RANDOM_SEED(size=n)
     allocate(seed(n))

     call SYSTEM_CLOCK(COUNT=clock)
 
     seed = seed_in + 37*(/(i-1,i=1,n)/)
     call RANDOM_SEED(PUT = seed)

     DEALLOCATE(seed)


   end subroutine


!!----------------------------------------------------------
!! Random Boundary Condition Setup
!! Put random number around the outer boundary region
!! bndx, bndz = width of outer boundary in x & z direction
!! nx,nz = total length of the grid (include outer boundary)
!! vp2 = velocity grid (size nz by nx)
!! maxv = bound on max velocity to ensure stable calculation

   subroutine rbcsetup(ishot, v2,Ran1, Ran2, bndx,bndz,nx,nz,seed, maxv)

      logical                :: found
      integer                :: ix, iz, ishot
      integer                :: bndx, bndz, nx, nz, seed, seed_it(2)
      real, dimension(:,:)   :: v2
      real, dimension(:,:,:) :: Ran1, Ran2
      real                   :: maxv, bx, bz, r, d, tempV, minV


      bx = (real(bndx))
      bz = (real(bndz))

      write(0,*) 'Random Boundary Setup'

      call init_random_seed(seed)

      write(0,*) 'after seed'



      do ix=1,nx

 
         do iz=1,nz
            found=.false.
! **************************
            minV = 2000.
! **************************
            do while(.not. found)

!if (.not. found) write(0,*) 'first thing', ishot, iz, ix

               ! Draw a random number
               call random_number(r)
               r = r-.5
               r=r
               
               ! Calculate distance to the inner boundary
               if (ix .le. bndx) then
                  if (iz .lt. bndz) then   ! Corner position
                     d = sqrt(1.*((ix-bndx)*(ix-bndx)/bx + (iz-bndz)*(iz-bndz)/bz ) )
	          else if (iz .gt. (nz-bndx)) then  ! corner position
                     d = sqrt(1.*((ix-bndx)*(ix-bndx)/bx + (iz-(nz-bndz))*(iz-(nz-bndz))/bz))
                  else ! edge
                     d = (real(bndx-ix)/bndx)
                  end if


                  tempV = v2(iz,ix) + r*min(minV**2*d,maxv*maxv)
                  Ran1(iz,ix,ishot) = tempV

! write(0,*) 'RAN 1', ishot, iz, ix

               else if (ix .gt. (nx-bndx)) then
                  if (iz .lt. bndz) then   ! Corner position
                     d = sqrt(1.*((ix-(nx-bndx))*(ix-(ix-bndx))/bx + (iz-bndz)*(iz-bndz)/bz ) )
	          else if (iz .gt. (nz-bndx)) then  ! corner position
                     d = sqrt(1.*((ix-(nx-bndx))*(ix-(nx-bndx))/bx + (iz-(nz-bndz))*(iz-(nz-bndz))/bz))
                  else ! edge
                     d = (real((nx-bndx)-ix)/bndx)
                  end if

                  tempV = v2(iz,ix) + r*min(minV*2*d,maxv*maxv)
		  Ran1(iz, ix-nx+2*bndx,ishot) = tempV

! write(0,*) 'RAN 2', ishot, iz, ix

               else if (iz .le. bndz) then ! edge

                     d = real(bndz-iz)/bndz

                  tempV = v2(iz,ix) + r*min(minV**2*d,maxv*maxv)

		  Ran2(iz, ix-bndx, ishot) = tempV

! write(0,*) 'RAN 3', ishot, iz, ix

               else if (iz .gt. (nz-bndz)) then  ! edge

                     d = real(iz-nz+bndz)/bndz

                  tempV = v2(iz,ix) + r*min(minV**2*d,maxv*maxv)

		  Ran2(iz-nz+2*bndz, ix-bndx, ishot) = tempV


! write(0,*) 'RAN 4', ishot, iz, ix

               else

		  tempV = 18.e6

! write(0,*) 'RAN 5', ishot, iz, ix
               end if
               

              if ((tempV .lt. 40.) .or.  (tempV .gt. 36.e6)) then
	         found=.false. 
                 write(0,*) 'v2, tmpV', ishot, iz, ix
                 write(0,*) 'found false', v2(iz,ix), tempV
              else
                 found=.true.
              end if

              
if (.not. found) write(0,*) 'last thing before enddo'
            enddo
         end do
       end do   
           

write(0,*) 'end of RAN setup'

   end subroutine

!!----------------------------------------------------------
!! take 2nd derivative of the wavelet

  subroutine SecDer(souIn,souOut,nt,dt)

    integer :: it, nt
    real    :: dt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

    allocate(temp(nt))

    temp = souIn

    do it=1+1,nt-1
!       write(0,*) 'it', it
       souOut(it) = (2.*temp(it) - temp(it+1) - temp(it-1))
    end do

    souOut= souOut/dt**2
    
    ! clip the edge to be zero
    souOut(1:2) = 0.;
    souOUt(nt-2:nt) = 0.; 

    deallocate(temp)


  end subroutine 


!!----------------------------------------------------------
!! take 1st derivative of the  wavelet

  subroutine FirstDer(adj,souIn,souOut,nt,dt)

    logical :: adj
    integer :: i, nt
    real    :: dt, rdt
    real, dimension(:) :: souIn, souOut
    real, dimension(:), allocatable :: temp

   allocate(temp(nt))
   temp(1:nt) = souIn(1:nt)
    
    souOut(1:nt) = 0.
    
    
    if (.not. adj) then
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1)+ temp(i)
         souOut(i)   = souOut(i) - temp(i)
      end do
      
    else
       
      do i=1,nt-1
         souOut(i+1) = souOut(i+1) + temp(i+1)-temp(i)
      end do
         
    end if    

    deallocate(temp)


  end subroutine 
  
!!---------------------------------------------------------- 
!! one-way absorbing BC setup

  subroutine abcsetup(nx,nz,vp,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1,sqed)

    logical  :: sqed
    integer  :: ix, iz, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real     :: delp, dt,dx, dz
    real,  dimension(:,:) :: vp


    if (sqed) then
       do ix=1,nx
        
          delp = sqrt(vp(1+nop,ix)) *dt/dz

         ! Want absorbtive top boundary
          bndz0(ix) = (1.-delp)/(1.+delp)

          delp = sqrt(vp(nz-nop,ix)) *dt/dz
          bndz1(ix) = (1.-delp)/(1.+delp)
       end do

       do iz=1,nz
          delp = sqrt(vp(iz,1+nop)) *dt/dx
          bndx0(iz) = (1.-delp)/(1.+delp)

          delp = sqrt(vp(iz,nx-nop)) *dt/dx
          bndx1(iz) = (1.-delp)/(1.+delp)
       end do
    else
       do ix=1,nx
          delp = vp(1+nop,ix) *dt/dz

         ! Want absorbtive top boundary
          bndz0(ix) = (1.-delp)/(1.+delp)

          delp = vp(nz-nop,ix) *dt/dz
          bndz1(ix) = (1.-delp)/(1.+delp)
       end do

       do iz=1,nz
          delp = vp(iz,1+nop) *dt/dx
          bndx0(iz) = (1.-delp)/(1.+delp)

          delp = vp(iz,nx-nop) *dt/dx
          bndx1(iz) = (1.-delp)/(1.+delp)
       end do

    end if

  end subroutine 

!!---------------------------------------------------------- 
!! one-way absorbing BC setup
!! Constant velocity

  subroutine abcsetupV2(nx,nz,wvel,dt,dx, dz, nop, bndz0,bndz1,bndx0,bndx1)

    integer  :: ix, iz, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real     :: delp, dt,dx, dz, wvel   

       delp = wvel*dt/dz
       do ix=1,nx
         ! Want absorbtive top boundary
          bndz0(ix) = (1.-delp)/(1.+delp)
          bndz1(ix) = (1.-delp)/(1.+delp)
       end do

       delp = wvel*dt/dx
       do iz=1,nz
          bndx0(iz) = (1.-delp)/(1.+delp)
          bndx1(iz) = (1.-delp)/(1.+delp)
       end do


  end subroutine 

!!---------------------------------------------------------- 

!! Absorbing BC 1 - absorbtive top boundary (used to propagate the source wavefield)

  subroutine absorbingBC1(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)

    integer  :: iop, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real, dimension(:,:) :: um,uo

       !! absorbing BC
       do iop=1,nop
          uo( 1+nop-iop  ,:) = &
               um( 1+nop-iop+1,:) + bndz0 * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))

          uo(nz-nop+iop  ,:) = &
               um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))

          uo(:, 1+nop-iop  ) = &
               um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))

          uo(:,nx-nop+iop  ) = &
               um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
       end do


  end subroutine


!!---------------------------------------------------------- 
!! Absorbing BC 2 - Reflective top boundary (used to image the mirror signal)
!!                - assumes padding at the top

  subroutine absorbingBC2(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)

    integer  :: iop, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real, dimension(:,:) :: um,uo

  !  write(0,*) 'Calling BC2'

       !! absorbing BC
       do iop=1,nop
          uo( 1+nop-iop  ,:) = &
               um( 1+nop-iop+1,:) + 1. * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))

          uo(nz-nop+iop  ,:) = &
               um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))

          uo(:, 1+nop-iop  ) = &
               um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))

          uo(:,nx-nop+iop  ) = &
               um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
       end do

  
  end subroutine

!!-------------------------------------------------------------
!!---------------------------------------------------------- 
!! Absorbing BC  3 - Reflective top boundary (used to image the mirror signal)
!!                 - assumes no padding on the top                  

  subroutine absorbingBC3(nz,nx,nop,bndz0,bndz1,bndx0,bndx1,uo,um)

    integer  :: iop, nx, nz, nop
    real, dimension(:) :: bndz0,bndz1,bndx0,bndx1
    real, dimension(:,:) :: um,uo

  !  write(0,*) 'Calling BC2'

       !! absorbing BC
       do iop=1,nop
!          uo( 1+nop-iop  ,:) = &
!               um( 1+nop-iop+1,:) + 1. * (um( 1+nop-iop,:)-uo( 1+nop-iop+1,:))

          uo(nz-nop+iop  ,:) = &
               um(nz-nop+iop-1,:) + bndz1 * (um(nz-nop+iop,:)-uo(nz-nop+iop-1,:))

          uo(:, 1+nop-iop  ) = &
               um(:, 1+nop-iop+1) + bndx0 * (um(:, 1+nop-iop)-uo(:, 1+nop-iop+1))

          uo(:,nx-nop+iop  ) = &
               um(:,nx-nop+iop-1) + bndx1 * (um(:,nx-nop+iop)-uo(:,nx-nop+iop-1))
       end do

  
  end subroutine


!!--------------------------------------------------------------

  subroutine absorbingBCHale(adj, vel_grid,dt, vp2, um,uo,up)
  
     type(velo)    :: vel_grid
     logical           :: adj
     integer           :: nx, nz, iz, ix
     real              :: dt, dx, dz
     real              :: ovs, ov, dpdx, dpdxs, dpdt, dpdts, cosa, beta, gamma
     real, dimension(:,:)   :: vp2, uo, um, up
     
     nx = vel_grid%x%n
     nz = vel_grid%z%n
     dx = vel_grid%x%d
     dz = vel_grid%z%d


    !! Top boundary     
    iz=3
    do ix=1,nx
    
       ovs = 1.0/vp2(iz,ix);
       ov  = sqrt(ovs);
       
       if (ix == 1) then
       
          dpdx = (uo(iz,2) - uo(iz, 1))/dx
       else if (ix == nx) then
          dpdx = (uo(iz,nx) - uo(iz,nx-1))/dx
       else 
          dpdx = (uo(iz,ix+1) - uo(iz,ix-1))/(2.*dx)
       end if
       
       dpdt = (up(iz,ix) - um(iz,ix))/(2.*dt)
       dpdxs = dpdx*dpdx
       dpdts = dpdt*dpdt
       
       if (ovs*dpdts .gt. dpdxs) then
          cosa=sqrt(1.0-dpdxs/(ovs*dpdts));
       else
          cosa = 0.0;
       end if
       
       beta = ov*cosa*dz/dt
       gamma = (1.0-beta)/(1.0+beta);
      
       up(iz,ix) = gamma*(up(iz+1,ix)-uo(iz,ix))+uo(iz+1,ix)
       
       if (ix .eq. nx/2) then
          write(0,*) ovs, dpdt, dpdx, sum(uo)  
      !   write(0,*) 'gamma', cosa, gamma       
       end if
       
        
    end do 

      
    !! Bottom boundary
    iz=nz-1
    do ix=1,nx
    
       ovs = 1.0/vp2(iz,ix);
       ov  = sqrt(ovs);
       
       if (ix == 1) then
       
          dpdx = (uo(iz,ix+1) - uo(iz, ix))/dx
       else if (ix == nx) then
          dpdx = (uo(iz,ix) - uo(iz,ix-1))/dx
       else 
          dpdx = (uo(iz,ix+1) - uo(iz,ix-1))/(2.*dx)
       end if
       
       dpdt = (up(iz,ix) - um(iz,ix))/(2.*dt)
       dpdxs = dpdx*dpdx
       dpdts = dpdt*dpdt
       
       if (ovs*dpdts .gt. dpdxs) then
          cosa=sqrt(1.0-dpdxs/(ovs*dpdts));
       else
          cosa = 0.0;
       end if
       
       beta = ov*cosa*dz/dt
       gamma = (1.0-beta)/(1.0+beta);
       
       up(iz+1,ix) = gamma*(up(iz,ix)-uo(iz+1,ix))+uo(iz,ix)
       
    end do   
     
  
  
  end subroutine
  
!!--------------------------------------------------------------

  subroutine ashotsetup(izsou,shotdir,nx,vp,dt,dz,nop,bndz2)
    integer  :: ix, izsou, nx, nop, shotdir
    real, dimension(:) :: bndz2
    real     :: delp, dt, dz
    real,  dimension(:,:) :: vp

	write(0,*) 'ashot setup'

      
    do ix=1,nx
       delp = vp(izsou+1+nop,ix) *dt/dz
       bndz2(ix) = (1.-delp)/(1.+delp)
       
    end do

  end subroutine 

!!--------------------------------------------------------------

  subroutine ashotcondition(izsou,shotdir,nop,bndz2,uo,um)

    integer  :: izsou, iop, nop, shotdir
    real, dimension(:) :: bndz2
    real, dimension(:,:) :: um,uo

	

       !! absorbing shot
    do iop=1,nop
       if (shotdir .eq. 1) then
          uo( izsou+1 +nop-iop  ,:) = um( izsou+1+nop-iop+1,:) + &
               bndz2 * (um( izsou+1+nop-iop,:)-uo( izsou+1+nop-iop+1,:))
       else
          uo(izsou-nop+iop  ,:) = &
               um(izsou-nop+iop-1,:) + bndz2 * (um(izsou-nop+iop,:)-uo(izsou-nop+iop-1,:))

       end if

    end do
  end subroutine

!!--------------------------------------------------------------       
!! Taper Setup Gaussian

  subroutine tapersetup(bndz,bndx,ztapco,xtapco,ztaper,xtaper)

    integer  :: iz, ix, bndz, bndx
    real, dimension(:) :: xtaper,ztaper
    real                        :: xtapco,ztapco


       do iz=1,bndz
          ztaper(iz) = exp(-ztapco * iz*iz)
       end do

       do ix=1,bndx
          xtaper(ix) = exp(-xtapco * ix*ix)
       end do

  end subroutine

!!--------------------------------------------------------------
!! Taper Setup Cosine

  subroutine tapersetupCos(bndz,bndx,ztaper,xtaper, ctaper, cent, A)

    integer                     :: iz, ix, bndz, bndx
    real, dimension(:)          :: xtaper,ztaper
    real, dimension(:,:)        :: ctaper
    real                       :: dmax, dist, cent, A
    real                        :: pi=3.14159265359


     dmax = max(bndx, bndz)
!    dmax = sqrt(pi*bndx*bndx + bndz*bndx)
    


       ! Left and right edge
       do iz=1,bndz
          ztaper(iz) = cent + A*cos(pi*(bndz-iz)/bndz)
       end do

       ! top and bottom edge
       do ix=1,bndx
          xtaper(ix) = cent + A*cos(pi*(bndx-ix)/bndx)
       end do


       ! Corner
       do ix=1,bndx
          do iz=1,bndz
              dist = sqrt(1.001*(bndx-ix)*(bndx-ix)/bndx/bndx + 0.999999*(bndz-iz)*(bndz-iz)/bndz/bndz)
              ctaper(iz,ix) =  cent + A*cos(pi*dist)
!	 	ctaper(iz,ix) = xtaper(ix)*ztaper(iz)
          end do
       end do

  end subroutine

!-----------------------------------------------------------------

 subroutine WmodeltaperSetup2(wdz, wdx, ztapeMod, xtapeMod, ctapeMod, cent, A)

   integer                    :: iz, ix, wdx, wdz, izero
   real, dimension(:)         :: xtapeMod, ztapeMod
   real, dimension(:,:)       :: ctapeMod
   real                      :: dist, pi=3.14159265359, cent, A
   integer                       :: izb, ize, ixb, ixe, denomx, denomz
   
   izero = 3  

   ! Left and right edge
   izb = izero
   ize = wdz
   denomz = ize-izb  
   do iz=izb,ize
      dist = denomz - (izb-iz)
      ztapeMod(iz) = cent + A *cos(pi*dist/denomz)
   end do

   ztapeMod(1:izb-1) = ztapeMod(izb)

   ! Top and bottom edge
   ixb = izero
   ixe = wdx
   denomx = ixe - ixb
   do ix=ixb,ixe
      dist = denomx - (ixb-ix)
      xtapeMod(ix) = cent + A*cos(pi*dist/denomx)
   end do
 
   xtapeMod(1:ixb-1) = xtapeMod(ixb)

   ctapeMod = 0.;
       ! Corner
       do ix=1+izero,wdx
          do iz=1+izero,wdz
              dist = sqrt(1.001*(wdx-ix)*(wdx-ix)/denomx/denomx + 0.999999*(wdz-iz)*(wdz-iz)/denomz/denomz)
              if (dist .gt. 1) dist = 1.
              ctapeMod(iz,ix) =  cent + A*cos(pi*dist)
          end do
       end do

 

 end subroutine

!-----------------------------------------------------------------------------
! Data space tapering, that tapers off the edge of the data 

  subroutine WDataTaperSetup(WData,nrx,nt,wdx,cent,A)
   integer                    :: iz, irx, nrx,nt, wdx, izero
   real, dimension(:,:)       :: WData
   real                      :: dist, pi=3.14159265359, cent, A, amp
   integer                       :: ixb, ixe, denomx

   izero = 2  
   ixb = izero
   ixe = wdx
   denomx = ixe - ixb
   amp = 0.

   ! Left and Right edge
     do irx=ixb,ixe
        dist = denomx - (ixb-irx)
        amp = (cent + A*cos(pi*dist/denomx))
        WData(irx,:) = amp*WData(irx,:)
        WData(nrx-irx+1,:) = amp*WData(nrx-irx+1,:) 
     end do

       WData(1:ixb-1,:)=0.
       WData(nrx-ixb+2:nrx,:) = 0.

 end subroutine


!!----------------------------------------------------------------
!! Tapering
!! Version 2 has reflective top boundary 
!! no padding is applied to the top boundary
!! nz = nnz+1*bndz + nop

  subroutine taperingCosV2(nz,nx,nop,bndz,bndx,ztaper,xtaper,ctaper,uo)

    ! array in z : nop+nnz+bndz+leftover

    integer  :: iz, ix, nz,nx,bndz,bndx,nnz, nop
    real, dimension(:) :: xtaper,ztaper
    real, dimension(:,:) :: uo, ctaper

    nnz = nz-nop-bndz

       ! Corner
       do ix=1,bndx
          do iz=1,bndz
              uo(nz-iz+1,nx-ix+1) = uo(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              uo(nz-iz+1,ix) = uo(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       do iz=nop+1,nz-bndz
          uo(iz, 1:bndx) = uo(iz, 1:bndx) * xtaper(1:bndx:+1);          ! left
	  uo(iz,nx-bndx+1:nx) = uo(iz,nx-bndx+1:nx) * xtaper(bndx:1:-1) ! right
       end do

       do ix=1,nx-bndx
          uo(nz-bndz+1:nz,ix) = uo(nz-bndz+1:nz,ix) * ztaper(bndz:1:-1) ! bottom
       end do

  end subroutine

!!----------------------------------------------------------------
!! Tapering


  subroutine taperingCos(nz,nx,bndz,bndx,ztaper,xtaper,ctaper,uo)

    integer  :: iz, ix, nz,nx,bndz,bndx
    real, dimension(:) :: xtaper,ztaper
    real, dimension(:,:) :: uo, ctaper

       ! Corner
       do ix=1,bndx
          do iz=1,bndz

              uo(iz,ix) = uo(iz,ix)*ctaper(iz,ix);         ! top left
              uo(iz,nx-ix+1) = uo(iz,nx-ix+1)*ctaper(iz,ix)          ! top right
              uo(nz-iz+1,nx-ix+1) = uo(nz-iz+1,nx-ix+1)*ctaper(iz,ix)   ! bottom right
              uo(nz-iz+1,ix) = uo(nz-iz+1,ix)*ctaper(iz,ix)          ! bottom left
          end do
       end do

       do iz=bndz+1,nz-bndz

          uo(iz, 1:bndx) = uo(iz, 1:bndx) * xtaper(1:bndx:+1);              !left
	  uo(iz,nx-bndx+1:nx) = uo(iz,nx-bndx+1:nx) * xtaper(bndx:1:-1)     !right
       end do

       do ix=bndx+1,nx-bndx

          uo(1:bndz,ix) = uo(1:bndz,ix) * ztaper(1:bndz:+1);             ! top
          uo(nz-bndz+1:nz,ix) = uo(nz-bndz+1:nz,ix) * ztaper(bndz:1:-1)  ! bottom
       end do

  end subroutine


!!--------------------------------------------------------------------------
!! 2nd Order Laplacian

  subroutine laplac2(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: idx, idz, dt, dt2, sum, xscale, zscale

    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz
    sum = 2.*(xscale+zscale)

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=2,nx-1       
          do iz=2,nz-1
          
             up(iz,ix) = 2.*uo(iz,ix) - um(iz,ix) + &
                  vp2(iz,ix)*((uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale  + &                 
                   (uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale -uo(iz,ix)*sum) 

          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=2,nx-1
          do iz=2,nz-1         

             up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + &
                         vp2(iz,ix)*((uo(iz,ix-1)+uo(iz,ix+1)-2.*uo(iz,ix))*xscale + &
                                     (uo(iz-1,ix)+uo(iz+1,ix)-2.*uo(iz,ix))*zscale )
                                     
!            up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + &
!               ((vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1) - 2.*vp2(iz,ix)*uo(iz,ix))*xscale + &
!                (vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  ) - 2.*vp2(iz,ix)*uo(iz,ix))*zscale)

                
          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine


!!--------------------------------------------------------------------------

!!--------------------------------------------------------------------------
!! 4th Order Laplacian
!! source has been scaled by density 'ro'


  subroutine laplac4(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: c0, c1, c2, idx, idz, dt, dt2
    real       :: xscale, zscale, sum

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz
    sum    = xscale+zscale

    if (iit .eq. 100) then
       write(0,*) 'c0, c1, c2', c0, c1, c2
       write(0,*) 'c0x, c1x, c2x', c0, c1*xscale, c2*xscale       
    end if

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=3,nx-2       
          do iz=3,nz-2
          

             up(iz,ix) = vp2(iz,ix)* &
                  (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale + &
                  c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*xscale + &
                  c0* uo(iz,  ix  ) * sum                  + &
                  c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale + &
                  c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*zscale) + &
   	          2*uo(iz,ix) - um(iz,ix)
   	          
!   	  if ((iit .eq. 100) .and. (iz .lt. 3*nz/4) .and. (iz .gt. nz/4) .and. (ix .lt. 3*nx/4) .and. (ix .gt. nx/4)) then
!   	     write(0,*)  iz,ix, up(iz,ix), um(iz,ix), up(iz,ix)           	          
!          end if
          
          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=3,nx-2
          do iz=3,nz-2         
            up(iz,ix) = &
               (c1*(vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1))*xscale + &
                c2*(vp2(iz,ix-2)*uo(iz  ,ix-2) + vp2(iz,ix+2)*uo(iz  ,ix+2))*xscale + &
                c0* vp2(iz,ix)*uo(iz,  ix  ) * sum                                  + &
                c1*(vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  ))*zscale + &
                c2*(vp2(iz-2,ix)*uo(iz-2,ix  ) + vp2(iz+2,ix)*uo(iz+2,ix  ))*zscale)+ & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine

!!--------------------------------------------------------------------------

!!--------------------------------------------------------------------------
!! 4th Order Laplacian
!! source has been scaled by density 'ro'
!! A faster version of laplac4 when velocity is constant


  subroutine laplac4V2(adj, um, uo,up, wvel, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um
    real       :: c0, c1, c2, idx, idz, dt, dt2, wvel
    real       :: xscale, zscale, total

    c0 = -30./12.
    c1 = +16./12.
    c2 = - 1./12.
    dt2 = dt**2
    xscale = wvel*dt2*idx
    zscale = wvel*dt2*idz
    total    = xscale+zscale

    

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=3,nx-2       
          do iz=3,nz-2
          

             up(iz,ix) =  &
                  (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale + &
                  c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*xscale + &
                  c0* uo(iz,  ix  ) * total                 + &
                  c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale + &
                  c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*zscale) + &
   	          2.*uo(iz,ix) - um(iz,ix)
          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=3,nx-2
          do iz=3,nz-2         
            up(iz,ix) = &
               (c1*(uo(iz  ,ix-1) + uo(iz  ,ix+1))*xscale + &
                c2*(uo(iz  ,ix-2) + uo(iz  ,ix+2))*xscale + &
                c0* uo(iz,  ix  ) * total                 + &
                c1*(uo(iz-1,ix  ) + uo(iz+1,ix  ))*zscale + &
                c2*(uo(iz-2,ix  ) + uo(iz+2,ix  ))*zscale)+ & 
                2.*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine

!!--------------------------------------------------------------------------
!!--------------------------------------------------------------------------
!! 6th Order approximation to the Laplacian
!! Coefficients are obtained from HaoHuan GPU code


  subroutine laplac6(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: idx, idz, dt, dt2
    real       :: c0, c1x, c2x, c3x, c4x, c1z, c2z, c3z, c4z
    real       :: xscale, zscale, total


    dt2 = dt**2
    xscale = 1.5*dt2*idx
    zscale = 1.5*dt2*idz


! 6th order in space
    c0 = -49./27.*(xscale+zscale)
    c1x = xscale
    c1z = zscale
    c2x = -1./10.*(xscale)
    c2z = -1./10.*(zscale)
    c3x =1./135.*(zscale)
    c3z =1./135.*(zscale)



    if (iit .eq. 100) then
       write(0,*) 'calling laplac6'
       write(0,*) 'c0, c1, c2, c3', c0, c1x, c2x, c3x
    end if

    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)))
                          			 

          
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop   

       	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*c0*uo(iz,ix) + &
			c1z*(vp2(iz+1,ix)*uo(iz+1,ix) + vp2(iz-1,ix)*uo(iz-1,ix)) + c1x*(vp2(iz,ix+1)*uo(iz,ix+1) + vp2(iz,ix-1)*uo(iz,ix-1)) + &
			c2z*(vp2(iz+2,ix)*uo(iz+2,ix) + vp2(iz-2,ix)*uo(iz-2,ix)) + c2x*(vp2(iz,ix+2)*uo(iz,ix+2) + vp2(iz,ix-2)*uo(iz,ix-2)) + &
			c3z*(vp2(iz+3,ix)*uo(iz+3,ix) + vp2(iz-3,ix)*uo(iz-3,ix)) + c3x*(vp2(iz,ix+3)*uo(iz,ix+3) + vp2(iz,ix-3)*uo(iz,ix-3))
                        
      

          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine

!!--------------------------------------------------------------------------
!!--------------------------------------------------------------------------
!! 6th Order approximation to the Laplacian
!! Coefficients are obtained from HaoHuan GPU code
!! This is a quick version where we have a constant velocity

  subroutine laplac6V2(adj, um, uo,up, vw, dt, nx, nz, idx, idz, nop, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um
    real       :: idx, idz, dt, dt2, vw
    real       :: c0, c1x, c2x, c3x, c4x, c1z, c2z, c3z, c4z
    real       :: xscale, zscale, total


    dt2 = dt**2
    xscale = 1.5*dt2*idx
    zscale = 1.5*dt2*idz


! 6th order in space
    c0 = -49./27.*(xscale+zscale)
    c1x = xscale
    c1z = zscale
    c2x = -1./10.*(xscale)
    c2z = -1./10.*(zscale)
    c3x =1./135.*(zscale)
    c3z =1./135.*(zscale)



    if (iit .eq. 100) then
       write(0,*) 'calling laplac6'
       write(0,*) 'c0, c1, c2, c3', c0, c1x, c2x, c3x
    end if

    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vw*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)))
                          			 

          
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop   

       	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vw*(c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)))
                        

          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine

!!--------------------------------------------------------------------------
!! 8th Order approximation to the Laplacian
!! Coefficients are obtained from Chris Leader/Abdullah's code


  subroutine laplac8V2(adj, um, uo,up, vw, dt, nx, nz, idx, idz, nop, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um
    real       :: idx, idz, dt, dt2, vw
    real       :: c0, c1x, c2x, c3x, c4x, c1z, c2z, c3z, c4z
    real       :: xscale, zscale, total


    dt2 = dt**2
    xscale = 1.6*dt2*idx
    zscale = 1.6*dt2*idz


    c0 = -1025./576.*(xscale+zscale)
    c1x = 1.*xscale
    c1z = 1.*zscale
    c2x = -1./8.*xscale
    c2z = -1./8.*zscale
    c3x = 1./63.*xscale
    c3z = 1./63.*zscale
    c4x = -1./896.*xscale
    c4z = -1./896.*zscale

     

    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vw*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)) + &
			c4z*(uo(iz+4,ix) + uo(iz-4,ix)) + c4x*(uo(iz,ix+4) + uo(iz,ix-4)))
                          			 

          
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop   

      	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vw*(c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)) + &
                        c4z*(uo(iz+4,ix) + uo(iz-4,ix)) + c4x*(uo(iz,ix+4) + uo(iz,ix-4)))


          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine
  
!!--------------------------------------------------------------------------
!! 8th Order approximation to the Laplacian
!! Coefficients are obtained from Chris Leader/Abdullah's code
!! This is a quick version with constant velocity

  subroutine laplac8(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: idx, idz, dt, dt2
    real       :: c0, c1x, c2x, c3x, c4x, c1z, c2z, c3z, c4z
    real       :: xscale, zscale, total


    dt2 = dt**2
    xscale = 1.6*dt2*idx
    zscale = 1.6*dt2*idz


    c0 = -1025./576.*(xscale+zscale)
    c1x = 1.*xscale
    c1z = 1.*zscale
    c2x = -1./8.*xscale
    c2z = -1./8.*zscale
    c3x = 1./63.*xscale
    c3z = 1./63.*zscale
    c4x = -1./896.*xscale
    c4z = -1./896.*zscale

     

 !   write(0,*) 'xscale, zscale, sum, vp2', xscale, zscale, sum, vp2(nz/2,nx/2) 

    if (iit .eq. 1) then
       write(0,*) 'Inside laplac 8'
       write(0,*) 'idx, idz, dt', idx, idz, dt
       write(0,*) 'min max V', minval(vp2), maxval(vp2)
    end if


    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)) + &
			c4z*(uo(iz+4,ix) + uo(iz-4,ix)) + c4x*(uo(iz,ix+4) + uo(iz,ix-4)))
                          			 

          
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop   

      	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*c0*uo(iz,ix) + &
			c1z*(vp2(iz+1,ix)*uo(iz+1,ix) + vp2(iz-1,ix)*uo(iz-1,ix)) + c1x*(vp2(iz,ix+1)*uo(iz,ix+1) + vp2(iz,ix-1)*uo(iz,ix-1)) + &
			c2z*(vp2(iz+2,ix)*uo(iz+2,ix) + vp2(iz-2,ix)*uo(iz-2,ix)) + c2x*(vp2(iz,ix+2)*uo(iz,ix+2) + vp2(iz,ix-2)*uo(iz,ix-2)) + &
			c3z*(vp2(iz+3,ix)*uo(iz+3,ix) + vp2(iz-3,ix)*uo(iz-3,ix)) + c3x*(vp2(iz,ix+3)*uo(iz,ix+3) + vp2(iz,ix-3)*uo(iz,ix-3)) + &
                        c4z*(vp2(iz+4,ix)*uo(iz+4,ix) + vp2(iz-4,ix)*uo(iz-4,ix)) + c4x*(vp2(iz,ix+4)*uo(iz,ix+4) + vp2(iz,ix-4)*uo(iz,ix-4))


          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine
  
!!--------------------------------------------------------------------------
!!--------------------------------------------------------------------------
!! 10th Order approximation to the Laplacian
!! Coefficients are obtained from 
!! John Etgen's code Acoustic2d


  subroutine laplac10(adj, um, uo,up, vp2, dt, nx, nz, idx, idz, nop)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2
    real       :: idx, idz, dt, dt2
    real       :: c0, c1x, c2x, c3x, c4x, c5x, c1z, c2z, c3z, c4z, c5z
    real       :: xscale, zscale, total


    dt2 = dt**2
    xscale = dt2*idx
    zscale = dt2*idz

    c0 = -2.927222222*(xscale+zscale)
    c1x = 3.333333333*xscale/2. 
    c1z = 3.333333333*zscale/2.
    c2x = -.4761904762*xscale/2.
    c2z = -.4761904762*zscale/2.   
    c3x = 0.793650794*xscale/2.
    c3z = 0.793650794*zscale/2.    
    c4x = -.0099206349*xscale/2.
    c4z = -.0099206349*zscale/2.
    c5x = 0.0006349206*xscale/2.
    c5z = 0.0006349206*zscale/2.


 !   write(0,*) 'xscale, zscale, sum, vp2', xscale, zscale, sum, vp2(nz/2,nx/2) 


    if (adj) then    !! RTM
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          

	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3)) + &
			c4z*(uo(iz+4,ix) + uo(iz-4,ix)) + c4x*(uo(iz,ix+4) + uo(iz,ix-4)) + &                            
			c5z*(uo(iz+5,ix) + uo(iz-5,ix)) + c5x*(uo(iz,ix+5) + uo(iz,ix-5)))
                          			 


!    write(0,*) 'c1*xscale*vp, vp', c1*xscale*vp2(iz,ix), vp2(iz,ix)
!     write(0,*) 'uo(ix+1) uo(ix-1)', uo(iz,ix-1)+uo(iz,ix+1)
          end do
       end do
!!    !$OMP end parallel do   
    else 
!!    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop   
!	    up(iz,ix) = 2*uo(iz,ix) - um(iz,ix) + xscale*( c0*vp2(iz,ix)*uo(iz,ix)  + &
!                        c1*(uo(iz+1,ix)*vp2(iz+1,ix) + uo(iz-1,ix)*vp2(iz-1,ix) + uo(iz,ix+1)*vp2(iz,ix+1) + uo(iz,ix-1)*vp2(iz,ix-1)) + &
!                        c2*(uo(iz+2,ix)*vp2(iz+2,ix) + uo(iz-2,ix)*vp2(iz-2,ix) + uo(iz,ix+2)*vp2(iz,ix+2) + uo(iz,ix-2)*vp2(iz,ix-2)) ) + &
!         c3*(uo(iz+2,ix+1)*vp2(iz+2,ix+1) + uo(iz+1,ix+2)*vp2(iz+1,ix+2) + uo(iz-2,ix+1)*vp2(iz-2,ix+1) + uo(iz-1,ix+2)*vp2(iz-1,ix+2)) + &
!         c3*(uo(iz-2,ix-1)*vp2(iz-2,ix-1) + uo(iz-1,ix-2)*vp2(iz-1,ix-2) + uo(iz+2,ix-1)*vp2(iz+2,ix-1) + uo(iz+1,ix-2)*vp2(iz+1,ix-2)) + &
                         
      

          end do
        end do
!!   !$OMP end parallel do
    end if

             

  end subroutine
  
!!--------------------------------------------------------------------------
!! Non-reflecting propgation with 4th Order Laplacian
!! Reference: Geophysics Vol 71 No. 6  (2006)  Fletcher et. al. 

  subroutine laplacNR4(adj, um, uo,up, vp2, vp, vpx, vpz, dt, nx, nz, dx, dz, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um, vp2,vpx, vpz, vp
    real       :: a0, a1x, a2x, a3x, a1z, a2z, a3z
    real       :: c0, c1x, c2x, c3x, c1z, c2z, c3z, dx, dz, dt, dt2
    real       :: xscale, zscale, xscale2, zscale2


    nop=3
    dt2 = dt**2
    xscale = dt2/(dx*dx)
    zscale = dt2/(dz*dz)
    xscale2= dt2/dx
    zscale2= dt2/dz


    c0 = -30./12.*(xscale+zscale)
    c1x = +16./12.*(xscale)
    c2x = - 1./12.*(xscale)
    c1z = +16./12.*(zscale)
    c2z = - 1./12.*(zscale)
    
    a0 = 0.
    a1x = 45./60.*xscale2
    a2x = -9./60.*xscale2
    a3x = 1./60.*xscale2
    a1z = 45./60.*zscale2
    a2z = -9./60.*zscale2
    a3z = 1./60.*zscale2

!    a0 = 0.
!    a1x = 0.
!    a2x = 0.
!    a3x = 0.
!    a1z = 0.
!    a2z = 0.
!    a3z = 0.
    

!    if (iit .gt. 151) then
!       write(0,*) 'a0, a1x, a2x, a3x', a0, a1x, a2x, a3x
!       call erexit('Exit here')
!    end if

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          
	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2))) + &
                        vp(iz,ix)*((a1x*(uo(iz, ix+1) - uo(iz ,ix-1)) + &
   	                            a2x*(uo(iz, ix+2) - uo(iz, ix-2)) + &
   	                            a3x*(uo(iz, ix+3) - uo(iz, ix-3)))*vpx(iz,ix) + &
   	                           (a1z*(uo(iz+1, ix) - uo(iz-1, ix)) + &
   	                            a2z*(uo(iz+2, ix) - uo(iz-2, ix))  + &
   	                            a3z*(uo(iz+3, ix) - uo(iz-3, ix)))*vpz(iz,ix))
   	            

          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop         
            up(iz,ix) = &
               (c1x*(vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1)) + &
                c2x*(vp2(iz,ix-2)*uo(iz  ,ix-2) + vp2(iz,ix+2)*uo(iz  ,ix+2)) + &
                c0* vp2(iz,ix)*uo(iz,  ix  )                                  + &
                c1z*(vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  )) + &
                c2z*(vp2(iz-2,ix)*uo(iz-2,ix  ) + vp2(iz+2,ix)*uo(iz+2,ix  )))+ & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine
!!--------------------------------------------------------------------------
!! Non-reflecting propgation with 6th Order Laplacian
!! Reference: Geophysics Vol 71 No. 6  (2006)  Fletcher et. al. 

  subroutine laplacNR6(adj, um, uo,up, vp2, vp, vpx, vpz, dt, nx, nz, dx, dz, iit)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz, iit
    real, dimension(:,:) :: uo,up, um, vp2,vpx, vpz, vp
    real       :: a0, a1x, a2x, a3x, a1z, a2z, a3z
    real       :: c0, c1x, c2x, c3x, c1z, c2z, c3z, dx, dz, dt, dt2
    real       :: xscale, zscale, xscale2, zscale2


    nop=3
    dt2 = dt**2
    xscale = dt2/(dx*dx)
    zscale = dt2/(dz*dz)
    xscale2= dt2/dx
    zscale2= dt2/dz

    c0 = -49./18.*(xscale+zscale)/2.
    c1x = 1.5*(xscale)/2.
    c2x = - 3./20.*(xscale)/2.
    c3x =1./90.*(zscale)/2.
    c1z = 1.5*(zscale)/2.
    c2z = -3./20.*(zscale)/2.
    c3z =1./90.*(zscale)/2.



    a0 = 0.
    a1x = 45./60.*xscale2
    a2x = -9./60.*xscale2
    a3x = 1./60.*xscale2
    a1z = 45./60.*zscale2
    a2z = -9./60.*zscale2
    a3z = 1./60.*zscale2

!    a0 = 0.
!    a1x = 0.
!    a2x = 0.
!    a3x = 0.
!    a1z = 0.
!    a2z = 0.
!    a3z = 0.
    

!    if (iit .gt. 151) then
!       write(0,*) 'a0, a1x, a2x, a3x', a0, a1x, a2x, a3x
!       call erexit('Exit here')
!    end if

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          
	      up(iz,ix) =  2.*uo(iz,ix) - um(iz,ix) + vp2(iz,ix)*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3))) + &
                        vp(iz,ix)*((a1x*(uo(iz, ix+1) - uo(iz ,ix-1)) + &
   	                            a2x*(uo(iz, ix+2) - uo(iz, ix-2)) + &
   	                            a3x*(uo(iz, ix+3) - uo(iz, ix-3)))*vpx(iz,ix) + &
   	                           (a1z*(uo(iz+1, ix) - uo(iz-1, ix)) + &
   	                            a2z*(uo(iz+2, ix) - uo(iz-2, ix))  + &
   	                            a3z*(uo(iz+3, ix) - uo(iz-3, ix)))*vpz(iz,ix))
   	            

          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop         
            up(iz,ix) = &
               (c1x*(vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1)) + &
                c2x*(vp2(iz,ix-2)*uo(iz  ,ix-2) + vp2(iz,ix+2)*uo(iz  ,ix+2)) + &
                c0* vp2(iz,ix)*uo(iz,  ix  )                                  + &
                c1z*(vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  )) + &
                c2z*(vp2(iz-2,ix)*uo(iz-2,ix  ) + vp2(iz+2,ix)*uo(iz+2,ix  )))+ & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine
  
!! ___________________________________________________________________________


!!--------------------------------------------------------------------------
!! Non-reflecting propgation with 6th Order Laplacian with DAMPING
!! Reference: Geophysics Vol 71 No. 6  (2006)  Fletcher et. al. 

  subroutine laplacNRDamp6(adj, um, uo,up, vp2, vp, vpx, vpz, eps, dt, nx, nz, dx, dz)

    logical    :: adj
    integer    :: iz, ix, nop, nx, nz
    real, dimension(:,:) :: uo,up, um, vp2,vpx, vpz, vp, eps
    real       :: a0, a1x, a2x, a3x, a1z, a2z, a3z, fac
    real       :: c0, c1x, c2x, c3x, c1z, c2z, c3z, dx, dz, dt, dt2
    real       :: xscale, zscale, xscale2, zscale2


    nop=3
    dt2 = dt**2
    xscale = dt2/(dx*dx)
    zscale = dt2/(dz*dz)
    xscale2= dt2/dx
    zscale2= dt2/dz

    c0 = -49./18.*(xscale+zscale)/2.
    c1x = 1.5*(xscale)/2.
    c2x = - 3./20.*(xscale)/2.
    c3x =1./90.*(zscale)/2.
    c1z = 1.5*(zscale)/2.
    c2z = -3./20.*(zscale)/2.
    c3z =1./90.*(zscale)/2.

    a0 = 0.
    a1x = 45./60.*xscale2
    a2x = -9./60.*xscale2
    a3x = 1./60.*xscale2
    a1z = 45./60.*zscale2
    a2z = -9./60.*zscale2
    a3z = 1./60.*zscale2


    a0 = 0.
    a1x = 0.
    a2x = 0.
    a3x = 0.
    a1z = 0.
    a2z = 0.
    a3z = 0.

    if (adj) then    !! RTM
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
       do ix=1+nop,nx-nop       
          do iz=1+nop,nz-nop
          
            fac = 2.*dt*eps(iz,ix)

	      up(iz,ix) =  2.*uo(iz,ix) - (fac -1.)*um(iz,ix) + vp2(iz,ix)*( c0*uo(iz,ix) + &
			c1z*(uo(iz+1,ix) + uo(iz-1,ix)) + c1x*(uo(iz,ix+1) + uo(iz,ix-1)) + &
			c2z*(uo(iz+2,ix) + uo(iz-2,ix)) + c2x*(uo(iz,ix+2) + uo(iz,ix-2)) + &
			c3z*(uo(iz+3,ix) + uo(iz-3,ix)) + c3x*(uo(iz,ix+3) + uo(iz,ix-3))) + &
                        vp(iz,ix)*((a1x*(uo(iz, ix+1) - uo(iz ,ix-1)) + &
   	                            a2x*(uo(iz, ix+2) - uo(iz, ix-2)) + &
   	                            a3x*(uo(iz, ix+3) - uo(iz, ix-3)))*vpx(iz,ix) + &
   	                           (a1z*(uo(iz+1, ix) - uo(iz-1, ix)) + &
   	                            a2z*(uo(iz+2, ix) - uo(iz-2, ix))  + &
   	                            a3z*(uo(iz+3, ix) - uo(iz-3, ix)))*vpz(iz,ix))
   	          
   	      up(iz,ix) = up(iz,ix)/(1.+fac)
   	          
          end do
       end do
    !$OMP end parallel do   
    else 
    !$OMP parallel do ordered schedule(dynamic,1) private(ix,iz)
        do ix=1+nop,nx-nop
          do iz=1+nop,nz-nop         
            up(iz,ix) = &
               (c1x*(vp2(iz,ix-1)*uo(iz  ,ix-1) + vp2(iz,ix+1)*uo(iz  ,ix+1)) + &
                c2x*(vp2(iz,ix-2)*uo(iz  ,ix-2) + vp2(iz,ix+2)*uo(iz  ,ix+2)) + &
                c0* vp2(iz,ix)*uo(iz,  ix  )                               + &
                c1z*(vp2(iz-1,ix)*uo(iz-1,ix  ) + vp2(iz+1,ix)*uo(iz+1,ix  )) + &
                c2z*(vp2(iz-2,ix)*uo(iz-2,ix  ) + vp2(iz+2,ix)*uo(iz+2,ix  )))+ & 
                2*uo(iz,ix) - um(iz,ix)

          end do
        end do
   !$OMP end parallel do
    end if

  end subroutine
  
!! ___________________________________________________________________________


end module
